



# liquidSVM Overview

This document describes core features and the learning process of liquidSVM as well as the 
design decisions behind it.



## Performance


liquidSVM provides state-of-the-art runtime performance for small, medium, and large scale problems.
This is achieved by the following ideas implemented in liquidSVM:




- Cross-validation: is integrated in the core of the package, since many of the calculations done for one hyper-parameter tuple can be reused for the next! Also, the solution of the previous tuple is a very good warm-start for the next tuple.
- Benefits of modern hardware: key parts are vectorized and multi-threaded. Moreover, GPUs can be used to compute kernel matrices.
- Scalability: several random and  spatial partitioning methods are included for data sets up to  millions of samples.
- Good default values: in many cases the default values already give very good results, making manual 
 fine-tuning in many cases unnecessary. However, there are also several levels of configurations for advanced users and experts.
- No offset: the offset of an SVM is not needed for common kernels like the Gaussian RBF and  omitting the offset gives much more flexibility for the design of solvers. This argument is false for linear kernels, and hence they 
are not included.

Most of these ideas are not only empirically verified but also based on the mathematical analysis of the last 15 years.



## The Learning Process: Three Phases

The general learning process of liquidSVM is split into three phases:

`train - the work horse`  calculates SVM solutions for all hyper-parameter combinations on all specified subsets of the data set and computes the resulting validation errors. During this phase, the data is automatically split consecutively into:

- Tasks: e.g.  multi-class problems are solved by several  binary classification problems, each defining a task. 
  For example, for a problem with labels 1, 2, 3, and 4, there are 6 one-vs-one tasks, or 4 one-vs-rest tasks.
  In the first case, the 1-vs-2 task only uses the samples with label 1 or 2, while each one-vs-rest task consists of the 
  entire data set.
  Another source for tasks is bootstrapping.

- cells: the data set of each task can be partitioned into cells to speed up, or just  enable training on large tasks. Random cells  are possible, but it is recommended to use cells that are based on spatial information. 
    Here, different choices are available each balancing speed, accuracy, and  theoretical guarantees  differently.


- Folds: every cell is split into k folds for k-fold cross validation. Several methods for creating the folds exists such as 
  random, stratified random, and random respecting group information.


On each combination of k-1 folds, the SVM solutions are then computed for all tuples in the specified hyper-parameter grid comprised of the regularization parameter, the kernel parameter, and in some cases also a weight parameter. 
These solutions are then 
validated on the remaining fold of the cell. Required distance and kernel matrices are automatically stored and re-used 
by default to avoid unnecessary repetitions of costly calculations. Alternatively, one can also invoke a caching 
of parts of the matrices.



`select - the cherry picker` identifies the hyper-parameter tuples with the best validation errors for each cell
and prepares the 
resulting SVM solutions. During this phase, the tasks, cells, and folds of the training phase are 
automatically
re-used, but 
additional tasks may also be created. Choices for this phase include:

- Best parameters - two possibilities: 
Either on each combination of k-1 folds  the best parameter tuple is identified
and the resulting k SVM solutions will be combined  during testing, or the validations errors for each tuple are averaged over the k combinations.
In this case,  the SVM solution will be re-computed with the resulting best hyper-parameter tuple.

- Use stored SVM solutions: It is possible to store all SVM solutions that are generated during the 
training phase. If you have done this, they can be simply be picked during the selection phase, otherwise,
they need to e re-computed. But be aware: while storing sounds attractive, usually 99% of the stored solutions
will never be touched again, while recomputing the required 1% typically costs only 10% to 15%  extra computing time.
Also, if you determine the best hyper-parameter tuple by averaging over the k validation errors, the store solutions
are anyway useless.

- New tasks: For example, quantile regression is hardly ever done for a single tau-quantile, but usually 
for several tau-quantiles, instead. During the training phase, the different values for tau are viewed as 
a hyper-parameter (as a weight to be precise) to avoid extra kernel matrix computations. During the selection phase, however, 
all the taus you pick create an individual, new task.


`test - the inference step`  combines the solutions found during the selection phase and applies this combination to test data. 
During this phase, the tasks, cells, and folds of the training, respectively selection, phase are 
automatically
re-used, and the 
combinations are consecutively constructed and applied to the test data set in this order:


- Folds: If you have chosen to use k SVM solutions, then these solutions are combined  to form a single predictor 
  by either (weighted) averaging
 for regression-like scenarios, or by (weighted) majority votes for classification like scenarios. The optional weighting 
 depends on the validation errors. In case you have chosen to use a single SVM solution during the selection phase, this will 
 serve as the predictor.

- Cells: Some partitioning strategies such as random partitioning or overlapping spatially defined cells require
  combining predictors of more than one cell. This is done by averaging or by a majority vote depending on the learning scenario. Some other spatial partition strategies actually create a disjoint partition of the input space into 
  cells, so that for each test sample, only the cell, in which the sample falls into, needs to be considered.


- Tasks: Some learning scenarios need to combine the predictions of all tasks. For example, 
  in the above one-vs-one multi-class problem with 4 labels, the resulting predictors of the 
  6 tasks need to be combined
  to produce a single predictor, and the same is true for the 4 predictors  of the one-vs-rest approach. Other tasks, such as quantile regression with more than one tau-quantile, do not have a single common task, and therefore 
  several predictors are created.

These steps are automatically processed without further adjustment, but you have the choice 
between weighted and unweighted combinations during the fold combination phase. 
In addition,  you may also choose the loss function that is
used to compute the prediction error on the test set.
 





## Configuration





There are several layers of configuration that address different levels of user expertise: 

- Scenario selection - the auto drive: Just pick the learning scenario you wish to solve, the rest is done automatically. 
Choices include (weighted) binary classification,  multi-class  classification, least squares regression, quantile regression, expectile regression, ROC estimation, and Neyman-Pearson type classification.
- High level configuration - switching gears: What is more important, speed or accuracy? Is memory an issue? This configuration level helps you finding   the right balance for your problem at hand.
- Advanced parameters - drive on your own: Do you have a learning scenario that does 
not fit well to the pre-defined ones? Do you want to 
have control over the hyper-parameter grid, the fold generation, or the combinations built during the test phase? 
In this level, you can make these choices and others on your own.
- Low-Level configuration - fine tune all aspects: Invoke GPUs or caching for matrix computations, fine tune solver parameters, pin liquidSVM to specific cores of your CPU, control warm start options, ... Now you have full control.




