// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (COMMAND_OPTIONS_CLUSTER_CLASS_CPP) 
	#define COMMAND_OPTIONS_CLUSTER_CLASS_CPP


	
	
#include "sources/cluster/command_line/command_line_cluster.h"


#include "sources/shared/basic_types/vector.h"

#include "sources/cluster/solver/cluster_compare.h" 


	
//**********************************************************************************************************************************


Tcommand_line_parser_cluster::Tcommand_line_parser_cluster()
{
	command_name = "dbc";
	
	set_tau = false;
}


//**********************************************************************************************************************************


void Tcommand_line_parser_cluster::parse()
{
	check_parameter_list_size();
	for(current_position=1; current_position<parameter_list_size; current_position++)
		if (Tcommand_line_parser::parse("-d-h-r-T") == false)
		{
			if(parameter_list[current_position][0] != '-') 
				break;
			if (string(parameter_list[current_position]).size() > 2)
				Tcommand_line_parser::exit_with_help(ERROR_clp_gen_unknown_option);
			
			switch(parameter_list[current_position][1])
			{
				case 'D':
					if (next_parameter_equals('[') == true)
					{
						cluster_control.deltas = get_next_list(ERROR_clp_clu_D, 0.0);
						if (cluster_control.deltas[argmin(cluster_control.deltas)] <= 0.0)
							Tcommand_line_parser::exit_with_help(ERROR_clp_clu_D);
						
						cluster_control.delta_steps = cluster_control.deltas.size();
						cluster_control.scale_delta_range = false;
					}
					else
					{
						cluster_control.min_delta = get_next_number_no_limits(ERROR_clp_clu_D, 0.0);
						if (next_parameter_is_number() == true)
						{
							cluster_control.max_delta = get_next_number_no_limits(ERROR_clp_clu_D, 0.0);
							cluster_control.delta_steps = get_next_number(ERROR_clp_clu_D, 2);
							if (next_parameter_is_number() == true)
								cluster_control.scale_delta_range = get_next_bool(ERROR_clp_clu_D);
							
							if (cluster_control.scale_delta_range == false)
								if ((cluster_control.max_delta <= cluster_control.min_delta) or (cluster_control.max_delta > 1.0))
									Tcommand_line_parser::exit_with_help(ERROR_clp_clu_D);
						}
						else
						{
							cluster_control.max_delta = cluster_control.min_delta;
							cluster_control.delta_steps = 1;
						}
					}
					break;
				case 'e':
					cluster_control.eps_factor = get_next_number_no_limits(ERROR_clp_clu_e, 1.0);
					break;
				case 'E':
					cluster_control.density_estimator_type = get_next_enum(ERROR_clp_clu_E, 0, DENSITY_ALGORITHMS_MAX-1);
					break;
				case 'H':
					cluster_control.hierarchical_depth = get_next_number(ERROR_clp_clu_H, UNLIMITED_CLUSTER_DEPTH);
					if ((cluster_control.hierarchical_depth < 0) and (cluster_control.hierarchical_depth != UNLIMITED_CLUSTER_DEPTH))
						Tcommand_line_parser::exit_with_help(ERROR_clp_clu_H);
					break;
				case 'P':
					cluster_control.domain_partitioner_type = get_next_enum(ERROR_clp_clu_P, 0, DOMAIN_PARTITIONER_TYPES_MAX-1);
					break;
				case 's':
					cluster_control.cluster_compare_method = get_next_number(ERROR_clp_clu_s, 0, CLUSTER_COMPARE_MODES_MAX-1);
					cluster_control.rho_compare_tolerance_factor = get_next_number(ERROR_clp_clu_s, 0.0);
					break;
				case 't':
					cluster_control.tau_factor = get_next_number(ERROR_clp_clu_t, 1.0);
					set_tau = true;
					break;
				default:
					Tcommand_line_parser::exit_with_help(ERROR_clp_gen_unknown_option);
			}
		}
		
	cluster_control.random_seed = get_random_seed();
		
	train_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
	result_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
	
	if (current_position < parameter_list_size)
	{
		test_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
		test_result_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
	}	
		
	if (train_file_format.weight_position != 0)
		cluster_control.density_estimator_type = DENSITY_TRUTH;
	else if (cluster_control.density_estimator_type == DENSITY_TRUTH)
		flush_exit(ERROR_COMMAND_LINE, "Option -E %d requires a weighted data set,", DENSITY_TRUTH);
}


//**********************************************************************************************************************************


void Tcommand_line_parser_cluster::display_help(unsigned error_code)
{
	Tcluster_control cluster_ctrl;
	

	if (error_code == ERROR_clp_clu_D)
	{
		display_separator("-D <option_1> ...",
						"-D <delta>\n"
						"-D <min_delta> <max_delta> <steps> [<scale>]\n"
						"-D <delta_list>");
		flush_info(INFO_1, 
		"Determines the considered delta widths for the density estimator. In the\n"
		"first variant, only the value <delta> is used, while in the second variant\n"
		"a grid with <steps> many grid points ranging from <min_delta> to <max_delta>\n"
		"is considered. If <scale> == 0, the intermediate grid points are arithmetic\n"
		"points in between, while in the case <scale> == 1, min_delta and max_delta\n"
		"are first multiplied by some numbers depending on the dataset's size and\n"
		"dimension and geometric grid between the resulting numbers is taken.\n"
		);
		
		display_specifics();
		flush_info(INFO_1, 
		"<scale>       Flag indicating whether <min_gamma> and <max_gamma> are scaled\n"
		"              based on the sample size and the dimension.\n");
		
		display_ranges();
		flush_info(INFO_1, "<delta>:       float in between 0.0 and 1.0\n");
		flush_info(INFO_1, "<min_delta>:   float > 0 (and <= 1.0 if <scale> == 0)\n");
		flush_info(INFO_1, "<max_delta>:   float > 0 (or between <min_delta> and 1.0 if <scale> == 0)\n");
		flush_info(INFO_1, "<steps>:       integer greater than 1\n");
		flush_info(INFO_1, "<scale>:       bool\n");
		
		display_defaults();
		flush_info(INFO_1, "<delta>        = not the default\n");
		flush_info(INFO_1, "<min_delta>    = %1.3f\n", cluster_ctrl.min_delta);
		flush_info(INFO_1, "<max_delta>    = %1.3f\n", cluster_ctrl.max_delta);
		flush_info(INFO_1, "<steps>        = %d\n", cluster_ctrl.delta_steps);
		flush_info(INFO_1, "<scale>        = %d\n", cluster_ctrl.scale_delta_range);
	}

	
	if (error_code == ERROR_clp_clu_e)
	{
		display_separator("-e <eps_factor>");
		flush_info(INFO_1, 
		"Sets the accepted vertical (statistical) uncertainty parameter epsilon, to the\n"
		"value:\n"
		"<eps_factor> * sqrt(max_density * sample_size * delta^dim / unit_ball_volume)\n"
		"In general, a small <eps_factor> lead to a higher vertical resolution, but bear\n"
		"the risk of interpreting statistical fluctuations as clusters. In addition,\n"
		"small values increase the runtime, so that changing this parameter from its\n"
		"default value should be done with care. More informaton on the vertical\n"
		"(statistical) uncertainty can be found in the paper\n"
		"Steinwart, Fully adaptive density-based clustering, Ann. Statist. 43, 2132-2167,\n"
		"2015.\n");
		
		display_ranges();
		flush_info(INFO_1, "<eps_factor>:  float >= 1.0\n");
		display_defaults();
		flush_info(INFO_1, "<eps_factor>   = %1.3f\n", cluster_ctrl.eps_factor);
	}
	
	if (error_code == ERROR_clp_clu_E)
	{
		display_separator("-E <density_estimator>");
		flush_info(INFO_1, 
		"Sets the underlying density estimation method to the specified one. If the\n"
		"method %d is chosen, the input data set must be weighted. Conversely, such\n"	
		"an input file type automatically sets <density_estimator> to %d.\n", DENSITY_TRUTH, DENSITY_TRUTH);
		
		display_specifics();
		flush_info(INFO_1, "<density_estimator> = %d  =>  treat weights in input file as true density values\n", DENSITY_TRUTH);
		flush_info(INFO_1, "<density_estimator> = %d  =>  histogram rule\n", DENSITY_HISTOGRAMM);
		flush_info(INFO_1, "<density_estimator> = %d  =>  moving window rule with squared ball\n", DENSITY_CONSTANT_SUP_KERNEL);
		flush_info(INFO_1, "<density_estimator> = %d  =>  moving window rule with Euclidean ball\n", DENSITY_CONSTANT_EUCLIDEAN_KERNEL);
		flush_info(INFO_1, "<density_estimator> = %d  =>  kernel density with triangular kernel\n", DENSITY_TRIANGULAR_KERNEL);
		flush_info(INFO_1, "<density_estimator> = %d  =>  kernel density with Epanechnikov kernel\n", DENSITY_EPANECHNIKOV);
		flush_info(INFO_1, "<density_estimator> = %d  =>  kernel density with biweight kernel\n", DENSITY_BIWEIGHT_KERNEL);
		flush_info(INFO_1, "<density_estimator> = %d  =>  kernel density with triweight kernel\n", DENSITY_TRIWEIGHT_KERNEL);
		

		display_ranges();
		flush_info(INFO_1, "<density_estimator>:         integer between 0 and %d.\n", DENSITY_ALGORITHMS_MAX-1);
		display_defaults();
		flush_info(INFO_1, "<density_estimator>         = %d\n", cluster_ctrl.density_estimator_type);
	}

	
	if (error_code == ERROR_clp_clu_H)
	{
		display_separator("-H <cluster_depth>");
		flush_info(INFO_1, 
		"Selects the maximal depth of the hierarchical cluster tree.\n");

		display_specifics();
		flush_info(INFO_1, "<cluster_depth> = %d  =>  unlimited depth\n", UNLIMITED_CLUSTER_DEPTH);

				
		display_ranges();
		flush_info(INFO_1, "<cluster_depth>:   unsigned integer or %d\n", UNLIMITED_CLUSTER_DEPTH);
		display_defaults();
		flush_info(INFO_1, "<cluster_depth>   = %d\n", cluster_ctrl.hierarchical_depth);
	}
	
	
	if (error_code == ERROR_clp_clu_P)
	{
		display_separator("-P <domain_partitioner>");
		flush_info(INFO_1, 
		"Sets the underlying domain partitioner method, which is used to compute the\n"
		"necessary neighborhoods in an efficient way.\n");
		
		display_specifics();
		flush_info(INFO_1, "<domain_partitioner> = %d  =>  compute all distannces by brute-force\n", NO_PARTITIONER);
		flush_info(INFO_1, "<domain_partitioner> = %d  =>  use a cover tree approach\n", COVER_TREE);
		flush_info(INFO_1, "<domain_partitioner> = %d  =>  pick one of the above partition methods by\n"
		"                              some heuristic focussing on speed\n", ADAPTIVE_PARTITIONER);
		
		display_ranges();
		flush_info(INFO_1, "<domain_partitioner>:         integer between 0 and %d.\n", DOMAIN_PARTITIONER_TYPES_MAX-1);
		display_defaults();
		flush_info(INFO_1, "<domain_partitioner>         = %d\n", cluster_ctrl.domain_partitioner_type);
	}
	
	if (error_code == ERROR_clp_clu_s)
	{
		display_separator("-s <compare_method> <rho_tolerance_factor>");
		flush_info(INFO_1, 
		"Determines the method for choosing from different clusters at similar rho\n"
		"levels.\n");
		
		display_specifics();
		flush_info(INFO_1, "<compare_method> = %d    =>  vompare by excess mass approach\n", EXCESS_MASS);
		flush_info(INFO_1, "<compare_method> = %d    =>  compare by number direct childs\n", GREEDY_CLUSTER);
		flush_info(INFO_1, "<compare_method> = %d    =>  compare by number of peaks\n", GREEDY_PEAKS);
		flush_info(INFO_1, "<rho_tolerance_factor>      describes, which rho are similar.\n", GREEDY_CLUSTER);
		
		display_ranges();
		flush_info(INFO_1, "<compare_method>:           integer between 0 and %d.\n", CLUSTER_COMPARE_MODES_MAX-1);
		flush_info(INFO_1, "<rho_tolerance_factor>:     float >= 0.0\n");
		
		display_defaults();
		flush_info(INFO_1, "<compare_method>          = %d\n", cluster_ctrl.cluster_compare_method);
		flush_info(INFO_1, "<rho_tolerance_factor>    = %1.3f\n", cluster_ctrl.rho_compare_tolerance_factor);
	}	
	
	if (error_code == ERROR_clp_clu_t)
	{
		display_separator("-t <tau_factor>");
		flush_info(INFO_1, 
		"Determines that the cluster algorithm looks for tau connetected components,\n"
		"where the value of tau is set to be:\n"
		"<tau_factor> * delta\n"	
		"Larger values of <tau_factor> glue more separated regions together so that\n"
		"clusters may only be detected at higher density levels, which in turn leads\n"
		"to smaller clusters and more samples that are not assigned to any cluster.\n"
		"Conversely, a small <tau_factor> bears the risk of too aggressively separating\n"
		"regions, so that the value of <tau_factor> should be changed with care. However\n"
		"values between 2.0 and 3.0 are in general fine, while values smaller than 2.0\n"
		"may lead to unintended effects.\n");

		display_ranges();
		flush_info(INFO_1, "<tau_factor>:   float >= 1.0\n");
		display_defaults();
		flush_info(INFO_1, "<tau_factor>   = %1.3f\n", cluster_ctrl.tau_factor);
	}	

	
	Tcommand_line_parser::display_help(error_code);
}



//**********************************************************************************************************************************


void Tcommand_line_parser_cluster::exit_with_help()
{
	flush_info(INFO_SILENCE,
	"\n\ndbc [options] <train_file> <result_file> [<test_file> <test_result_file>]\n"
	"\nClusters the data in <train_file> according to the specified options and writes"
	"\nthe resulting clustered data into <result_file>. If the optional two test file"
	"\nnames are specified, too, then the clustering obtained from <train_file> is also"
	"\napplied to the data in <test_file> and written to <test_result_file>.\n"
	"\nAllowed extensions:\n"
		"<train_file>:         .csv and .lsv\n"
		"<train_result_file>:  .csv and .lsv\n"
		"<test_file>:          .csv and .lsv\n"
		"<test_result_file>:   .csv and .lsv\n"
	);
	display_help_file_formats();

	if (full_help == false)
		flush_info(INFO_SILENCE, "\nOptions:");
	display_help(ERROR_clp_gen_d);
	display_help(ERROR_clp_clu_D);
	display_help(ERROR_clp_clu_e);
	display_help(ERROR_clp_clu_E);
	display_help(ERROR_clp_gen_h);
	display_help(ERROR_clp_clu_H);
	display_help(ERROR_clp_clu_P);
	display_help(ERROR_clp_gen_r);
	display_help(ERROR_clp_clu_s);
	display_help(ERROR_clp_clu_t);
	display_help(ERROR_clp_gen_T);
	             
	flush_info(INFO_SILENCE,"\n\n");
	copyright();
	flush_exit(ERROR_SILENT, "");
};





#endif

