// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (CLUSTER_AFFILIATION_CLASS_CPP) 
 #define CLUSTER_AFFILIATION_CLASS_CPP


#include "sources/cluster/solver/cluster_affiliation.h" 




//**********************************************************************************************************************************


Tcluster_affiliation::Tcluster_affiliation()
{
	clear();
}

//**********************************************************************************************************************************


void Tcluster_affiliation::clear()
{
	cluster_affiliations.clear();
	clusters.clear();
	cluster_maxima.clear();
	current_cluster_numbers.clear();
	unassigned_samples.clear();
	samples_with_new_cluster_number.clear();
	
	min_cluster_number = 0.0;
	max_cluster_number = 0.0;
}



//**********************************************************************************************************************************


void Tcluster_affiliation::reserve(unsigned size)
{
	unsigned i;
	
	
	clear();
	cluster_affiliations.assign(size, 0.0);
	for (i=0; i<cluster_affiliations.size(); i++)
		assign_sample_to_cluster(i, 0.0, -1.0);
}


//**********************************************************************************************************************************

void Tcluster_affiliation::new_level_set()
{
	samples_with_new_cluster_number.clear();
}

//**********************************************************************************************************************************

void Tcluster_affiliation::assign_sample_to_cluster(unsigned index, double cluster_number, double rho)
{
	Tpeak peak;
	cluster_maxima_store_type::iterator it;
	
	
	if (index >= cluster_affiliations.size())
		flush_exit(ERROR_DATA_MISMATCH, "Cannot insert sample with number %d into a cluster affiliation object of size %d.", index, cluster_affiliations.size());


	if (cluster_number != 0.0)
	{
		peak.height = rho;
		peak.indices.insert(index);
	
		create_new_cluster(cluster_number);
		clusters[cluster_number].push_back(index);
		unassigned_samples.erase(index);
		
		it = cluster_maxima.find(cluster_number);
		if (it == cluster_maxima.end())
			cluster_maxima[cluster_number] = peak;
		else
		{
			if (rho > cluster_maxima[cluster_number].height)
				cluster_maxima[cluster_number] = peak;
			else if (rho == cluster_maxima[cluster_number].height)
				cluster_maxima[cluster_number].indices.insert(index);
		}
		
		if ((cluster_number > 0.0) and (cluster_affiliations[index] != cluster_number))
			samples_with_new_cluster_number.insert(index);
		
	}
	else
		unassigned_samples.insert(index);
	cluster_affiliations[index] = cluster_number;
}

//**********************************************************************************************************************************

void Tcluster_affiliation::merge_clusters(double cluster_number_1, double cluster_number_2, double& new_cluster_number)
{
	double min_cluster_number;
	double max_cluster_number;
	cluster_store_type::iterator jt_1;
	cluster_store_type::iterator jt_2;
	

	if ((cluster_number_1 == 0.0) or (cluster_number_2 == 0.0))
		flush_exit(ERROR_DATA_MISMATCH, "Cannot merge clusters with numbers %2.4f and %2.4f.", cluster_number_1, cluster_number_2);
	jt_1 = clusters.find(cluster_number_1);
	if (jt_1 == clusters.end())
		flush_exit(ERROR_DATA_MISMATCH, "Cannot merge clusters since there is no cluster with number %2.4f.", cluster_number_1);
	jt_2 = clusters.find(cluster_number_2);
	if (jt_2 == clusters.end())
		flush_exit(ERROR_DATA_MISMATCH, "Cannot merge clusters since there is no cluster with number %2.4f.", cluster_number_2);
	
	
	min_cluster_number = min(cluster_number_1, cluster_number_2);
	max_cluster_number = max(cluster_number_1, cluster_number_2);
	if (min_cluster_number > 0.0)
	{
		flush_info(INFO_DEBUG, "\nMerging cluster %d into cluster %d.", unsigned(max_cluster_number), unsigned(min_cluster_number));
		move_cluster_numbers(max_cluster_number, min_cluster_number);
		new_cluster_number = min_cluster_number;
	}
	else
	{
		flush_info(INFO_DEBUG, "\nMerging cluster %d into cluster %d.", unsigned(min_cluster_number), unsigned(max_cluster_number));
		move_cluster_numbers(min_cluster_number, max_cluster_number);
		new_cluster_number = max_cluster_number;
	}
}


//**********************************************************************************************************************************

void Tcluster_affiliation::change_cluster_number(double old_cluster_number, double new_cluster_number)
{
	list <unsigned> new_list;
	cluster_store_type::iterator jt_old;
	cluster_store_type::iterator jt_new;
	
	
	jt_old = clusters.find(old_cluster_number);
	if (jt_old == clusters.end())
		flush_exit(ERROR_DATA_MISMATCH, "Cannot change cluster number since there is no cluster with number %2.4f.", old_cluster_number);
	
	jt_new = clusters.find(new_cluster_number);
	if (jt_new != clusters.end())
		flush_exit(ERROR_DATA_MISMATCH, "Cannot change cluster number since there already exists a cluster with number %2.4f.", new_cluster_number);
	

	flush_info(INFO_DEBUG, "\nChanging cluster number %d to cluster number %d.", unsigned(old_cluster_number), unsigned(new_cluster_number));
	create_new_cluster(new_cluster_number);
	move_cluster_numbers(old_cluster_number, new_cluster_number);
}


//**********************************************************************************************************************************

double Tcluster_affiliation::get_largest_inactive_maximum()
{
	double largest_maximum;
	cluster_maxima_store_type::iterator it;
	
	
	largest_maximum = -1.0;
	for (it = cluster_maxima.begin(); it != cluster_maxima.end(); ++it)
		if (it->first < 0.0)
			largest_maximum = max(largest_maximum,  it->second.height); 

	return largest_maximum;
}


//**********************************************************************************************************************************

void Tcluster_affiliation::display()
{
	Tpeak peak;
	cluster_store_type::iterator jt;
	my_unordered_set <unsigned>::iterator it;
	
	for (jt = clusters.begin(); jt != clusters.end(); ++jt) 
		if (jt->first > 0.0)
		{
			peak = cluster_maxima[jt->first];
			flush_info(INFO_1, "\nCluster with number %1.0f contains %d samples and its maximum is %2.4f at: ", jt->first, jt->second.size(), peak.height);
			for (it = peak.indices.begin(); it != peak.indices.end(); ++it)
				flush_info(INFO_1, "%d ", *it);
		}
}


//**********************************************************************************************************************************

void Tcluster_affiliation::delete_cluster(double cluster_number)
{
	cluster_store_type::iterator jt;

	
	if (cluster_number != 0.0)
	{
		jt = clusters.find(cluster_number);
		jt->second.clear();
		clusters.erase(jt);
		current_cluster_numbers.erase(cluster_number);
		cluster_maxima.erase(cluster_number);
	}
}



//**********************************************************************************************************************************

void Tcluster_affiliation::create_new_cluster(double cluster_number)
{
	list <unsigned> new_list;
	cluster_store_type::iterator it;
	
	
	if (cluster_number != 0.0)
	{
		it = clusters.find(cluster_number);
		if (it == clusters.end())
		{
			clusters[cluster_number] = new_list;
			current_cluster_numbers.insert(cluster_number);
			min_cluster_number = min(min_cluster_number, cluster_number);
			max_cluster_number = max(max_cluster_number, cluster_number);
		}
	}
}





//**********************************************************************************************************************************

void Tcluster_affiliation::move_cluster_numbers(double source_cluster_number, double target_cluster_number)
{
	Tpeak source_max;
	Tpeak target_max;
	list <unsigned>::iterator it;
	list <unsigned>::iterator it_start;
	list <unsigned>::iterator it_stop;
	cluster_store_type::iterator jt;
	

	if (source_cluster_number == target_cluster_number)
		return;
	
	it_start = clusters[source_cluster_number].begin();
	it_stop = clusters[source_cluster_number].end();
	
	if (target_cluster_number != 0.0)
	{
		for (it = it_start; it != it_stop; ++it)
			cluster_affiliations[*it] = target_cluster_number;
		
		if (target_cluster_number > 0.0)
			for (it = it_start; it != it_stop; ++it)
				samples_with_new_cluster_number.insert(*it);
	
		jt = clusters.find(target_cluster_number);
		jt->second.splice(jt->second.end(), clusters[source_cluster_number]);
		
		if (source_cluster_number != 0.0) 
		{
			source_max = cluster_maxima[source_cluster_number];
			target_max = cluster_maxima[target_cluster_number];
			
			if (source_max.height > target_max.height)
				cluster_maxima[target_cluster_number] = source_max;
			else if (source_max.height < target_max.height)
				cluster_maxima[target_cluster_number] = target_max;
			else if (source_max.height == target_max.height)
			{
				target_max.indices.insert(source_max.indices.begin(), source_max.indices.end());
				cluster_maxima[target_cluster_number] = target_max;
			}
		}
	}
	else
		for (it = it_start; it != it_stop; ++it)
		{
			cluster_affiliations[*it] = 0.0;
			unassigned_samples.insert(*it);
		}
	
	delete_cluster(source_cluster_number);
}







//**********************************************************************************************************************************



#endif
