// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (CLUSTER_COMPARE_CPP) 
 #define CLUSTER_COMPARE_CPP


#include "sources/cluster/solver/cluster_compare.h" 




#include "sources/shared/basic_functions/extra_string_functions.h"


const double NOT_ASSIGNED = -1.0;

//**********************************************************************************************************************************

Tcluster_compare::Tcluster_compare()
{
	clear();
}


//**********************************************************************************************************************************

void Tcluster_compare::clear()
{
	rho_star = std::numeric_limits<double>::max();
	rho_step = 0.0;
	excess_mass = NOT_ASSIGNED;
	density_integral = NOT_ASSIGNED;
	constant_integral = NOT_ASSIGNED;
	
	rho_tolerance_factor = 1.0;
	
	number_of_clusters = 0;
	number_of_peaks = 0;
	core_samples = 0;
	all_samples = 0;
	
	working_set_size = 0;
	dataset_size = 0;
	
	delta = NOT_ASSIGNED;
	
	merge = NULL;
	cluster_compare_method = GREEDY_CLUSTER;
}

//**********************************************************************************************************************************

void Tcluster_compare::display(unsigned info_level)
{
	string output;

		
	output = displaystring_detailed();
	output = output + "\n";
	flush_info(info_level, output.c_str());
}


//**********************************************************************************************************************************

bool Tcluster_compare::operator > (const Tcluster_compare& cluster_compare) const
{
	bool i_am_the_better_cluster;
	
	
	
	if (rho_stars_within_tolerance(cluster_compare) == false)
		return (rho_star < cluster_compare.rho_star);
	
	
	i_am_the_better_cluster = false;
	switch (cluster_compare.cluster_compare_method) 
	{
		case EXCESS_MASS:  
			i_am_the_better_cluster = (excess_mass > cluster_compare.excess_mass);
			break;  
		case GREEDY_CLUSTER:  
			i_am_the_better_cluster = (number_of_clusters > cluster_compare.number_of_clusters);
			i_am_the_better_cluster = i_am_the_better_cluster or ((number_of_clusters == cluster_compare.number_of_clusters) and (core_samples > cluster_compare.core_samples));
			break;  
		case GREEDY_PEAKS:  
			i_am_the_better_cluster = (number_of_peaks > cluster_compare.number_of_peaks);
			i_am_the_better_cluster = i_am_the_better_cluster or ((number_of_peaks == cluster_compare.number_of_peaks) and (core_samples > cluster_compare.core_samples));
			break;  
		default:  
			flush_exit(ERROR_DATA_FALLS_OUTSIDE_SAFE_PARAMETERS, "Cluster compare method %d does not exist.", cluster_compare.cluster_compare_method);
	}
	
	return i_am_the_better_cluster;
}


//**********************************************************************************************************************************

bool Tcluster_compare::rho_stars_within_tolerance(const Tcluster_compare& cluster_compare) const
{
	double rho_diff;
	

	rho_diff = abs(rho_star - cluster_compare.rho_star);

	return (rho_diff <= rho_tolerance_factor * min(rho_step, cluster_compare.rho_step));
}



//**********************************************************************************************************************************

string Tcluster_compare::displaystring_detailed() const
{
	string output;


	output = output + "del: " + number_to_string(delta, 4, DISPLAY_FLOAT);
	if (rho_star >= 0.0)
		output = output + "  rho*: " + number_to_string(rho_star, 4, DISPLAY_FLOAT);
	else
		output = output + "  rho*:    ---";
	output = output + "  rho+: " + number_to_string(rho_step, 4, DISPLAY_FLOAT);
	
	if (excess_mass >= 0.0)
		output = output + "  exma: " + number_to_string(excess_mass, 4, DISPLAY_FLOAT);
	else
		output = output + "  exma:    ---";	
	if (density_integral >= 0.0)
		output = output + "  prob: " + number_to_string(density_integral, 4, DISPLAY_FLOAT);
	else
		output = output + "  prob:    ---";
	if (constant_integral >= 0.0)
		output = output + "  vol: " + number_to_string(constant_integral, 4, DISPLAY_FLOAT);
	else
		output = output + "  vol:    ---";
	output = output + "  clu: " + pos_number_to_string(number_of_clusters, 2);
	output = output + "  pea: " + pos_number_to_string(number_of_peaks, 2);
	output = output + "  cor: " + pos_number_to_string(core_samples, 5);
	output = output + "  all: " + pos_number_to_string(all_samples, 5);

	return output;
}


#endif




