// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (CLUSTER_CONTROL_CLASS_CPP) 
	#define CLUSTER_CONTROL_CLASS_CPP


#include "sources/cluster/solver/cluster_control.h"



#include "sources/shared/basic_functions/math_misc.h"

#include "sources/shared/basic_types/sample.h"
#include "sources/shared/basic_types/vector.h"

#include "sources/cluster/solver/cluster_compare.h" 

//**********************************************************************************************************************************


Tdensity_control::Tdensity_control()
{
	density_estimator_type = DENSITY_CONSTANT_EUCLIDEAN_KERNEL;
	domain_partitioner_type = ADAPTIVE_PARTITIONER;
	
	min_delta = 0.02;
	max_delta = 0.1;
	delta_steps = 50;
	scale_delta_range = false;
	max_delta_neighborhood = 0.0;
}



//**********************************************************************************************************************************

unsigned Tdensity_control::get_density_estimator_category() const
{
	if ((density_estimator_type == DENSITY_TRUTH) or (density_estimator_type == DENSITY_HISTOGRAMM))
		return density_estimator_type;
	else
		return DENSITY_GENERIC_KERNEL;
}


//**********************************************************************************************************************************

unsigned Tdensity_control::get_metric_of_estimator() const
{
	if ((density_estimator_type == DENSITY_TRUTH) or (density_estimator_type == DENSITY_HISTOGRAMM) or (density_estimator_type == DENSITY_CONSTANT_SUP_KERNEL))
		return SUP_NORM;
	else
		return EUCLIDEAN;
}

//**********************************************************************************************************************************

double Tdensity_control::get_estimator_integral(unsigned dim) const
{
	double factor;
	
	if (density_estimator_type == DENSITY_TRUTH)
		return 1.0;
	

	switch(density_estimator_type)
	{
		case DENSITY_HISTOGRAMM:
			factor = 1.0;
			break;

		case DENSITY_CONSTANT_SUP_KERNEL:
			factor = 1.0;
			break;
			
		case DENSITY_CONSTANT_EUCLIDEAN_KERNEL:
			factor = 1.0;
			break;
			
		case DENSITY_TRIANGULAR_KERNEL:
			factor = 1.0 / (double(dim) + 1.0);
			break;
			
		case DENSITY_EPANECHNIKOV:
			factor = 2.0 / (double(dim) + 2.0);
			break;
			
		case DENSITY_BIWEIGHT_KERNEL:
			factor = 8.0 / ((double(dim) + 2.0) * (double(dim) + 4.0));
			break;
			
		case DENSITY_TRIWEIGHT_KERNEL:
			factor = 48.0 / ((double(dim) + 2.0) * (double(dim) + 4.0) * (double(dim) + 6.0));;
			break;
			
		default:
			flush_exit(ERROR_DATA_FALLS_OUTSIDE_SAFE_PARAMETERS, "Density estimator of type %d does not exist.", density_estimator_type);
			return 0.0;
			break;
	}
	
	return factor * unit_ball_volume(dim);
}



//**********************************************************************************************************************************

double Tdensity_control::kernel_function(double distance) const
{
	double value;
	double tmp_value;
	
	if ((density_estimator_type == DENSITY_TRUTH) or (density_estimator_type == DENSITY_HISTOGRAMM))
	{
		flush_exit(ERROR_DATA_FALLS_OUTSIDE_SAFE_PARAMETERS, "Density estimator of type %d does not have a kernel.", density_estimator_type);
		return 1.0;
	}
	
// To avoid numerical issues, the constant kernels need to assume that the value of distance
// is not greater than 1.0 (remember, when calling this function, we usually scale the true
// distance by 1.0 / delta, so that we may end up being slightly above 1.0 although the 
// original distance was not greater than delta. The other kernels are less sensitive to this
// effect since they are continuous.

	switch(density_estimator_type)
	{
		case DENSITY_CONSTANT_SUP_KERNEL:
			value = 1.0;
			break;
			
		case DENSITY_CONSTANT_EUCLIDEAN_KERNEL:
			value = 1.0;
			break;
			
		case DENSITY_TRIANGULAR_KERNEL:
			value = max(0.0, 1.0 - distance);
			break;
			
		case DENSITY_EPANECHNIKOV:
			value = max(0.0, 1.0 - distance * distance);
			break;
			
		case DENSITY_BIWEIGHT_KERNEL:
			tmp_value = max(0.0, 1.0 - distance * distance);
			value = tmp_value * tmp_value;
			break;
			
		case DENSITY_TRIWEIGHT_KERNEL:
			tmp_value = max(0.0, 1.0 - distance * distance);
			value = tmp_value * tmp_value * tmp_value;
			break;
			
		default:
			flush_exit(ERROR_DATA_FALLS_OUTSIDE_SAFE_PARAMETERS, "Density estimator of type %d does not exist.");
			return 0.0;
	}
	
	return value;
}



//**********************************************************************************************************************************

double Tdensity_control::unit_ball_volume(unsigned dim) const
{
	unsigned i;
	double twopi;
	double volume;

	
	if (get_metric_of_estimator() == EUCLIDEAN)
	{
		twopi = 2.0 * 3.14159265359;
		volume = 1.0;
		if (dim % 2 == 0)
			for (i=0; i<dim; i=i+2)
				volume = volume * twopi / double(i+2);
		else
		{
			for (i=1; i<dim; i=i+2)
				volume = volume * twopi / double(i+2);
			volume = 2.0 * volume;
		}
	}
	else
		volume = pow(2.0, double(dim));
	
	return volume;
}



//**********************************************************************************************************************************


void Tdensity_control::scale_deltas(unsigned data_size, unsigned dim)
{
	unsigned i;
	double delta;
	double delta_max_scaled;
	double delta_min_scaled;
	
	if (deltas.size() > 0)
		deltas_scaled = deltas;
	else
	{
		deltas_scaled.clear();
		deltas_scaled.resize(delta_steps);
		
		if (delta_steps == 1)
			deltas_scaled[0] = min_delta;
		else
		{
			if (scale_delta_range == false)
			{
				delta = max_delta;
				for (i=0; i<delta_steps; i++)
				{
					deltas_scaled[i] = delta;
					delta = delta - (max_delta - min_delta)/(double(delta_steps) - 1.0);
				}
			}
			else
			{
				delta_max_scaled = max_delta * pow(log(double(data_size)), -1.0/double(dim));
				delta_min_scaled = min_delta * pow(log(double(data_size))/double(data_size), 1.0/double(dim));
			
				for (i=0; i<delta_steps; i++)
					deltas_scaled[i] = delta_min_scaled * pow((delta_max_scaled/delta_min_scaled), double(delta_steps - i - 1)/(double(delta_steps) - 1.0));
			}
			sort_down(deltas_scaled);
		}
	}
}

//**********************************************************************************************************************************


Tcluster_control::Tcluster_control()
{
	tau_factor = 2.0;
	eps_factor = 3.0;

	hierarchical_depth = UNLIMITED_CLUSTER_DEPTH;
	
	cluster_compare_method = GREEDY_PEAKS;
	rho_compare_tolerance_factor = 1.0;
	
	test_set_size = 5000;
	random_seed = 1;
}


//**********************************************************************************************************************************


double Tcluster_control::get_tau_from_delta(double delta) const
{
	double tau;
	
	switch(get_density_estimator_category())
	{
		case DENSITY_HISTOGRAMM:
			tau = (tau_factor + 0.000001) * delta;
// 			tau = (cluster_control.tau + 2.0) * delta;
			break;
			
		case DENSITY_GENERIC_KERNEL:
			tau = tau_factor * delta;
			break;
			
		case DENSITY_TRUTH:
			tau = delta;
			break;

		default:
			tau = 0.0;
			break;
	}

	return tau;
}







#endif
