// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (CLUSTER_TREE_CLASS_CPP) 
 #define CLUSTER_TREE_CLASS_CPP

#include "sources/cluster/solver/cluster_tree.h" 




//**********************************************************************************************************************************


bool Tpeak::operator == (const Tpeak& peak) const
{
	return ((height == peak.height) and (cluster_number == peak.cluster_number));
}


//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************






Tcluster::Tcluster(const Tcluster& cluster)
{
	copy(cluster);
}


//**********************************************************************************************************************************


Tcluster::Tcluster()
{
	clear();
}


//**********************************************************************************************************************************


void Tcluster::clear()
{
	rho = -1.0;
	rho_star = -1.0;
	cluster_number = 0.0;
	parent_cluster_number = 0.0;
	width = 0.0;
	size = 0;
	max_number_of_core_samples = 0;
	
	peaks.clear();
	core_samples.clear();
}


//**********************************************************************************************************************************


void Tcluster::copy(const Tcluster& source_cluster)
{
	rho = source_cluster.rho;
	rho_star = source_cluster.rho_star;
	cluster_number = source_cluster.cluster_number;
	parent_cluster_number = source_cluster.parent_cluster_number;

	width = source_cluster.width;
	size = source_cluster.size;
	max_number_of_core_samples = source_cluster.max_number_of_core_samples;
	
	peaks = source_cluster.peaks;
	core_samples = source_cluster.core_samples;
}


//**********************************************************************************************************************************





Tcluster& Tcluster::operator = (const Tcluster& cluster)
{
	copy(cluster);
	return *this;
}

//**********************************************************************************************************************************

void Tcluster::merge(const Tcluster& cluster)
{
	flush_info(INFO_PEDANTIC_DEBUG, "\nMerging cluster %2.4f with %d core samples into cluster %2.4f with %d core samples.", cluster.cluster_number, cluster.core_samples.size(), cluster_number, core_samples.size()); 
	
	if ((width != cluster.width) or (cluster.width == 0.0))
		flush_exit(ERROR_DATA_MISMATCH, "Cannot merge cluster %2.4f with width %2.4f into cluster %2.4f with width %2.4f.", cluster.cluster_number, cluster.width, cluster_number, width);
	
	if (cluster_number == cluster.cluster_number)
		flush_exit(ERROR_DATA_MISMATCH, "Cannot merge cluster %2.4f into cluster %2.4f.", cluster.cluster_number, cluster_number);
	
	
	peaks.insert(peaks.end(), cluster.peaks.begin(), cluster.peaks.end());
	cluster_number = min(cluster_number, cluster.cluster_number);
	
	if (cluster.cluster_number < cluster_number)
		core_samples = cluster.core_samples;
}

//**********************************************************************************************************************************

void Tcluster::merge(const Tcluster& cluster_1, const Tcluster& cluster_2)
{
	flush_info(INFO_PEDANTIC_DEBUG, "\nMerging clusters %2.4f and %2.4f with %d and %d core samples.", cluster_1.cluster_number, cluster_2.cluster_number,  cluster_1.core_samples.size(), cluster_2.core_samples.size()); 
	
	if ((cluster_1.width != cluster_2.width) or (cluster_1.width == 0.0))
		flush_exit(ERROR_DATA_MISMATCH, "Cannot merge clusters %2.4f and %2.4f with widths %2.4f and %2.4f.", cluster_1.cluster_number, cluster_2.cluster_number, cluster_1.width, cluster_2.width);
	
	if (cluster_1.cluster_number == cluster_2.cluster_number)
		flush_exit(ERROR_DATA_MISMATCH, "Cannot merge clusters %2.4f and %2.4f.", cluster_1.cluster_number, cluster_2.cluster_number);

	
	peaks = cluster_1.peaks;
	peaks.insert(peaks.end(), cluster_2.peaks.begin(), cluster_2.peaks.end());
	cluster_number = min(cluster_1.cluster_number, cluster_2.cluster_number);

	if (cluster_1.cluster_number < cluster_2.cluster_number)
		core_samples = cluster_1.core_samples;
	else
		core_samples = cluster_2.core_samples;
	
	width = cluster_1.width;
	max_number_of_core_samples = cluster_1.max_number_of_core_samples;
}


//**********************************************************************************************************************************


void Tcluster::display(unsigned info_level)
{
	flush_info(info_level, "\nCluster %2.4f with parent %2.4f.", cluster_number, parent_cluster_number);
	flush_info(info_level, "\nNumber of all peaks is %d.", peaks.size());
	if (rho >= 0.0)
	{
		flush_info(info_level, "\nMerge happened at    %2.4f.", rho);
		flush_info(info_level, "\nLevel above cluster is %2.4f.", rho_star);
	}
	flush_info(info_level, "\nUsed width is %2.4f.", width);
	flush_info(info_level, "\nCluster size is %d.", size);
	flush_info(info_level, "\nCluster core size is %d", core_samples.size());
	flush_info(info_level, "\nMax number of core samples is %d.", max_number_of_core_samples);
}



#endif



