// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (DENSITY_CLUSTER_MANAGER_CPP) 
 #define DENSITY_CLUSTER_MANAGER_CPP


#include "sources/cluster/solver/density_cluster_manager.h" 
 
 
 

#include "sources/shared/system_support/timing.h"

#include "sources/shared/basic_functions/math_misc.h"
#include "sources/shared/basic_functions/random_subsets.h" 
#include "sources/shared/basic_functions/random_generator.h"
 
#include "sources/shared/basic_types/vector.h"
#include "sources/shared/basic_types/my_unordered_map_and_set.h"



//**********************************************************************************************************************************



const double NOT_IN_SUBCLUSTER = -1.0;
const double IN_SUBCLUSTER = std::numeric_limits<double>::max();

enum THREAD_TASKS_CLUSTER {NEIGHBORHOOD_BUILD, NEIGHBORHOOD_SORT, CLUSTER_TREE_COMPUTE, EXCESS_MASS_COMPUTE, TEST_SET_ASSIGN};
 



//**********************************************************************************************************************************



Tdensity_cluster_manager::Tdensity_cluster_manager()
{
	clear();
}

//**********************************************************************************************************************************


Tdensity_cluster_manager::~Tdensity_cluster_manager()
{
	clear();
}


//**********************************************************************************************************************************


void Tdensity_cluster_manager::clear()
{
	clear_threads();
	
	dataset.clear();
	neighborhood_lists.neighbors.clear();
	neighborhood_lists.distances.clear();
}



//**********************************************************************************************************************************


void Tdensity_cluster_manager::load(const Tdataset& dataset, const Tcluster_control& cluster_control, Tcluster_info& cluster_info, Tparallel_control parallel_ctrl)
{
	unsigned i;
	double min_distance;
	Tdataset random_set;
	Tsample random_sample;
	vector <double> random_vector;

	
	cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
	
	clear();
	reserve_threads(parallel_ctrl);

	Tdensity_cluster_manager::dataset = dataset;
	Tdensity_cluster_manager::dataset.enforce_ownership();
	Tdensity_cluster_manager::cluster_control = cluster_control;
	Tdensity_cluster_manager::cluster_control.scale_deltas(dataset.size(), dataset.dim());
	
	excess_mass_of_thread.resize(get_team_size());
	cluster_info_of_thread.resize(get_team_size());
	
	switch(cluster_control.get_density_estimator_category())
	{
		case DENSITY_HISTOGRAMM:
			cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
			break;

		case DENSITY_GENERIC_KERNEL:
			domain_partitioner.load(dataset);
			cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
			neighborhood_build(cluster_control.get_tau_from_delta(Tdensity_cluster_manager::cluster_control.deltas_scaled[0]), cluster_info);
			break;

		case DENSITY_TRUTH:
			domain_partitioner.load(dataset);
			Tdensity_cluster_manager::cluster_control.eps_factor = 0.0;

			min_distance = std::numeric_limits<double>::max();
			for (i=1; i<dataset.size(); i++)
				min_distance = min(min_distance, sup_distance(dataset.sample(0), dataset.sample(i)));

			min_distance = 1.00001 * min_distance;
			Tdensity_cluster_manager::cluster_control.min_delta = min_distance;
			Tdensity_cluster_manager::cluster_control.max_delta = min_distance;
			Tdensity_cluster_manager::cluster_control.delta_steps = 1;
			Tdensity_cluster_manager::cluster_control.scale_deltas(dataset.size(), dataset.dim());
			
			cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
			neighborhood_build(cluster_control.get_tau_from_delta(min_distance), cluster_info);
			break;
	}
	
	
	if (cluster_control.get_density_estimator_category() != DENSITY_TRUTH)
	{
		cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
		flush_info(INFO_1, "\nGenerating %d random samples ...                     ", Tdensity_cluster_manager::cluster_control.test_set_size);
		
		random_set.enforce_ownership();
		init_random_generator(Tdensity_cluster_manager::cluster_control.random_seed);
		
		for (i=0; i<Tdensity_cluster_manager::cluster_control.test_set_size; i++)
		{
			get_uniform_ball_vector(random_vector, dataset.dim(), 1.0, Tdensity_cluster_manager::cluster_control.get_metric_of_estimator());
			random_sample = Tsample(random_vector);
			random_set.push_back(random_sample);
		}
		
		flush_info(INFO_1, "done!");
		cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
	}

	for (i=0; i<Tdensity_cluster_manager::cluster_control.delta_steps; i++)
		density_cluster_tree_estimator[Tdensity_cluster_manager::cluster_control.deltas_scaled[i]].load(Tdensity_cluster_manager::dataset, random_set, Tdensity_cluster_manager::cluster_control, cluster_info);
}


//**********************************************************************************************************************************


void Tdensity_cluster_manager::thread_entry()
{
	switch(thread_task)
	{
		case NEIGHBORHOOD_BUILD:
			neighborhood_build_threaded();
			break;
		
		case NEIGHBORHOOD_SORT:
			neighborhood_sort_threaded();
			break;
			
		case CLUSTER_TREE_COMPUTE:
			compute_all_cluster_trees_threaded();
			break;
			
		case EXCESS_MASS_COMPUTE:
			compute_excess_mass_etc_threaded();
			break;
			
		case TEST_SET_ASSIGN:
			assign_cluster_to_test_set_threaded();
			break;
	}
}



//**********************************************************************************************************************************


void Tdensity_cluster_manager::neighborhood_build(double max_distance, Tcluster_info& cluster_info)
{
	const unsigned cover_tree_threshold = 1024;
	Tworking_set_control working_set_control;

	
	cluster_info.tau_neighbor_build_time = get_wall_time_difference(cluster_info.tau_neighbor_build_time);
	
	if (cluster_control.domain_partitioner_type == ADAPTIVE_PARTITIONER)
	{
		use_cover_tree = (dataset.size() > cover_tree_threshold); 
		flush_info(INFO_1, "\nBuilding neighborhood information by tree approach ... ");
	}
	else
	{
		use_cover_tree = (cluster_control.domain_partitioner_type == COVER_TREE);
		flush_info(INFO_1, "\nBuilding neighborhood information by brute force ...");
	}

	
	neighborhood_lists.neighbors.clear();
	neighborhood_lists.distances.clear();
	neighborhood_lists.neighbors.resize(dataset.size());
	neighborhood_lists.distances.resize(dataset.size());
	neighborhood_tau = max_distance;
	
	if (use_cover_tree == true)
	{
		working_set_control.random_seed = 1;
		working_set_control.size_of_cells = 50;
		working_set_control.partition_method = VORONOI_TREE_BY_SIZE;
		domain_partitioner.build_cover_tree(working_set_control, cluster_control.get_metric_of_estimator());
		
		thread_task = NEIGHBORHOOD_BUILD;
		start_threads();
	}
	else
		domain_partitioner.build_neighborhood_lists(0, dataset.size(), max_distance, cluster_control.get_metric_of_estimator(), neighborhood_lists);

	flush_info(INFO_1, "done!");
	cluster_info.tau_neighbor_build_time = get_wall_time_difference(cluster_info.tau_neighbor_build_time);


	neighborhood_sort(cluster_info);
}


//**********************************************************************************************************************************


void Tdensity_cluster_manager::neighborhood_build_threaded()
{
	Tthread_chunk thread_chunk;

	if (use_cover_tree == false)
		flush_exit(ERROR_RUNTIME, "Threaded neighborhood building requires a cover_tree"); 
	
	thread_chunk = get_thread_chunk(dataset.size());
	
	domain_partitioner.build_neighborhood_lists(thread_chunk.start_index, thread_chunk.stop_index, neighborhood_tau, cluster_control.get_metric_of_estimator(), neighborhood_lists);
}



//**********************************************************************************************************************************


void Tdensity_cluster_manager::neighborhood_sort(Tcluster_info& cluster_info)
{
	cluster_info.tau_neighbor_sort_time = get_wall_time_difference(cluster_info.tau_neighbor_sort_time);
	flush_info(INFO_1, "\nSorting neighbors ...                                  ");
	
	thread_task = NEIGHBORHOOD_SORT;
	start_threads();
	flush_info(INFO_1, "done!");
	cluster_info.tau_neighbor_sort_time = get_wall_time_difference(cluster_info.tau_neighbor_sort_time);
}


//**********************************************************************************************************************************


void Tdensity_cluster_manager::neighborhood_sort_threaded()
{
	unsigned i;
	unsigned j;
	Tthread_chunk thread_chunk;
	
	
	thread_chunk = get_thread_chunk(dataset.size());
	
	
	for (i=thread_chunk.start_index; i<thread_chunk.stop_index; i++)
		if (neighborhood_lists.neighbors[i].size() > 0)
		{
			sort_up(neighborhood_lists.distances[i], neighborhood_lists.neighbors[i]);

			if (use_cover_tree == true)
			{
				j = 0;
				while ((j+1 < neighborhood_lists.neighbors[i].size()) and (neighborhood_lists.neighbors[i][j] != i))
					j++;

				neighborhood_lists.neighbors[i].erase(neighborhood_lists.neighbors[i].begin() + j);
				neighborhood_lists.distances[i].erase(neighborhood_lists.distances[i].begin() + j);
			}
		}
}



//**********************************************************************************************************************************


void Tdensity_cluster_manager::compute_all_cluster_trees(Tcluster_compare& best_cluster_compare, Tcluster_info& cluster_info)
{
	unsigned i;
	vector <unsigned> new_task_numbers;
	unsigned index_of_best_cluster;

	
// First we compute all cluster trees
	
	cluster_infos.clear();
	cluster_compare_list.clear();
	cluster_compare_list.resize(cluster_control.delta_steps);
	
	
// First we need to set-up the task scheduler. We follow a greedy approach
// by assuming that larger delta values require more computation due to the
// larger neighborhoods. Consequently, we assume that these large values 
// are at the front fo cluster_control.deltas_scaled, so that they are
// processed first. A few experiments suggested that this may save up to 
// 20-25% of the overall run time
	
	new_task_numbers = id_permutation(cluster_control.delta_steps);
	set_task_numbers(new_task_numbers);

	
	flush_info(INFO_1, "\nComputing %d cluster trees by %d threads.\n", cluster_control.delta_steps, get_team_size());
	thread_task = CLUSTER_TREE_COMPUTE;
	start_threads();

	
// Once we are done with this, we look for the best first split.

	cluster_compare_candidates = cluster_compare_list;
	identify_best_cluster(index_of_best_cluster, cluster_info);
	best_cluster_compare = cluster_compare_list[index_of_best_cluster];
	
	if (best_cluster_compare.rho_star == std::numeric_limits<double>::max())
		best_cluster_compare.rho_star = NO_RHO;
	

// Finally, collect the timings of the different threads;
	
	for (i=0; i<get_team_size(); i++)
		cluster_info = max(cluster_info, cluster_info_of_thread[i]);
}






//**********************************************************************************************************************************

void Tdensity_cluster_manager::identify_best_cluster(unsigned& index_of_best_cluster, Tcluster_info& cluster_info)
{
	unsigned i;
	unsigned ii;
	double max_rho_star;
	vector <double> core_cluster_affiliation;
	vector <double> cluster_affiliation_empty;
	vector <double> cluster_affiliation_tmp;
	Tcluster_compare best_cluster_compare;
	

// We first find the smallest rho_star within the candidates
	
	for (i=0; i<cluster_compare_candidates.size(); i++)
		if (cluster_compare_candidates[i].number_of_peaks > 1)
		{
			if (cluster_compare_candidates[i].rho_star <= best_cluster_compare.rho_star)
			{
				best_cluster_compare.rho_star = cluster_compare_candidates[i].rho_star;
				best_cluster_compare.rho_step = cluster_compare_candidates[i].rho_step;
				best_cluster_compare.number_of_peaks = cluster_compare_candidates[i].number_of_peaks;
			}
		}
		else
			if (cluster_compare_candidates[i].number_of_peaks == 1)
				best_cluster_compare.number_of_peaks = max(unsigned(1), best_cluster_compare.number_of_peaks);


// Next, we collect all rho_stars that (essentially) achieve the
// smallest rho_star.
	
	candidate_runs.clear();
	candidate_runs.reserve(cluster_control.delta_steps);
	max_rho_star = best_cluster_compare.rho_star + best_cluster_compare.rho_step;
	if (best_cluster_compare.number_of_peaks > 1)
	{
		for (i=0; i<cluster_compare_candidates.size(); i++)
			if ((cluster_compare_candidates[i].number_of_peaks > 1) 
				and best_cluster_compare.rho_stars_within_tolerance(cluster_compare_candidates[i]))
			{
				candidate_runs.push_back(i);
				max_rho_star = min(max_rho_star, best_cluster_compare.rho_star + cluster_compare_candidates[i].rho_step);
			}
	}
	else if (best_cluster_compare.number_of_peaks == 1)
		for (i=0; i<cluster_compare_candidates.size(); i++)
			if (cluster_compare_candidates[i].number_of_peaks == 1)
				candidate_runs.push_back(i);


// Depending on the cluster compare method, we next collect additional information
// about the clusters of the collected candidates.


	flush_info(INFO_1, "\n\nConsidering %d candidates out of %d initial candidates.\n", candidate_runs.size(), cluster_compare_candidates.size());
	
	if (cluster_control.cluster_compare_method == EXCESS_MASS)
		compute_excess_mass_etc(cluster_info);
	
	cluster_affiliation_empty.assign(dataset.size(), 0.0);
	for (i=0; i<candidate_runs.size(); i++)
	{
		ii = candidate_runs[i];
// 		delta = cluster_compare_candidates[ii].delta;

		if (cluster_control.cluster_compare_method == EXCESS_MASS)
			cluster_compare_candidates[ii].excess_mass = cluster_compare_candidates[ii].density_integral - max_rho_star * cluster_compare_candidates[ii].constant_integral;
		else
			get_level_set(ii);
		
// // Next we compute some further information.
// 		
// 		core_cluster_affiliation.assign(cluster_compare_candidates[ii].working_set_size, 0.0);
// 		get_core_sample_affiliation_from_forest(&cluster_tree[delta], core_cluster_affiliation);
// 
// 		cluster_affiliation_tmp = density_cluster_tree_estimator[delta].expand_from_core_to_train_samples(delta, 0.0, -1.0, cluster_info, core_cluster_affiliation, cluster_affiliation_empty);
// 
// 		cluster_compare_candidates[ii].not_assigned_samples = count(cluster_affiliation_tmp.begin(), cluster_affiliation_tmp.end(), 0.0);
	}

	
// Now that we have all information for the final comparison of the 
// candidates, we identify the best candidate.
		
	best_cluster_compare.clear();
	best_cluster_compare.cluster_compare_method = cluster_control.cluster_compare_method;

	index_of_best_cluster = candidate_runs[0];
	best_cluster_compare = cluster_compare_candidates[candidate_runs[0]];
	best_cluster_compare.cluster_compare_method = cluster_control.cluster_compare_method;
	best_cluster_compare.display(INFO_1);
	
	for (i=1; i<candidate_runs.size(); i++)
	{
		ii = candidate_runs[i];
		cluster_compare_candidates[ii].display(INFO_1);
		if (cluster_compare_candidates[ii] > best_cluster_compare)
		{
			index_of_best_cluster = ii;
			best_cluster_compare = cluster_compare_candidates[ii];
			best_cluster_compare.cluster_compare_method = cluster_control.cluster_compare_method;
		}
	}
	
	flush_info(INFO_1, "\nNew best rho is = %2.4f with %d clusters for delta = %2.4f.", best_cluster_compare.rho_star, best_cluster_compare.number_of_clusters, best_cluster_compare.delta);
}




//**********************************************************************************************************************************


void Tdensity_cluster_manager::compute_all_cluster_trees_threaded()
{
	unsigned i;
	int ii;
	double delta;
	double rho_star;
	unsigned thread_id;
	unsigned team_size;
	Tcluster_info cluster_info;
	list <Tcluster_info> current_cluster_infos;
	Tcluster_tree empty_tree;
	Tcluster_tree current_cluster_tree;
	std::list <Tcluster_info>::iterator it;
	
	
	thread_id = get_thread_id();
	team_size = get_team_size();
	
	cluster_info_of_thread[thread_id].clear();
	for (i=0; i<cluster_control.delta_steps; i++)
	{
		ii = get_next_task_number();
		if (ii >= 0)
		{
			delta = cluster_control.deltas_scaled[ii];
			if (team_size > 1) 
				density_cluster_tree_estimator[delta].set_info_level(INFO_SHUT_UP);

			lock_mutex();
			cluster_tree[delta] = empty_tree;
			unlock_mutex();
			
			density_cluster_tree_estimator[delta].import_spatial_information(&domain_partitioner, &neighborhood_lists);
			density_cluster_tree_estimator[delta].compute_cluster_tree(delta, current_cluster_tree, cluster_info, rho_star, current_cluster_infos);
			cluster_info_of_thread[thread_id] = cluster_info_of_thread[thread_id] + cluster_info;
			
			lock_mutex();
			cluster_tree[delta] = current_cluster_tree;

			if (team_size > 1)
			{
				flush_info(INFO_1, "\nThread %d computed cluster tree for delta = %1.4f,  tau = %1.4f, and eps = %1.4f", thread_id, delta, cluster_info.tau, cluster_info.eps);

				for (it=current_cluster_infos.begin(); it != current_cluster_infos.end(); ++it)
					if ((it->rho == 0.0) or (it->rho_star != NO_RHO))
						it->display(TRAIN_INFO_DISPLAY_FORMAT_REGULAR, INFO_1);
					
				if (cluster_info.number_of_peaks > 1)
					flush_info(INFO_1, "\nLargest clusters are detected at level rho = %2.4f.\n", rho_star);
				else if (cluster_info.number_of_peaks == 1)
					flush_info(INFO_1, "\nSingle cluster detected.\n");
				else
					flush_info(INFO_1, "\nNo cluster detected.\n");
			}

			cluster_infos.insert(cluster_infos.end(), current_cluster_infos.begin(), current_cluster_infos.end());

			cluster_compare_list[ii].delta = delta;
			cluster_compare_list[ii].rho_star = rho_star;
			cluster_compare_list[ii].rho_step = cluster_info.min_rho_step;
			cluster_compare_list[ii].number_of_clusters = cluster_info.number_of_clusters;
			cluster_compare_list[ii].number_of_peaks = cluster_info.number_of_peaks;
			
			cluster_compare_list[ii].rho_tolerance_factor = cluster_control.rho_compare_tolerance_factor;

			cluster_compare_list[ii].working_set_size = cluster_info.working_set_size;
			cluster_compare_list[ii].dataset_size = cluster_info.dataset_size;
			unlock_mutex();
		}
	}
}


//**********************************************************************************************************************************


void Tdensity_cluster_manager::compute_excess_mass_etc(Tcluster_info& cluster_info)
{
	unsigned ii;
	
	if (candidate_runs.size() == 1)
	{
		ii = candidate_runs[0];
		get_level_set(ii);
		cluster_compare_candidates[ii].density_integral = 0.0;
		cluster_compare_candidates[ii].constant_integral = 0.0;
	}
	else
	{
		random_level_subset.resize(cluster_control.test_set_size);
		
		thread_task = EXCESS_MASS_COMPUTE;
		start_threads();
	}
}



//**********************************************************************************************************************************

void Tdensity_cluster_manager::compute_excess_mass_etc_threaded()
{
	unsigned i;
	unsigned j;
	unsigned ii;
	double delta;
	unsigned thread_id;
	Tthread_chunk thread_chunk;
	

	thread_id = get_thread_id();
	thread_chunk = get_thread_chunk(cluster_control.test_set_size);
	
	for (i=0; i<candidate_runs.size(); i++)
	{
		ii = candidate_runs[i];
		delta = cluster_compare_candidates[ii].delta;
		if (is_first_team_member() == true)
		{
			density_cluster_tree_estimator[delta].import_spatial_information(&domain_partitioner, &neighborhood_lists);
			get_level_set(ii);
			init_random_generator(cluster_control.random_seed, 0);
			for (j=0; j<cluster_control.test_set_size; j++)
				random_level_subset[j] = get_random_number(0, int(level_set.size()) - 1);

			density_cluster_tree_estimator[delta].prepare_test_set(cluster_control.test_set_size, level_set, cluster_info_of_thread[thread_id]);
		}
		sync_threads();

		density_cluster_tree_estimator[delta].construct_test_set(thread_chunk.start_index, thread_chunk.stop_index, delta, random_level_subset, cluster_info_of_thread[thread_id]);
		excess_mass_of_thread[thread_id] = density_cluster_tree_estimator[delta].get_excess_mass(thread_chunk.start_index, thread_chunk.stop_index, cluster_info_of_thread[thread_id]);

		sync_threads();
		if (is_last_team_member() == true)
		{
			cluster_compare_candidates[ii].density_integral = 0.0;
			cluster_compare_candidates[ii].constant_integral = 0.0;
			for (j=0; j<get_team_size(); j++)
			{
				cluster_compare_candidates[ii].density_integral = cluster_compare_candidates[ii].density_integral + excess_mass_of_thread[j].density_integral;
				cluster_compare_candidates[ii].constant_integral = cluster_compare_candidates[ii].constant_integral + excess_mass_of_thread[j].constant_integral;
			}
		}
	}
}


//**********************************************************************************************************************************


void Tdensity_cluster_manager::get_level_set(unsigned candidate_index)
{
	unsigned j;
	double delta;
	map < double, Ttree_node <Tcluster> >::iterator root_it;
	
	
	level_set.clear();
	delta = cluster_compare_candidates[candidate_index].delta;
	
	if (cluster_compare_candidates[candidate_index].merge == NULL)
		for (root_it = cluster_tree[delta].root_nodes.begin(); root_it != cluster_tree[delta].root_nodes.end(); ++root_it)
			level_set.insert(level_set.end(), root_it->second.core_samples.begin(), root_it->second.core_samples.end());
	else
		for (j=0; j<cluster_compare_candidates[candidate_index].merge->child_nodes.size(); j++)
			level_set.insert(level_set.end(), cluster_compare_candidates[candidate_index].merge->child_nodes[j]->core_samples.begin(), cluster_compare_candidates[candidate_index].merge->child_nodes[j]->core_samples.end());
		
	cluster_compare_candidates[candidate_index].core_samples = level_set.size();
}


//**********************************************************************************************************************************


vector <double> Tdensity_cluster_manager::get_best_cluster_tree(Tcluster_compare& best_cluster_compare, Tcluster_info& cluster_info)
{
	unsigned c;
	double delta;
	double best_higher_rho_star;
	double largest_cluster_number;
	double parent_cluster_number;
	double current_cluster_number;
	unsigned number_of_clusters;
	unsigned number_of_peaks;
	map <double, double> cluster_number_transformation;
	vector <double> cluster_affiliation;
	vector <double> cluster_affiliation_empty;
	vector <double> cluster_affiliation_tmp;
	vector <double> core_cluster_affiliation;
	vector <unsigned> new_core_samples;
	double best_higher_delta;
	map < double, Ttree_node <Tcluster> >::iterator root_it;
	Ttree_node <Tcluster>* best_merge;
	Ttree_node <Tcluster>* current_node;
	Ttree_node <Tcluster>* current_node_in_final_tree;
	multimap <double, Ttree_node <Tcluster>* > subclusters_to_be_considered;	
	vector <double> true_labels;
	
	

	compute_all_cluster_trees(best_cluster_compare, cluster_info);

	cluster_trafo.clear();
	final_cluster_tree.clear();
	cluster_affiliation_empty.assign(dataset.size(), 0.0);
	

	if (best_cluster_compare.number_of_peaks == 0)
	{
		flush_info(INFO_1, "\n\nNo cluster detected.");
		
		return cluster_affiliation_empty;
	}
	
	delta = best_cluster_compare.delta;
	flush_info(INFO_1, "\n\nThe best delta = %2.4f achieved a rho* = %2.4f.", delta, best_cluster_compare.rho_star);
	flush_info(INFO_1, "\nThe initial cluster forest has %d trees.", cluster_tree[delta].root_nodes.size());
	
	true_labels = dataset.get_labels();
	dataset.set_labels(0.0);

	
	core_cluster_affiliation.assign(best_cluster_compare.working_set_size, 0.0);
	get_core_sample_affiliation_from_forest(&cluster_tree[delta], core_cluster_affiliation);
	cluster_affiliation = density_cluster_tree_estimator[delta].expand_from_core_to_train_samples(delta, 0.0, -1.0, cluster_info, core_cluster_affiliation, cluster_affiliation_empty);
	
	cluster_info.cluster_tree_build_time = get_wall_time_difference(cluster_info.cluster_tree_build_time);
	
	current_cluster_number = 0.0;
	for (root_it = cluster_tree[delta].root_nodes.begin(); root_it != cluster_tree[delta].root_nodes.end(); ++root_it)
	{
		current_cluster_number++;
		
		current_node = new Ttree_node <Tcluster> (root_it->second);
		current_node->depth = min(max(int(cluster_tree[delta].root_nodes.size()) - 1, 0), 1);
		current_node->width = delta;
		current_node->size = count(cluster_affiliation.begin(), cluster_affiliation.end(), current_cluster_number);
		current_node->cluster_number = current_cluster_number;
		current_node->parent_cluster_number = 0.0; 
		current_node->parent_node = NULL;
		current_node->child_nodes.clear();

		final_cluster_tree.root_nodes.insert(std::pair<double, Ttree_node <Tcluster> >(current_cluster_number, *(current_node)));
		subclusters_to_be_considered.insert(std::pair<double, Tcluster_tree_node* > (best_cluster_compare.rho_star, current_node));
	}
	largest_cluster_number = current_cluster_number;
	

	best_merge = NULL;
	number_of_peaks = 0;
	number_of_clusters = subclusters_to_be_considered.size();
	if (cluster_control.hierarchical_depth == UNLIMITED_CLUSTER_DEPTH)
		cluster_control.hierarchical_depth = std::numeric_limits<int>::max();

	while (subclusters_to_be_considered.empty() == false)
	{
		best_higher_rho_star = (*subclusters_to_be_considered.begin()).first;
		current_node = (*subclusters_to_be_considered.begin()).second;
		current_cluster_number = current_node->cluster_number;
		parent_cluster_number = current_node->parent_cluster_number;
		delta = current_node->width;
		
		subclusters_to_be_considered.erase(subclusters_to_be_considered.begin());

		current_node_in_final_tree = final_cluster_tree.start();
		while (current_node_in_final_tree->cluster_number != current_cluster_number)
			current_node_in_final_tree = final_cluster_tree.next();

		flush_info(INFO_1, "\n\nConsidering cluster %2.4f with %d core samples and parent cluster %2.4f and delta %2.4f and depth %d.", current_cluster_number, current_node->core_samples.size(), parent_cluster_number, delta, current_node->depth);
		
		
		assign_core_samples_of_cluster(current_node, current_node_in_final_tree, core_cluster_affiliation, parent_cluster_number, current_cluster_number, delta, cluster_affiliation, cluster_info);
		
		cluster_affiliation = density_cluster_tree_estimator[delta].expand_from_core_to_train_samples(delta, parent_cluster_number, current_cluster_number, cluster_info, core_cluster_affiliation, cluster_affiliation);

		if (int(current_node->depth) < cluster_control.hierarchical_depth)
		{
			find_higher_rho_stars(cluster_affiliation, best_higher_rho_star, current_cluster_number, best_higher_delta, best_merge, cluster_info);
			if (best_merge != NULL)
			{
				current_node_in_final_tree->rho_star = best_higher_rho_star;
				current_node_in_final_tree->child_nodes.resize(best_merge->child_nodes.size());
				for (c=0; c<best_merge->child_nodes.size(); c++)
				{
					current_node_in_final_tree->child_nodes[c] = new Ttree_node <Tcluster> (*best_merge->child_nodes[c]);
					
					current_node_in_final_tree->child_nodes[c]->depth = current_node->depth + 1;
					current_node_in_final_tree->child_nodes[c]->width = best_higher_delta;
					current_node_in_final_tree->child_nodes[c]->cluster_number = largest_cluster_number + 1.0 + double(c);
					current_node_in_final_tree->child_nodes[c]->parent_cluster_number = current_cluster_number;
					current_node_in_final_tree->child_nodes[c]->parent_node = current_node_in_final_tree;
					current_node_in_final_tree->child_nodes[c]->child_nodes.clear();
					
					subclusters_to_be_considered.insert(std::pair<double, Tcluster_tree_node* > (best_merge->rho_star, current_node_in_final_tree->child_nodes[c]));
				}
				largest_cluster_number = largest_cluster_number + double(best_merge->child_nodes.size());
					
				flush_info(INFO_1, "\nThe following subclusters have been found:");
				for (c=0; c<current_node_in_final_tree->child_nodes.size(); c++)
				{
					current_node_in_final_tree->child_nodes[c]->size = current_node_in_final_tree->child_nodes[c]->core_samples.size();
					
					flush_info(INFO_1, "\nCluster %2.4f with %d core samples at rho* %2.4f.", current_node_in_final_tree->child_nodes[c]->cluster_number, current_node_in_final_tree->child_nodes[c]->size, best_merge->rho_star);
				}

				number_of_clusters = number_of_clusters + current_node_in_final_tree->child_nodes.size();
			}
			else
				number_of_peaks++;
		}
		else
			number_of_peaks++;
	}
	flush_info(INFO_1, "\n\nAssigned %d cluster numbers and found %d peaks.", number_of_clusters, number_of_peaks); 

	
	cluster_info.cluster_tree_build_time = get_wall_time_difference(cluster_info.cluster_tree_build_time);
	
	
	dataset.set_labels(true_labels);
	if (dataset.is_classification_data() == true)
		change_cluster_number_according_to_labels(cluster_affiliation);


	return cluster_affiliation;
}


//**********************************************************************************************************************************

void Tdensity_cluster_manager::get_core_sample_affiliation_from_forest(Tcluster_tree* cluster_tree, vector <double>& core_cluster_affiliation)
{
	unsigned i;
	unsigned j;
	double current_cluster_number;
	Ttree_node <Tcluster>* current_node;
	map <double, double> cluster_number_transformation;
	map < double, Ttree_node <Tcluster> >::iterator root_it;


	current_cluster_number = 0.0;
	for (root_it = cluster_tree->root_nodes.begin(); root_it != cluster_tree->root_nodes.end(); ++root_it)
	{
		current_cluster_number++;
		current_node = &(root_it->second);
		for (i=0; i<current_node->core_samples.size(); i++)
		{
			j = current_node->core_samples[i];
			core_cluster_affiliation[j] = current_node->cluster_number;
		}
		
		cluster_number_transformation[root_it->first] = current_cluster_number;
	}
	
	if (cluster_number_transformation.size() > 1)
		for (i=0; i<core_cluster_affiliation.size(); i++)
			if (core_cluster_affiliation[i] > 0.0)
				core_cluster_affiliation[i] = cluster_number_transformation[core_cluster_affiliation[i]];
}


//**********************************************************************************************************************************

void Tdensity_cluster_manager::assign_core_samples_of_cluster(Ttree_node <Tcluster>* cluster_node, Ttree_node <Tcluster>* target_cluster_node, vector <double>& core_cluster_affiliation, double parent_cluster_number, double cluster_number, double delta, const vector <double> cluster_affiliation, Tcluster_info& cluster_info)
{
	if (parent_cluster_number == 0.0)
		return;
	
	if (cluster_control.density_estimator_type == DENSITY_HISTOGRAMM)
		core_cluster_affiliation.assign(cluster_node->max_number_of_core_samples, 0.0);

	target_cluster_node->core_samples = density_cluster_tree_estimator[delta].reassign_core_samples(cluster_node->core_samples, core_cluster_affiliation, parent_cluster_number, cluster_number, cluster_affiliation);
}



//**********************************************************************************************************************************

void Tdensity_cluster_manager::find_higher_rho_stars(const vector <double>& cluster_affiliation, double rho_star, double cluster_number, double& best_delta, Tcluster_tree_node*& best_merge, Tcluster_info& cluster_info)
{
	unsigned i;
	unsigned index_of_best_cluster;
	double current_key;
	double current_delta;
	double current_rho_star;
	unsigned peaks_in_node;
	unsigned peaks_in_cluster;
	Tcluster_tree_node* current_node;
	map < double, Ttree_node <Tcluster> >::iterator jt;
	my_unordered_map <double, Tcluster_tree >::iterator it;
	Tcluster_compare current_cluster_compare;
	Tcluster_compare best_cluster_compare;
	
	
	
	flush_info(INFO_1, "\nLooking for best subcluster of cluster %2.4f above %1.4f.", cluster_number, rho_star);
	
	
// Our first goal is to identify those deltas, for which there is a split in their cluster tree above rho_star

	cluster_compare_candidates.clear();
	for (it = cluster_tree.begin(); it != cluster_tree.end(); ++it)
	{
		current_delta = it->first;
		for (i=0; i<cluster_compare_list.size(); i++)
			if (current_delta == cluster_compare_list[i].delta)
				break;
			
		current_cluster_compare = cluster_compare_list[i];
		current_cluster_compare.merge = NULL;
		current_cluster_compare.rho_star = std::numeric_limits<double>::max();
		
		for (jt = it->second.root_nodes.begin(); jt != it->second.root_nodes.end(); ++jt)
		{
			current_key = jt->first;
			current_node = cluster_tree[current_delta].start(current_key);
			
			while (current_node != NULL)
			{
				current_rho_star = current_node->rho_star;
				if ((current_rho_star >= rho_star) and (current_rho_star <= current_cluster_compare.rho_star))
				{
					peaks_in_node = current_node->peaks.size();
					if (peaks_in_node > 1)
					{
						peaks_in_cluster = count_peaks_in_cluster(&(current_node->peaks), cluster_affiliation, cluster_number, rho_star);

						if (peaks_in_cluster == peaks_in_node)
						{
							current_cluster_compare.merge = current_node;
							current_cluster_compare.rho_star = current_rho_star;
							current_cluster_compare.number_of_peaks = peaks_in_cluster;
						}
					}
				}
				current_node = cluster_tree[current_delta].next();
			}
		}
		
		if (current_cluster_compare.rho_star < std::numeric_limits<double>::max())
			cluster_compare_candidates.push_back(current_cluster_compare);
	}

		
// Next we identify the best cluster among the identified candidates

	if (cluster_compare_candidates.size() == 0)
	{
		best_merge = NULL;
		best_delta = -1.0;
		flush_info(INFO_1, " Found none.");
	}
	else
	{
		identify_best_cluster(index_of_best_cluster, cluster_info);
		
		best_merge = cluster_compare_candidates[index_of_best_cluster].merge;
		best_delta = cluster_compare_candidates[index_of_best_cluster].delta;
	}
}





//**********************************************************************************************************************************

bool Tdensity_cluster_manager::peak_in_cluster(Tpeak* peak, const vector <double>& cluster_affiliation, double cluster_number, double rho_star)
{
	unsigned peak_samples_in_cluster;
	
	bool inside;
	my_unordered_set <unsigned>::iterator jt;
	

	if (peak->height < rho_star)
		return false;

	inside = true;
	peak_samples_in_cluster = 0;
	for (jt = peak->indices.begin(); jt != peak->indices.end(); ++jt)
		if (cluster_affiliation[*jt] == cluster_number)
			peak_samples_in_cluster++;
		
	if (2 * peak_samples_in_cluster > peak->indices.size())
		inside = true;
	else
		inside = false;
		
	return inside;
}



//**********************************************************************************************************************************

unsigned Tdensity_cluster_manager::count_peaks_in_cluster(vector <Tpeak>* peaks, const vector <double>& cluster_affiliation, double cluster_number, double rho_star)
{
	unsigned i;
	unsigned count;
	
	
	count = 0;
	for (i=0; i<peaks->size(); i++)
		if (peak_in_cluster(&((*peaks)[i]), cluster_affiliation, cluster_number, rho_star) == true)
			count++;
		
	return count;
}

//**********************************************************************************************************************************


void Tdensity_cluster_manager::assign_cluster_to_test_set_threaded()
{
	unsigned i;
	double delta;
	double cluster_number;
	double parent_cluster_number;
	vector <double> core_cluster_affiliation;
	Tthread_chunk thread_chunk;
	


	thread_chunk = get_thread_chunk(test_set.size());
	
	if (is_first_team_member() == true)
		current_node_for_all_threads = final_cluster_tree.start();
	sync_threads();
	
	while (current_node_for_all_threads != NULL)
	{
		delta = current_node_for_all_threads->width;
		cluster_number = current_node_for_all_threads->cluster_number;
		parent_cluster_number = current_node_for_all_threads->parent_cluster_number;
		
		if (cluster_control.density_estimator_type == DENSITY_HISTOGRAMM)
		{
			core_cluster_affiliation.assign(current_node_for_all_threads->max_number_of_core_samples, 0.0);
			for (i=0; i<current_node_for_all_threads->core_samples.size(); i++)
				core_cluster_affiliation[current_node_for_all_threads->core_samples[i]] = cluster_number;
			
			
			// Unfortunately, the expand routine for histograms is currently NOT THREAD SAFE.
			// We thus let the master thread do all the hard work. 
			
			if (is_first_team_member() == true)
				density_cluster_tree_estimator[delta].expand_from_core_to_test_samples(0, test_set.size(), delta, parent_cluster_number, cluster_number, new_test_affiliation, test_set, test_affiliation, current_node_for_all_threads->core_samples, core_cluster_affiliation);
			sync_threads();
		}
		else
			density_cluster_tree_estimator[delta].expand_from_core_to_test_samples(thread_chunk.start_index, thread_chunk.stop_index, delta, parent_cluster_number, cluster_number, new_test_affiliation, test_set, test_affiliation, current_node_for_all_threads->core_samples, core_cluster_affiliation);
		 
		for (i=thread_chunk.start_index; i<thread_chunk.stop_index; i++)
			test_affiliation[i] = new_test_affiliation[i];

		if (is_first_team_member() == true)
			current_node_for_all_threads = final_cluster_tree.next();
		sync_threads();
	}
	
	if (cluster_trafo.size() != 0)
		for (i=thread_chunk.start_index; i<thread_chunk.stop_index; i++)
			test_affiliation[i] = cluster_trafo[test_affiliation[i]];
}

//**********************************************************************************************************************************

vector <double> Tdensity_cluster_manager::assign_cluster_to_test_set(const Tdataset& test_set, Tcluster_info& cluster_info)
{
	cluster_info.test_time = get_wall_time_difference(cluster_info.test_time);
	
	flush_info(INFO_1, "\n\nAssigning %d test samples to clustering ...                     ", test_set.size());
	
	test_affiliation.assign(test_set.size(), 0.0);
	new_test_affiliation.assign(test_set.size(), 0.0);
	Tdensity_cluster_manager::test_set = test_set;
	
	
	thread_task = TEST_SET_ASSIGN;
	start_threads();
	
	flush_info(INFO_1, "done!");
	cluster_info.test_time = get_wall_time_difference(cluster_info.test_time);

	return test_affiliation;
}

//**********************************************************************************************************************************


vector <double> Tdensity_cluster_manager::get_cluster_numbers_of_clustering(const vector <double>& cluster_affiliation)
{
	unsigned i;
	my_unordered_set <double> cluster_numbers_tmp;
	vector <double> cluster_numbers;

	for (i = 0; i<cluster_affiliation.size(); i++)
		if (cluster_affiliation[i] != 0.0)
			cluster_numbers_tmp.insert(cluster_affiliation[i]);
		
	copy(cluster_numbers_tmp.begin(), cluster_numbers_tmp.end(), inserter(cluster_numbers, cluster_numbers.begin()));
	
	return cluster_numbers;
}

//**********************************************************************************************************************************

void Tdensity_cluster_manager::change_cluster_number_according_to_labels(vector <double>& cluster_affiliation)
{
	unsigned i;
	unsigned j;
	vector <int> labels;
	unsigned start_label_number;
	double max_cluster_number;
	Tdataset_info dataset_info;
	unsigned number_of_clusters;
	vector <double> cluster_numbers;
	vector <bool> leaf_cluster_number;
	vector <my_unordered_map <double, unsigned> > label_count_in_cluster;
	unsigned highest_label_occurence;
	Ttree_node <Tcluster>* current_node;
	std::my_unordered_set <unsigned>::iterator it;
	
	
	cluster_trafo.clear();
	cluster_numbers = get_cluster_numbers_of_clustering(cluster_affiliation);
	number_of_clusters = cluster_numbers.size();
	max_cluster_number = cluster_numbers[argmax(cluster_numbers)];

	dataset_info = Tdataset_info(dataset, true);
	labels = dataset_info.label_list;

	label_count_in_cluster.resize(max_cluster_number + 1);
	for (i=0; i<number_of_clusters; i++)
		for (j=0; j<labels.size(); j++)
			label_count_in_cluster[unsigned(cluster_numbers[i])].insert(std::pair <double, unsigned> (double(labels[j]), unsigned(0)));

		
	if (cluster_control.density_estimator_type == DENSITY_HISTOGRAMM)
	{
		for (i=0; i<dataset.size(); i++)
			if (cluster_affiliation[i] > 0.0)
			label_count_in_cluster[unsigned(cluster_affiliation[i])][dataset.sample(i)->label]++;
	}
	else
	{
		leaf_cluster_number.resize(max_cluster_number + 1);
		current_node = final_cluster_tree.start();
		while (current_node != NULL)
		{
			if (current_node->child_nodes.size() == 0)
				leaf_cluster_number[unsigned(current_node->cluster_number)] = true;
			else
				leaf_cluster_number[unsigned(current_node->cluster_number)] = false;
			current_node = final_cluster_tree.next();
		}
		
		flush_info(INFO_1, "\n");
		current_node = final_cluster_tree.start();
		while (current_node != NULL)
		{
			if (current_node->child_nodes.size() == 0)
			{
				flush_info(INFO_1, "\nDetermining new cluster number at the peaks of leaf with number %1.0f.", current_node->cluster_number);
				
				for (it = current_node->peaks[0].indices.begin(); it != current_node->peaks[0].indices.end(); ++it)
					label_count_in_cluster[current_node->cluster_number][dataset.sample(*it)->label]++;
			}
			else
			{
				flush_info(INFO_1, "\nDetermining new cluster number at the core samples of branch with number %1.0f.", current_node->cluster_number);
				
				for (j=0; j<current_node->core_samples.size(); j++)
				{
					i = current_node->core_samples[j];
					if ((current_node->cluster_number == cluster_affiliation[i]) and (leaf_cluster_number[unsigned(cluster_affiliation[i])] == false))
						label_count_in_cluster[current_node->cluster_number][dataset.sample(i)->label]++;
				}
			}
			current_node = final_cluster_tree.next();
		}
	}
	
		
	if (labels[0] != 0)
		start_label_number = 0;
	else
		start_label_number = 1;

	for (i=0; i<number_of_clusters; i++)
		if (start_label_number < labels.size())
		{
			highest_label_occurence = label_count_in_cluster[cluster_numbers[i]][labels[start_label_number]];
			cluster_trafo[cluster_numbers[i]] = double(labels[start_label_number]); 

			for (j = start_label_number + 1; j<labels.size(); j++)
				if (label_count_in_cluster[cluster_numbers[i]][labels[j]] > highest_label_occurence)
				{
					cluster_trafo[cluster_numbers[i]] = double(labels[j]);
					highest_label_occurence = label_count_in_cluster[cluster_numbers[i]][labels[j]];
				}
		}
		else
			cluster_trafo[cluster_numbers[i]] = cluster_numbers[i];
	

	flush_info(INFO_1, "\n");
	for (i=0; i<number_of_clusters; i++)
		flush_info(INFO_1, "\nChanging cluster number %1.0f to %1.0f.", cluster_numbers[i], cluster_trafo[cluster_numbers[i]]);

	
	for (i=0; i<cluster_affiliation.size(); i++)
		cluster_affiliation[i] = cluster_trafo[cluster_affiliation[i]];
}


//**********************************************************************************************************************************

vector <double> Tdensity_cluster_manager::compute_test_error(const vector <double>& cluster_affiliation, const vector <double>& true_cluster_affiliation, vector <unsigned>& true_cluster_sizes)
{
	unsigned i;
	unsigned cn;
	double min_cluster_number;
	double max_cluster_number;
	vector <double> cluster_numbers;
	vector <double> errors_on_clusters;
	
	
	
	if (cluster_affiliation.size() != true_cluster_affiliation.size())
		flush_exit(ERROR_DATA_MISMATCH, "Cannot compare cluster affiliations of sizes %d and %d.", cluster_affiliation.size(), true_cluster_affiliation.size());
	
	cluster_numbers = get_unique_entries(true_cluster_affiliation);

	min_cluster_number = min(0.0, cluster_numbers[argmin(cluster_numbers)]);
	max_cluster_number = cluster_numbers[argmax(cluster_numbers)];

	true_cluster_sizes.assign(unsigned(max_cluster_number - min_cluster_number) + 1, 0);
	errors_on_clusters.assign(unsigned(max_cluster_number - min_cluster_number) + 1, 0.0);

	for (i=0; i<true_cluster_affiliation.size(); i++)
	{
		cn = unsigned(true_cluster_affiliation[i] - min_cluster_number);
		true_cluster_sizes[cn]++;
		if (true_cluster_affiliation[i] != cluster_affiliation[i])
			errors_on_clusters[cn] = errors_on_clusters[cn] + 1.0;
	}

	flush_info(INFO_1, "\n");
	for (cn=0; cn<true_cluster_sizes.size(); cn++)
		if ((true_cluster_sizes[cn] > 0) and (cn != unsigned(- min_cluster_number)))
			flush_info(INFO_1, "\nCluster %2.0f: Missed %4d out of %4d samples.", double(cn) + min_cluster_number, unsigned(errors_on_clusters[cn]), true_cluster_sizes[cn]);
		
	if (true_cluster_sizes[unsigned(- min_cluster_number)] > 0)
		flush_info(INFO_1, "\nAssigned %d samples from neutral region of size %d to clusters.", unsigned(errors_on_clusters[unsigned(- min_cluster_number)]), true_cluster_sizes[unsigned(- min_cluster_number)]);
	
	for (i=0; i<true_cluster_sizes.size(); i++)
		errors_on_clusters[i] = errors_on_clusters[i] / double(true_cluster_affiliation.size());
		
	return errors_on_clusters;
}




//**********************************************************************************************************************************

void Tdensity_cluster_manager::show_clustering_statistics(const vector <double>& cluster_affiliation)
{
	unsigned c;
	vector <double> cluster_numbers;
	
	cluster_numbers = get_unique_entries(cluster_affiliation);
	sort_up(cluster_numbers);
	
	
	for (c=0; c<cluster_numbers.size(); c++)
		if (cluster_numbers[c] > 0.0)
			flush_info(INFO_1, "\nCluster %2.4f has %d samples.", cluster_numbers[c], count(cluster_affiliation.begin(), cluster_affiliation.end(), cluster_numbers[c]));
	flush_info(INFO_1, "\n");
}




//**********************************************************************************************************************************

void Tdensity_cluster_manager::check_tree_consistency(Tcluster_tree* cluster_tree)
{
	unsigned peaks_in_roots;
	unsigned leafs;
	double cluster_number;
	double parent_cluster_number;
	double true_parent_cluster_number;
	Ttree_node <Tcluster>* current_node;
	my_unordered_set <double> cluster_numbers;
	std::my_unordered_set <double>::iterator it;
	std::map < double, Ttree_node <Tcluster> >::iterator jt;
	
	
	peaks_in_roots = 0;
	for (jt = cluster_tree->root_nodes.begin(); jt != cluster_tree->root_nodes.end(); ++jt)
	{
		peaks_in_roots = peaks_in_roots + jt->second.peaks.size();
		flush_info(INFO_1, "\nRoot number %2.4f with cluster number %2.4f, %d peaks, and %d children", jt->first, jt->second.cluster_number, jt->second.peaks.size(), jt->second.child_nodes.size());
	}
	
	leafs = 0;
	current_node = cluster_tree->start();
	while (current_node != NULL)
	{
		cluster_number = current_node->cluster_number;
		parent_cluster_number = current_node->parent_cluster_number;
		if (current_node->child_nodes.size() == 0)
			leafs++;

		flush_info(INFO_1, "\nCluster %2.4f with parent %2.4f at rho* %2.4f at address %p with %d core samples, %d peaks, and %d children", cluster_number, parent_cluster_number, current_node->rho, current_node, current_node->core_samples.size(), current_node->peaks.size(), current_node->child_nodes.size());
		
		it = cluster_numbers.find(cluster_number);
		if ((it != cluster_numbers.end()) and (cluster_tree == &final_cluster_tree))
			flush_exit(ERROR_DATA_STRUCTURE, "Cluster %2.4f appears twice.", cluster_number);
		else
			cluster_numbers.insert(cluster_number);
		
		if (current_node->parent_node != NULL)
			true_parent_cluster_number = current_node->parent_node->cluster_number;
		else
			true_parent_cluster_number = 0.0;
		
		if (true_parent_cluster_number != parent_cluster_number)
			flush_exit(ERROR_DATA_STRUCTURE, "Cluster %2.4f stored %2.4f as parent, but parent has cluster number %2.4f.", cluster_number, parent_cluster_number, true_parent_cluster_number);
		
		if (cluster_number <= 0.0)
			flush_exit(ERROR_DATA_STRUCTURE, "Cluster tree contains cluster with number %2.4f.", cluster_number);
		
		if (current_node->core_samples.size() == 0)
			flush_exit(ERROR_DATA_STRUCTURE, "Cluster tree contains clusters without core samples.");
		
		current_node = cluster_tree->next();
	}
	
	if (cluster_tree != &final_cluster_tree)
	{
		if (peaks_in_roots != leafs)
			flush_exit(ERROR_DATA_STRUCTURE, "Cluster tree has %d peaks but %d leafs.", peaks_in_roots, leafs);
	
		flush_info(INFO_1, "\nCluster tree has %d peaks and leafs.", peaks_in_roots);
	}
}




#endif
