// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (DENSITY_CLUSTER_TREE_CPP) 
 #define DENSITY_CLUSTER_TREE_CPP


#include "sources/cluster/solver/density_cluster_tree.h" 

 
 
#include "sources/shared/basic_types/vector.h"
#include "sources/shared/basic_types/my_unordered_map_and_set.h"
#include "sources/shared/system_support/timing.h"
#include "sources/shared/basic_functions/random_subsets.h"


#include <stack>



//**********************************************************************************************************************************

Tdensity_cluster_tree::Tdensity_cluster_tree()
{
	clear();
}

//**********************************************************************************************************************************


Tdensity_cluster_tree::~Tdensity_cluster_tree()
{
	clear();
}


//**********************************************************************************************************************************


void Tdensity_cluster_tree::clear()
{
	Tdensity_estimator::clear();

	true_size = 0;
	
	old_rho = NO_RHO;

	density_estimate.clear();
	density_ordering.clear();
	
	cluster_affiliations.clear();
}



//**********************************************************************************************************************************


void Tdensity_cluster_tree::load(const Tdataset& dataset, const Tdataset& random_set, const Tcluster_control& cluster_control, Tcluster_info& cluster_info)
{
// This routine assumes that cluster_control has alreadybeen correctly set for ground truth data
// (and potentially for other density estimators as well)!
	
	cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
	clear();

	Tdensity_cluster_tree::cluster_control = cluster_control;
	
	switch(cluster_control.get_density_estimator_category())
	{
		case DENSITY_HISTOGRAMM:
			Tdensity_estimator::load(dataset, random_set, cluster_control, false);
			dataset_info = Tdataset_info(Tdensity_cluster_tree::dataset, false);
			break;

		case DENSITY_GENERIC_KERNEL:
			Tdensity_estimator::load(dataset, random_set, cluster_control, true);
			dataset_info = Tdataset_info(dataset, true);
			break;
	
		case DENSITY_TRUTH:
			Tdensity_estimator::load(dataset, random_set, cluster_control, true);
			dataset_info = Tdataset_info(dataset, true);
			break;
	}
	dataset_info_computed = true;
	cluster_info.data_load_time = get_wall_time_difference(cluster_info.data_load_time);
}




//**********************************************************************************************************************************


double Tdensity_cluster_tree::get_rho_step(double current_rho) const
{
	unsigned i;
	unsigned level_set_start_tmp;
	
	if (cluster_control.density_estimator_type == DENSITY_TRUTH)
	{
		if (current_rho == NO_RHO)
			return density_estimate[density_ordering[1]] - density_estimate[density_ordering[0]];
		
		level_set_start_tmp = level_set_stop;
		for (i=level_set_start_tmp; i<working_set_size; i++)
			if (density_estimate[density_ordering[i]] < current_rho)
				break;
		if (i == working_set_size)
			return density_estimate[density_ordering[int(working_set_size) - 1]] + 1.0;
		else
			return current_rho - density_estimate[density_ordering[i]];
	}
	else
		return get_scale_factor(get_current_delta(), true_size);
}


//**********************************************************************************************************************************


double Tdensity_cluster_tree::get_max_rho_precision() const
{
	if (cluster_control.density_estimator_type == DENSITY_TRUTH)
		return 0.0000000001;
	else
		return get_rho_step(NO_RHO)/2.0;
}



//**********************************************************************************************************************************


double Tdensity_cluster_tree::get_start_rho() const
{
	double rho_step;
	
	if (cluster_control.density_estimator_type == DENSITY_TRUTH)
		return density_estimate[density_ordering[0]];
	else
	{
		rho_step = get_rho_step(NO_RHO);
		return floor((get_max_density() - get_eps_uncertainty())/rho_step) * rho_step;
	}
}


//**********************************************************************************************************************************


double Tdensity_cluster_tree::get_eps_uncertainty() const
{
	if (density_estimate.size() > 0)
	{
// 	The default factor 3.0 is a rather aggressive value, which also assumes 
// 		a bounded density.  See e.g. Theorem 3.1 in Steinwart, AoS 2015, for 
// 		concentration bounds related to the histogram rule.

		return cluster_control.eps_factor * sqrt(get_max_density() * get_scale_factor(get_current_delta(), true_size));
	}
	else
		return 0.0;
}



//**********************************************************************************************************************************


void Tdensity_cluster_tree::provide_density_estimate(double delta, Tcluster_info& cluster_info)
{
	vector <double> density_tmp;
	
	
	density_control.max_delta_neighborhood = cluster_control.get_tau_from_delta(delta);
	compute_density_estimate(delta, cluster_info);
	
	level_set_start = 0;
	level_set_stop = 0;
	
	cluster_info.density_compute_time = get_wall_time_difference(cluster_info.density_compute_time);
	density_ordering = id_permutation(working_set_size);
	density_tmp = density_estimate;
	sort_down(density_tmp, density_ordering);
	cluster_info.density_compute_time = get_wall_time_difference(cluster_info.density_compute_time);
}


//**********************************************************************************************************************************


void Tdensity_cluster_tree::get_level_set(double rho)
{
	unsigned i;



	level_set_start = level_set_stop;
	if (cluster_control.density_estimator_type != DENSITY_TRUTH)
	{
		for (i=level_set_start; i<working_set_size; i++)
			if (density_estimate[density_ordering[i]] < rho - get_max_rho_precision())
				break;
	}
	else
	{
		for (i=level_set_start; i<working_set_size; i++)
			if (density_estimate[density_ordering[i]] < density_estimate[density_ordering[level_set_start]])
				break;
	}

	level_set_stop = i;
}


//**********************************************************************************************************************************

void Tdensity_cluster_tree::get_cluster(double rho, Tcluster_info& cluster_info)
{
	unsigned i;
	unsigned ii;
	unsigned j;
	unsigned k;
	unsigned merge_type;
	double eps;
	double tau;
	double density;
	double cluster_of_sample;
	double previous_largest_cluster_number;
	double new_cluster_number;
	double target_cluster_number;
	double omit_cluster_number;
	double current_cluster_number;
	unsigned new_samples_to_level;
	unsigned assigned_samples;
	cluster_number_store_type current_clusters;
	cluster_number_store_type::iterator it;
	vector <bool> cluster_target_of_merge;
	stack <unsigned> neighbors;

	
// 	Get the current level set

	cluster_info.level_set_compute_time = get_wall_time_difference(cluster_info.level_set_compute_time);
	if ((old_rho == NO_RHO) or (rho > old_rho))
		cluster_affiliations.reserve(working_set_size);
	cluster_affiliations.new_level_set();

	cluster_info.number_of_new_peaks = 0;
	cluster_info.number_of_new_merges = 0;
	cluster_info.changed_level_set = false;
	cluster_info.rho = rho;

	get_level_set(rho);
	new_samples_to_level = level_set_stop - level_set_start;
	cluster_info.level_set_compute_time = get_wall_time_difference(cluster_info.level_set_compute_time);

	if (new_samples_to_level == 0)
		return;

	
// 	Look, if new clusters appear
	
	cluster_info.cluster_search_time = get_wall_time_difference(cluster_info.cluster_search_time);
	
	tau = cluster_control.get_tau_from_delta(get_current_delta());
	eps = get_eps_uncertainty();
	
	previous_largest_cluster_number = cluster_affiliations.get_largest_cluster_number();
	cluster_target_of_merge.resize(previous_largest_cluster_number + 1.0, false);

	current_clusters = cluster_affiliations.get_current_cluster_numbers();
	for (it = current_clusters.begin(); it!= current_clusters.end(); ++it)
		if ((*it < 0.0) and (cluster_affiliations.get_maximum_of_cluster(*it).height >= rho + eps - get_max_rho_precision()))
		{
			new_cluster_number =  1.0 + cluster_affiliations.get_largest_cluster_number();
			cluster_affiliations.change_cluster_number(*it, new_cluster_number);
			flush_info(INFO_2, "\nNew cluster with number %2.4f and %d samples appears at rho = %2.4f.", new_cluster_number, cluster_affiliations.get_size_of_cluster(new_cluster_number), rho);
			
			insert_root_in_cluster_tree(new_cluster_number, rho);
		}
	cluster_info.cluster_search_time = get_wall_time_difference(cluster_info.cluster_search_time);

	
// 	Now, assign all new samples to their clusters
	
	assigned_samples = 0;
	cluster_info.changed_level_set = true;
	do
	{
// 	Look for the next sample in the level set, that has not been assigned
		
		cluster_info.cluster_search_time = get_wall_time_difference(cluster_info.cluster_search_time);
		for (i=level_set_start; i<level_set_stop; i++)
			if (cluster_affiliations.get_cluster_number(density_ordering[i]) == 0.0)
				break;
		if (i==level_set_stop)
			break;
		ii = density_ordering[i];
		cluster_info.cluster_search_time = get_wall_time_difference(cluster_info.cluster_search_time);


// 	Assign a new cluster number to the unassigned sample
		
		cluster_info.cluster_connect_time = get_wall_time_difference(cluster_info.cluster_connect_time);
		neighbors.push(ii);
		assigned_samples++;

		if (density_estimate[ii] >= rho + eps - get_max_rho_precision())
		{
			current_cluster_number = 1.0 + cluster_affiliations.get_largest_cluster_number();
			flush_info(INFO_2, "\nNew cluster with number %2.4f appears at rho %2.4f.", current_cluster_number, rho);
		}
		else
			current_cluster_number = -1.0 + cluster_affiliations.get_smallest_cluster_number();
		
		cluster_affiliations.assign_sample_to_cluster(ii, current_cluster_number, density_estimate[ii]);
		if (current_cluster_number > 0.0)
			insert_root_in_cluster_tree(current_cluster_number, rho);
		

// 	Find connected component of unassigned sample
		
		do
		{
// 		Fix the sample we are currently working with
			
			ii = neighbors.top();
			neighbors.pop();
			if (neighborhood_lists->neighbors[ii].size() > 0)
			{
// 			Look at all tau-neighbors of the current sample. 
				
				j = 0;
				while (neighborhood_lists->distances[ii][j] <= tau)
				{
					k = neighborhood_lists->neighbors[ii][j];
					density = density_estimate[k];
					
// 				If the current tau-neighbor is in the rho-level set, we need 
// 				to work it, otherwise we can ignore it
					
					if (density >= rho - get_max_rho_precision())
					{
						cluster_of_sample = cluster_affiliations.get_cluster_number(k);

// 					If the current tau-neighbor has not been assigned to a cluster, it 
// 					is now assigned to the cluster with the current_cluster_number.
						
						if (cluster_of_sample == 0.0)
						{
							neighbors.push(k);
							cluster_affiliations.assign_sample_to_cluster(k, current_cluster_number, density);

							assigned_samples++;
							if ((density >= rho + eps - get_max_rho_precision()) and (current_cluster_number < 0.0))
							{
								new_cluster_number =  1.0 + cluster_affiliations.get_largest_cluster_number();
								cluster_affiliations.change_cluster_number(current_cluster_number, new_cluster_number);
								insert_root_in_cluster_tree(new_cluster_number, rho);
								
								flush_info(INFO_2, "\nChanged cluster number %2.4f to %2.4f for %d samples at rho = %2.4f.", current_cluster_number, new_cluster_number, cluster_affiliations.get_size_of_cluster(new_cluster_number), rho);
								
								current_cluster_number = new_cluster_number;
							}
						}
						else if (cluster_of_sample != current_cluster_number)
						{
// 						Otherwise, we need to merge the clusters ...
// 						Interesting new merges of clusters are only those with positive cluster 
// 						number, which have been around before the current call of get_clusters(...)

							target_cluster_number = min(cluster_of_sample, current_cluster_number);
							omit_cluster_number = max(cluster_of_sample, current_cluster_number);
							
							if (target_cluster_number > 0.0)
							{
								flush_info(INFO_2, "\nMerging cluster with number %2.4f into cluster with number %2.4f at rho = %2.4f.", omit_cluster_number, target_cluster_number, rho);
								
								
								if (omit_cluster_number <= previous_largest_cluster_number)
								{
									merge_type = unsigned(cluster_target_of_merge[unsigned(omit_cluster_number)]) + 2 * unsigned(cluster_target_of_merge[unsigned(target_cluster_number)]);
									
									switch(merge_type)
									{
										case 0:
											cluster_tree.merge_roots(target_cluster_number, cluster_of_sample, current_cluster_number);
											break;
										case 1:
											cluster_tree.unite_roots(target_cluster_number, target_cluster_number, omit_cluster_number);
											break;
										case 2:
											cluster_tree.merge_roots(target_cluster_number, omit_cluster_number);
											break;
										case 3:
											cluster_tree.unite_roots(target_cluster_number, omit_cluster_number);
											break;
									}
									cluster_target_of_merge[unsigned(target_cluster_number)] = true;
									cluster_tree.root_nodes[target_cluster_number].rho = rho;
									cluster_tree.root_nodes[target_cluster_number].rho_star = old_rho;

									cluster_info.number_of_new_merges++;
								}
								else
									cluster_tree.root_nodes.erase(omit_cluster_number);
							}

							cluster_affiliations.merge_clusters(cluster_of_sample, current_cluster_number, new_cluster_number);
							current_cluster_number = new_cluster_number;
						}
					}
					j++;
					if (j == neighborhood_lists->distances[ii].size())
						break;
				}
			}
		}
		while (neighbors.empty() == false);
		
		cluster_info.cluster_connect_time = get_wall_time_difference(cluster_info.cluster_connect_time);
	}
	while (assigned_samples < new_samples_to_level);
	

	current_clusters = cluster_affiliations.get_current_cluster_numbers();
	for (it = current_clusters.begin(); it!= current_clusters.end(); ++it)
		if (*it > previous_largest_cluster_number)
			cluster_info.number_of_new_peaks++;
		
		
	cluster_info.number_of_peaks = cluster_info.number_of_peaks + cluster_info.number_of_new_peaks;
	cluster_info.number_of_merges = cluster_info.number_of_merges + cluster_info.number_of_new_merges;
	cluster_info.number_of_clusters = cluster_affiliations.get_number_of_current_active_cluster();
	cluster_info.number_of_assigned_samples = cluster_affiliations.get_number_of_assigned_samples();
	cluster_info.number_of_samples_with_new_cluster_number = cluster_affiliations.get_number_of_samples_with_new_cluster_number();
	cluster_info.number_of_newly_assigned_samples = new_samples_to_level;
	
	cluster_info.working_set_size = working_set_size;
	cluster_info.dataset_size = true_size;
	
	old_rho = rho;
}




//**********************************************************************************************************************************

void Tdensity_cluster_tree::convert_mid_points_to_peak_samples()
{
	unsigned i;
	unsigned j;
	std::my_unordered_set <unsigned>::iterator jt;
	Ttree_node <Tcluster>* current_node;
	vector <double> current_peak_samples;
	list <unsigned> samples_in_peak;
	

	current_node = cluster_tree.start();
	while (current_node != NULL)
	{
		for (i=0; i<current_node->peaks.size(); i++)
		{
			current_peak_samples.clear();
			for (jt = current_node->peaks[i].indices.begin(); jt != current_node->peaks[i].indices.end(); ++jt)
			{
				samples_in_peak = histogram.get_indices_in_cell(*working_set.sample(*jt));
				current_peak_samples.insert(current_peak_samples.begin(), samples_in_peak.begin(),samples_in_peak.end());
			}
			current_node->peaks[i].indices.clear();
			for (j=0; j<current_peak_samples.size(); j++)
				current_node->peaks[i].indices.insert(current_peak_samples[j]);
		}
		current_node = cluster_tree.next();
	}

}



//**********************************************************************************************************************************

void Tdensity_cluster_tree::compute_cluster_tree(double delta, Tcluster_tree& new_cluster_tree, Tcluster_info& cluster_info, double& rho_star, list <Tcluster_info>& all_cluster_infos)
{
	double min_rho;
	unsigned number_of_rho_steps;
	Tcluster_info cluster_info_tmp;
	cluster_number_store_type::iterator it;
	Tcluster_tree empty_cluster_tree;
	
	
	cluster_info.clear();
	cluster_infos.clear();
	provide_density_estimate(delta, cluster_info);
	
	flush_info(info_level, "\nComputing cluster tree for delta = %1.4f,  tau = %1.4f, and eps = %1.4f", delta, cluster_control.get_tau_from_delta(delta), get_eps_uncertainty());
	cluster_affiliations.reserve(working_set_size);
	
	old_rho = NO_RHO;
	cluster_tree = empty_cluster_tree;
	
	if (cluster_control.density_estimator_type == DENSITY_TRUTH)
		cluster_info.min_rho_step = 0.0;
	else
		cluster_info.min_rho_step = get_scale_factor(delta, true_size);

	if ((cluster_control.density_estimator_type == DENSITY_TRUTH) and (get_min_density() == 0.0))
		min_rho = 2.0 * get_max_rho_precision();
	else
		min_rho = 0.0;
	
	rho_star = NO_RHO;
	do
	{
		get_next_merge(min_rho, cluster_info_tmp, number_of_rho_steps);

		if (cluster_info_tmp.rho_star != NO_RHO)
			rho_star = cluster_info_tmp.rho_star;

		if ((cluster_info_tmp.rho_start >= -get_max_rho_precision()) and (cluster_infos.empty() == false))
			cluster_infos.pop_back();
		cluster_infos.push_back(cluster_info_tmp);

		cluster_info = cluster_info + cluster_info_tmp;
		cluster_info_tmp.display(TRAIN_INFO_DISPLAY_FORMAT_REGULAR, info_level);
	}
	while (cluster_info_tmp.rho_stop > min_rho);

	
	cluster_info.copy_hyper_parameters(cluster_info_tmp);
	cluster_info.copy_cluster_results(cluster_info_tmp);
	
	if (cluster_control.density_estimator_type == DENSITY_HISTOGRAMM)
		convert_mid_points_to_peak_samples();

	if (cluster_info.number_of_peaks > 1)
		flush_info(info_level, "\nLargest clusters are detected at level rho = %2.4f.\n", rho_star);
	else if (cluster_info.number_of_peaks == 1)
		flush_info(info_level, "\nSingle cluster detected.\n");
	else
		flush_info(info_level, "\nNo cluster detected.\n");
	

	new_cluster_tree = cluster_tree;
	all_cluster_infos = cluster_infos;
}




//**********************************************************************************************************************************

void Tdensity_cluster_tree::insert_root_in_cluster_tree(double cluster_number, double rho)
{
	Ttree_node <Tcluster>* new_root;
	
	
	cluster_tree.new_root(cluster_number);
	
	new_root = &(cluster_tree.root_nodes[cluster_number]);
	
	new_root->cluster_number = cluster_number;
	new_root->width = get_current_delta();
	new_root->peaks.push_back(cluster_affiliations.get_maximum_of_cluster(cluster_number));
	new_root->max_number_of_core_samples = working_set_size;
}




//**********************************************************************************************************************************

void Tdensity_cluster_tree::get_next_merge(double min_rho, Tcluster_info& cluster_info, unsigned& number_of_rho_steps)
{
	unsigned i;
	double rho;
	double old_rho_step;
	double rho_step;
	unsigned number_of_new_peaks;
	unsigned index;
	double cluster_number;
	double largest_hidden_maximum;
	vector <double> cluster_affiliation;
	std::map < double, Ttree_node <Tcluster> >::iterator cit;


	rho_step = get_rho_step(old_rho);
	old_rho_step = rho_step;

	if (cluster_info.rho_start == NO_RHO)
		cluster_info.rho_start = get_start_rho();
	else
		cluster_info.rho_start = cluster_info.rho_stop - cluster_info.rho_step - rho_step;


	cluster_info.clear_time();
	cluster_info.rho_step = rho_step;
	cluster_info.rho_stop = NO_RHO;
	cluster_info.width = get_current_delta();
	cluster_info.tau = cluster_control.get_tau_from_delta(get_current_delta());
	cluster_info.eps = get_eps_uncertainty();
	cluster_info.rho_star = NO_RHO;
	
	rho = cluster_info.rho_start;
	number_of_new_peaks = 0;
	if (rho >= min_rho - get_max_rho_precision())
		flush_info(INFO_2, "\nStarting at level rho = %2.4f with step size %2.4f", rho, rho_step);

	number_of_rho_steps = 0;
	while (rho >= min_rho - get_max_rho_precision())
	{
		old_rho_step = rho_step;
		rho_step = get_rho_step(rho);
		get_cluster(rho, cluster_info);

		if (cluster_info.number_of_new_merges > 0)
			cluster_info.rho_star = rho + old_rho_step;
		cluster_infos.push_back(cluster_info);
		cluster_info.display(TRAIN_INFO_DISPLAY_FORMAT_REGULAR, INFO_3);
		
		cluster_info.cluster_admin_time = get_wall_time_difference(cluster_info.cluster_admin_time);
		
		if (cluster_info.number_of_new_merges == 0)
		{
			// 	This part is currently wrong: one needs to store and get information which clusters were gone
			// 	and which one were merged. This explain the if-clause at the beginning.
			
			cluster_affiliations.get_sample_with_new_cluster_number(index, cluster_number);
			while (cluster_number > 0.0)
			{
				cluster_tree.root_nodes[cluster_number].core_samples.push_back(index);
				cluster_affiliations.get_sample_with_new_cluster_number(index, cluster_number);
			}
		}
		else
		{
			cluster_affiliation = cluster_affiliations.get_cluster_affiliations();

			for (cit = cluster_tree.root_nodes.begin(); cit != cluster_tree.root_nodes.end(); ++cit)
				cit->second.core_samples.clear();
			for (i=0; i<cluster_affiliation.size(); i++)
				if (cluster_affiliation[i] > 0.0)
					cluster_tree.root_nodes[cluster_affiliation[i]].core_samples.push_back(i);
		}
			
		cluster_info.cluster_admin_time = get_wall_time_difference(cluster_info.cluster_admin_time);

		number_of_new_peaks = number_of_new_peaks + cluster_info.number_of_new_peaks;
		if (cluster_info.number_of_new_merges > 0)
			break;

		
// 	Try to take a larger rho_step if possible
		
		if (cluster_control.density_estimator_type != DENSITY_TRUTH)
			if (cluster_info.number_of_clusters == 1)
			{
				largest_hidden_maximum = cluster_affiliations.get_largest_inactive_maximum();
				if (largest_hidden_maximum == -1.0)
					rho_step = max(rho_step, rho - max(0.0, rho_step * floor((rho - cluster_info.eps)/rho_step)));
				else
					rho_step = max(rho_step, rho - max(0.0, rho_step * floor((largest_hidden_maximum - cluster_info.eps)/rho_step)));
			}
		
		rho = rho - rho_step;
		number_of_rho_steps++;
	}

	cluster_info.cluster_admin_time = get_wall_time_difference(cluster_info.cluster_admin_time);
	
	cluster_info.rho_step = old_rho_step;
	if ((rho + old_rho_step > min_rho + get_max_rho_precision()) and (old_rho > get_max_rho_precision()))
		cluster_info.rho_stop = rho + old_rho_step;
	else
		cluster_info.rho_stop = min_rho;
	cluster_info.number_of_new_peaks = number_of_new_peaks;

	if ((cluster_info.number_of_new_merges == 0) and (cluster_info.number_of_clusters > 1))
		cluster_info.rho_star = min_rho;

	if (flush_info_will_show(INFO_3) == true)
	{
		flush_info(INFO_3, "\nThe following clusters have been found:");
		for (cit = cluster_tree.root_nodes.begin(); cit != cluster_tree.root_nodes.end(); ++cit)
			if (cit->second.core_samples.size() == 0)
				flush_exit(ERROR_DATA_MISMATCH, "Cluster %2.4f has no core samples", cit->second.cluster_number);
			else
				flush_info(INFO_3, "\nCluster %2.4f with %d core samples and %d peaks", cit->second.cluster_number, cit->second.core_samples.size(), cit->second.peaks.size());
	}

	cluster_info.cluster_admin_time = get_wall_time_difference(cluster_info.cluster_admin_time);
}


 
//**********************************************************************************************************************************


#endif
