// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (DENSITY_ESTIMATOR_CPP) 
 #define DENSITY_ESTIMATOR_CPP



#include "sources/cluster/solver/density_estimator.h" 



#include "sources/shared/system_support/timing.h"

#include "sources/shared/basic_functions/math_misc.h"
#include "sources/shared/basic_functions/random_generator.h"


#include "sources/shared/basic_types/vector.h"

//**********************************************************************************************************************************



Tdensity_estimator::Tdensity_estimator()
{
	clear();
}


//**********************************************************************************************************************************


Tdensity_estimator::~Tdensity_estimator()
{
}

//**********************************************************************************************************************************


void Tdensity_estimator::clear()
{
	info_level = INFO_1;
	
	dim = 0;
	true_size = 0;
	working_set_size = 0;
	
	dataset_info_computed = false;
	empty_density_estimate();
	
	working_set.clear();
	dataset.clear();
	
	delta = 0.0;
	
	histogram.clear();
	
	internal_domain_partitioner.clear();
	internal_neighborhood_lists.neighbors.clear();
	internal_neighborhood_lists.distances.clear();
	
	domain_partitioner = NULL;
	neighborhood_lists = NULL;
}


//**********************************************************************************************************************************


void Tdensity_estimator::set_info_level(unsigned info_level)
{
	Tdensity_estimator::info_level = info_level;
}

//**********************************************************************************************************************************


void Tdensity_estimator::load(const Tdataset& dataset, const Tdataset& random_set, const Tdensity_control& density_control, bool set_working_set)
{
	clear();
	
	Tdensity_estimator::dataset = dataset;
	Tdensity_estimator::dataset.enforce_ownership();
	Tdensity_estimator::random_set = random_set;
	Tdensity_estimator::random_set.enforce_ownership();
	
	Tdensity_estimator::density_control = density_control;
	
	
	if (set_working_set == true)
	{
		working_set = dataset;
		working_set.enforce_ownership();
		working_set_size = dataset.size();
	}
		
	dim = dataset.dim();
	true_size = dataset.size();
	
	if (true_size == 0)
		flush_exit(ERROR_DATA_FALLS_OUTSIDE_SAFE_PARAMETERS, "Cannot load an empty data set into a density estimator.");
}



//**********************************************************************************************************************************


void Tdensity_estimator::internal_neighborhood_build(double max_distance, double& tau_neighbor_build_time, double& tau_neighbor_sort_time)
{
	unsigned i;
	unsigned j;
	unsigned metric;
	bool use_cover_tree;
	const unsigned cover_tree_threshold = 1024;
	Tworking_set_control working_set_control;

	
	tau_neighbor_build_time = get_wall_time_difference(tau_neighbor_build_time);
	
	internal_domain_partitioner.load(working_set);
	internal_neighborhood_lists.neighbors.clear();
	internal_neighborhood_lists.distances.clear();
	internal_neighborhood_lists.neighbors.resize(working_set_size);
	internal_neighborhood_lists.distances.resize(working_set_size);

	metric = density_control.get_metric_of_estimator();
	
	if (density_control.domain_partitioner_type == ADAPTIVE_PARTITIONER)
		use_cover_tree = (working_set_size > cover_tree_threshold); 
	else
		use_cover_tree = (density_control.domain_partitioner_type == COVER_TREE);
	
	if (use_cover_tree == true)
	{
		working_set_control.random_seed = 1;
		working_set_control.size_of_cells = 50;
		working_set_control.partition_method = VORONOI_TREE_BY_SIZE;
		
		internal_domain_partitioner.build_cover_tree(working_set_control, metric);
	}

	internal_domain_partitioner.build_neighborhood_lists(0, working_set_size, max_distance, metric, internal_neighborhood_lists);

	tau_neighbor_build_time = get_wall_time_difference(tau_neighbor_build_time);

	tau_neighbor_sort_time = get_wall_time_difference(tau_neighbor_sort_time);

	for (i=0; i<working_set_size; i++)
		if (internal_neighborhood_lists.neighbors[i].size() > 0)
		{
			sort_up(internal_neighborhood_lists.distances[i], internal_neighborhood_lists.neighbors[i]);

			if (use_cover_tree == true)
			{
				j = 0;
				while ((j+1 < internal_neighborhood_lists.neighbors[i].size()) and (internal_neighborhood_lists.neighbors[i][j] != i))
					j++;
				internal_neighborhood_lists.neighbors[i].erase(internal_neighborhood_lists.neighbors[i].begin() + j);
				internal_neighborhood_lists.distances[i].erase(internal_neighborhood_lists.distances[i].begin() + j);
			}
		}
	
	domain_partitioner = &internal_domain_partitioner;
	neighborhood_lists = &internal_neighborhood_lists;
	
	
	tau_neighbor_sort_time = get_wall_time_difference(tau_neighbor_sort_time);
}

//**********************************************************************************************************************************


void Tdensity_estimator::import_spatial_information(Tdomain_partitioner* domain_partitioner, Tneighborhood_lists* neighborhood_lists)
{
	if ((density_control.get_density_estimator_category() == DENSITY_GENERIC_KERNEL) or (density_control.get_density_estimator_category() == DENSITY_TRUTH))
	{
		Tdensity_estimator::domain_partitioner = domain_partitioner;
		Tdensity_estimator::neighborhood_lists = neighborhood_lists;
	}
}

//**********************************************************************************************************************************


double Tdensity_estimator::get_min_density() const
{
	return min_density;
}


//**********************************************************************************************************************************


double Tdensity_estimator::get_max_density() const
{
	return max_density;
}


//**********************************************************************************************************************************

vector <double> Tdensity_estimator::get_density_estimate(double delta, Tdensity_info& density_info)
{
	compute_density_estimate(delta, density_info);
	
	return density_estimate;
}


//**********************************************************************************************************************************


void Tdensity_estimator::compute_density_estimate(double delta, Tdensity_info& density_info)
{
	if (delta != Tdensity_estimator::delta)
	{
		Tdensity_estimator::delta = delta;
		switch(density_control.get_density_estimator_category())
		{
			case DENSITY_HISTOGRAMM:
				density_estimate_histogram(density_info.density_compute_time, density_info.histogram_build_time);
				internal_neighborhood_build(density_control.max_delta_neighborhood, density_info.tau_neighbor_build_time, density_info.tau_neighbor_sort_time);
				break;
				
			case DENSITY_GENERIC_KERNEL:
				density_estimate_kernel(neighborhood_lists->distances, density_info.density_compute_time);
				break;
				
			case DENSITY_TRUTH:
				density_estimate_true(density_info.density_compute_time);
				break;
		}
	}
}




//**********************************************************************************************************************************


Texcess_mass Tdensity_estimator::get_excess_mass(unsigned start_index, unsigned stop_index, Tdensity_info& density_info)
{
	Texcess_mass excess_mass;
	
	
	switch(density_control.get_density_estimator_category())
	{
		case DENSITY_HISTOGRAMM:
			excess_mass = excess_mass_histogram(start_index, stop_index, density_info);
			break;
			
		case DENSITY_GENERIC_KERNEL:
			excess_mass = excess_mass_kernel(start_index, stop_index, density_info);
			break;
			
		case DENSITY_TRUTH:
			excess_mass = excess_mass_true(start_index, stop_index, density_info);
			break;
	}
	return excess_mass;
}


//**********************************************************************************************************************************


double Tdensity_estimator::get_scale_factor(double delta, unsigned size) const
{
	double factor;

	if (size > 0)
		factor = 1.0 / (double(size) * pow(delta, double(dim)));
	else 
		factor = 0.0;
	
	factor = factor / density_control.get_estimator_integral(dim);

	return factor;
}


 
//**********************************************************************************************************************************


void Tdensity_estimator::density_estimate_histogram(double& density_time, double& histogram_time)
{
	unsigned i;
	double scale_factor;
	vector <double> density_tmp;
	

	if (true_size == 0)
	{
		empty_density_estimate();
		return;
	}
	
// Prepare and load histogram with the original dataset
	
	histogram_time = get_wall_time_difference(histogram_time);
	if (dataset_info_computed == false)
	{
		dataset_info = Tdataset_info(dataset, false);
		dataset_info_computed = true;
	}
	histogram.reserve(dataset_info, 2.0 * delta);
	histogram_time = get_wall_time_difference(histogram_time);
	histogram.assign(dataset, histogram_time);
	
	
// Get mid_point dataset of histogram with labels being counts of cell hits
	
	density_time = get_wall_time_difference(density_time);
	histogram.get_mid_point_dataset(working_set);
	
	flush_info(info_level,"\n\nDataset of size %d filled %d cells out of %1.4g possible cells.", dataset.size(), working_set.size(), histogram.get_number_of_cells());

	
// Compute density estimate on the mid_points
	
	reset_density();
	scale_factor = get_scale_factor(delta, true_size);
	for (i=0; i<working_set_size; i++)
	{
		density_estimate[i] = scale_factor * working_set.sample(i)->label;
		max_density = max(max_density, density_estimate[i]);
		min_density = min(min_density, density_estimate[i]);
	}
	
	density_time = get_wall_time_difference(density_time);
}
 
 
 
//**********************************************************************************************************************************


void Tdensity_estimator::density_estimate_kernel(const vector <vector <double> >& lists_of_distances, double& density_time)
{
	unsigned i;
	unsigned j;
	double sum;
	double scale_factor;
	double delta_inverse;
	
	if (true_size == 0)
	{
		empty_density_estimate();
		return;
	}
	
	density_time = get_wall_time_difference(density_time);
	reset_density();
	scale_factor = get_scale_factor(delta, working_set_size);
	
	if (working_set_size != lists_of_distances.size())
		flush_exit(ERROR_DATA_MISMATCH, "Kernel density estimator expects a list of neighborhood distances of size %d but received on of size %d.", working_set_size, lists_of_distances.size());
	
	
// This loop assumes that the list of distances are ordered in an ascending way
// so that one can stop the summation as soon as a distance is greater than delta.
	
	delta_inverse = 1.0 / delta;
	for (i=0; i<working_set_size; i++)
	{
		j = 0;
		sum = density_control.kernel_function(0.0);
		if (lists_of_distances[i].size() > 0)
			while (lists_of_distances[i][j] <= delta)
			{
				if (j + 1 == lists_of_distances[i].size())
					break;
				j++;
				sum = sum + density_control.kernel_function(lists_of_distances[i][j] * delta_inverse);
			}
		density_estimate[i] = scale_factor * sum;
		max_density = max(max_density, density_estimate[i]);
		min_density = min(min_density, density_estimate[i]);
	}

	density_time = get_wall_time_difference(density_time);
}
 
 
 
//**********************************************************************************************************************************


void Tdensity_estimator::density_estimate_true(double& density_time)
{
	unsigned i;
	double scale_factor;
	double weight_sum;
	
	
	if (true_size == 0)
	{
		empty_density_estimate();
		return;
	}
	
	density_time = get_wall_time_difference(density_time);
	
	
// First we need to compute the sum of the weights to find the normalizing
// scale_factor. Note that we reset the density here to make sure that we 
// have the right working set size. 
	
	reset_density();
	weight_sum = 0.0;
	for (i=0; i<working_set_size; i++)
		weight_sum = weight_sum + working_set.sample(i)->weight;
	scale_factor = 1.0 / weight_sum;

	
// Now we can use the normalized weights as density values.
	
	for (i=0; i<working_set_size; i++)
	{
		density_estimate[i] = scale_factor * working_set.sample(i)->weight;
		max_density = max(max_density, density_estimate[i]);
		min_density = min(min_density, density_estimate[i]);
	}
	
	density_time = get_wall_time_difference(density_time);
} 
 
 
 
//**********************************************************************************************************************************


Texcess_mass Tdensity_estimator::excess_mass_histogram(unsigned start_index, unsigned stop_index, Tdensity_info& density_info)
{
	unsigned i;
	unsigned start_index_rescaled;
	unsigned stop_index_rescaled;
	Texcess_mass excess_mass;
	

	compute_density_estimate(delta, density_info);
	
	start_index_rescaled = (start_index * level_set.size())/test_set.size();
	stop_index_rescaled = (stop_index * level_set.size())/test_set.size();

	excess_mass.density_integral = 0.0;
	excess_mass.constant_integral = double(stop_index_rescaled - start_index_rescaled);

	for (i=start_index_rescaled; i<stop_index_rescaled; i++)
		excess_mass.density_integral = excess_mass.density_integral +
		density_estimate[level_set[i]];

	excess_mass.density_integral = excess_mass.density_integral * pow(2.0 * delta, double(dim));
	excess_mass.constant_integral = excess_mass.constant_integral * pow(2.0 * delta, double(dim));
	
	return excess_mass;
}


//**********************************************************************************************************************************


void Tdensity_estimator::prepare_test_set(unsigned test_set_size, const Tsubset_info& level_set_core_samples, Tdensity_info& density_info)
{
	test_set.clear();
	test_set.resize(test_set_size);

	level_set = get_unique_entries(level_set_core_samples);
}


//**********************************************************************************************************************************


void Tdensity_estimator::construct_test_set(unsigned start_index, unsigned stop_index, double delta, const Tsubset_info& random_level_subset, Tdensity_info& density_info)
{
	unsigned i;
	Tsample random_sample;


	for (i=start_index; i<stop_index; i++)
	{
		random_sample = delta * Tsample(random_set.sample(i)); 
		random_sample = random_sample + Tsample(dataset.sample(level_set[random_level_subset[i]]));
		
		test_set[i] = random_sample;
	}
}



//**********************************************************************************************************************************


Texcess_mass Tdensity_estimator::excess_mass_kernel(unsigned start_index, unsigned stop_index, Tdensity_info& density_info)
{
	unsigned j;
	unsigned k;
	unsigned metric;
	double delta_inverse;
	Tsubset_info subset_info;
	vector <double> distances;
	vector <double> mask;
	Texcess_mass excess_mass;
	double test_set_level_set_ratio;
	double inverse_subset_size;
	double local_sum;
	

	excess_mass.density_integral = 0.0;
	excess_mass.constant_integral = 0.0;
	delta_inverse = 1.0 / delta;
	
	metric = density_control.get_metric_of_estimator();
	mask.assign(working_set_size, TRUE_MASK_ENTRY);
	test_set_level_set_ratio = double(test_set.size()) / double(level_set.size());
	
	
	for (j=start_index; j<stop_index; j++)
	{
		domain_partitioner->get_tau_neighborhood(&test_set[j], delta, subset_info, distances, metric, mask);
		
		if (subset_info.size() > 0)
		{
			inverse_subset_size = 1.0 / double(subset_info.size());
			excess_mass.constant_integral = excess_mass.constant_integral + inverse_subset_size;

			local_sum = 0.0;
			for (k=0; k<subset_info.size(); k++)
				local_sum = local_sum + density_control.kernel_function(distances[k] * delta_inverse);
			excess_mass.density_integral = excess_mass.density_integral + local_sum * inverse_subset_size;
		}
	}

	excess_mass.density_integral = excess_mass.density_integral / test_set_level_set_ratio;
	excess_mass.density_integral = excess_mass.density_integral * get_scale_factor(delta, working_set_size); 
	excess_mass.density_integral = excess_mass.density_integral * (pow(delta, dim) * density_control.unit_ball_volume(dim));
	
	excess_mass.constant_integral = excess_mass.constant_integral / test_set_level_set_ratio;
	excess_mass.constant_integral = excess_mass.constant_integral * (pow(delta, dim) * density_control.unit_ball_volume(dim));

	return excess_mass;
}


//**********************************************************************************************************************************


Texcess_mass Tdensity_estimator::excess_mass_true(unsigned start_index, unsigned stop_index, Tdensity_info& density_info)
{
	Texcess_mass excess_mass;
	
	
	return excess_mass;
}
 
 
//**********************************************************************************************************************************


void Tdensity_estimator::empty_density_estimate()
{
	if (true_size == 0)
	{
		max_density = 0.0;
		min_density = 0.0;
		
		working_set.clear();
		density_estimate.clear();
	}
	else
		flush_exit(ERROR_DATA_MISMATCH, "Cannot create empty density estimate from a dataset of size %d.", true_size);
}


//**********************************************************************************************************************************


void Tdensity_estimator::reset_density()
{
	max_density = 0.0;
	min_density = std::numeric_limits<double>::max();
	
	working_set_size = working_set.size();
	density_estimate.resize(working_set_size);
}



//**********************************************************************************************************************************


vector <unsigned> Tdensity_estimator::reassign_core_samples(const Tsubset_info& core_samples, vector <double>& core_sample_affiliation, double active_affiliation_number, double new_affiliation_number, const vector <double> cluster_affiliation)
{
	unsigned i;
	unsigned j;
	vector <unsigned> new_core_samples;
	unsigned samples_with_active_number_in_cell;
	list <unsigned> samples_in_cell;
	std::list <unsigned>::iterator it;
	
	
	if (density_control.density_estimator_type == DENSITY_HISTOGRAMM)
	{		
		for (i=0; i<core_samples.size(); i++)
		{
			j = core_samples[i];
			samples_in_cell = histogram.get_indices_in_cell(working_set.sample(j));
			samples_with_active_number_in_cell = 0;
			for (it = samples_in_cell.begin(); it != samples_in_cell.end(); ++it)
				if (cluster_affiliation[*it] == active_affiliation_number)
					samples_with_active_number_in_cell++;

			if (2 * samples_with_active_number_in_cell > samples_in_cell.size())
			{
				new_core_samples.push_back(j);
				core_sample_affiliation[j] = new_affiliation_number;
			}
		}
	}
	else
		for (i=0; i<core_samples.size(); i++)
		{
			j = core_samples[i];
			if (cluster_affiliation[j] == active_affiliation_number)
			{
				new_core_samples.push_back(j);
				core_sample_affiliation[j] = new_affiliation_number;
			}
		}
	
	return new_core_samples;
}


//**********************************************************************************************************************************


bool Tdensity_estimator::is_valid_neighbor(double own_affiliation, double neighbor_affiliation) const
{
	bool valid_neighbor;
	
	
	if (own_affiliation < 0.0)
		valid_neighbor = (neighbor_affiliation > 0.0);
	else
		valid_neighbor = (neighbor_affiliation == own_affiliation);
	
	return valid_neighbor;
}

//**********************************************************************************************************************************

vector <double> Tdensity_estimator::expand_from_core_to_train_samples(double sigma, double active_affiliation_number, double new_affiliation_number, Tdensity_info& density_info, const vector <double>& core_samples_affiliation, const vector <double>& current_affiliation)
{
	bool assign;
	unsigned i;
	unsigned j;
	unsigned added_samples;
	double neighbor_affiliation;
	vector <double> final_affiliation;

	
	switch(density_control.get_density_estimator_category())
	{
		case DENSITY_HISTOGRAMM:
			final_affiliation.resize(dataset.size());
			expand_from_core_histogram(0, dataset.size(), sigma, new_affiliation_number, final_affiliation, dataset, core_samples_affiliation, current_affiliation);
			break;
		
		case DENSITY_GENERIC_KERNEL:
			final_affiliation = current_affiliation;
			
			for (i=0; i<true_size; i++)
			{
				if (is_valid_neighbor(new_affiliation_number, core_samples_affiliation[i]) == true)
					final_affiliation[i] = core_samples_affiliation[i];
				else
				{
					assign = (neighborhood_lists->neighbors[i].size() > 0);
					if (new_affiliation_number >= 0.0)
						assign = assign and (current_affiliation[i] == active_affiliation_number);
					
					if (assign == true)
					{
						j = 0;
						while (neighborhood_lists->distances[i][j] <= sigma)
						{
							neighbor_affiliation = core_samples_affiliation[neighborhood_lists->neighbors[i][j]];
							if (is_valid_neighbor(new_affiliation_number, neighbor_affiliation) == true)
							{
								final_affiliation[i] = neighbor_affiliation;
								break;
							}
							j++;
							if (j == neighborhood_lists->distances[i].size())
								break;
						};
					}
				}
			}
			break;
			
		case DENSITY_TRUTH:
			final_affiliation = core_samples_affiliation;
			break;
	}

	if (new_affiliation_number >= 0.0)
	{
		added_samples = count_differences(final_affiliation, current_affiliation);
		flush_info(INFO_2, "\nAdded %d non-core samples with number %1.0f.", added_samples, new_affiliation_number);
	}
	

	return final_affiliation;
}


//**********************************************************************************************************************************

void Tdensity_estimator::expand_from_core_to_test_samples(unsigned start_index, unsigned stop_index, double sigma, double active_affiliation_number, double new_affiliation_number, vector <double>& new_affiliation, const Tdataset& test_set, const vector <double>& current_affiliation, const vector <unsigned>& current_core_samples, const vector <double>& core_samples_affiliation)
{
	switch(density_control.get_density_estimator_category())
	{
		case DENSITY_HISTOGRAMM:
			expand_from_core_histogram(start_index, stop_index, sigma, new_affiliation_number, new_affiliation, test_set, core_samples_affiliation, current_affiliation);
			break;
	
		case DENSITY_GENERIC_KERNEL:
			expand_from_core_to_test_kernel(start_index, stop_index, sigma, active_affiliation_number, new_affiliation_number, new_affiliation, test_set, current_affiliation, current_core_samples, core_samples_affiliation);
			break;
			
		case DENSITY_TRUTH:
			expand_from_core_to_test_kernel(start_index, stop_index, sigma, active_affiliation_number, new_affiliation_number, new_affiliation, test_set, current_affiliation, current_core_samples, core_samples_affiliation);
			break;
	}
}


//**********************************************************************************************************************************

void Tdensity_estimator::expand_from_core_to_test_kernel(unsigned start_index, unsigned stop_index, double sigma, double active_affiliation_number, double new_affiliation_number, vector <double>& new_affiliation, const Tdataset& test_set, const vector <double>& current_affiliation, const vector <unsigned>& current_core_samples, const vector <double>& core_samples_affiliation)
{
	unsigned i;
	unsigned j;
	unsigned jj;
	unsigned metric;
	bool use_cover_tree;
	double distance;
	double sigma_distance_compare;
	vector <double> neighbor_distances;
	vector <unsigned> neighbor_indices;
	vector <double> mask;
	
	
	if (new_affiliation.size() != current_affiliation.size())
		flush_exit(ERROR_DATA_MISMATCH, "Size %d of new affiliation does not mtch the size %d of current affiliation.", new_affiliation.size(),  current_affiliation.size());

	metric = density_control.get_metric_of_estimator();
	use_cover_tree = domain_partitioner->is_cover_tree_built();
	
	if (use_cover_tree == true)
	{
		// Do some preparations ensuring that only core samples are considered 
		// in the correct neighborhood.
		
		mask.assign(working_set_size, FALSE_MASK_ENTRY);
		for (j=0; j<current_core_samples.size(); j++)
			mask[current_core_samples[j]] = TRUE_MASK_ENTRY;
		
		if (density_control.get_density_estimator_category() == DENSITY_TRUTH)
			sigma_distance_compare = 0.5 * sigma;
		else
			sigma_distance_compare = sigma;
		
		
		// Now, look in the neighborhood and make an assignment if necessary 
		
		for (i=start_index; i<stop_index; i++)
			if (current_affiliation[i] == active_affiliation_number)
			{
				domain_partitioner->get_tau_neighborhood(test_set.sample(i), sigma_distance_compare, neighbor_indices, neighbor_distances, metric, mask);
					
				if (neighbor_indices.size() > 0)
					new_affiliation[i] = max(current_affiliation[i], new_affiliation_number);
				else
					new_affiliation[i] = current_affiliation[i];
			}
			else
				new_affiliation[i] = current_affiliation[i];
	}
	else
	{
		// Do some preparations ensuring that the correct neighborhood is considered.
		
		if (density_control.get_density_estimator_category() == DENSITY_TRUTH)
			sigma_distance_compare = 0.5 * sigma;
		else if (metric == EUCLIDEAN)
			sigma_distance_compare = sigma * sigma;
		else
			sigma_distance_compare = sigma;
		
		
		// Now, look in the neighborhood and make an assignment if necessary 
		
		for (i=start_index; i<stop_index; i++)
		{
			new_affiliation[i] = current_affiliation[i];
			if (current_affiliation[i] == active_affiliation_number)
				for (j=0; j<current_core_samples.size(); j++)
				{
					jj = current_core_samples[j];
					if (metric == EUCLIDEAN)
						distance = squared_distance(test_set.sample(i), working_set.sample(jj));
					else
						distance = sup_distance(test_set.sample(i), working_set.sample(jj));
					
					if (distance <= sigma_distance_compare)
					{
						new_affiliation[i] = max(current_affiliation[i], new_affiliation_number);
						break;
					}
				}
		}
	}
}


//**********************************************************************************************************************************

void Tdensity_estimator::expand_from_core_histogram(unsigned start_index, unsigned stop_index, double sigma, double new_affiliation_number, vector <double>& new_affiliation, const Tdataset& sample_set, const vector <double>& core_samples_affiliation, const vector <double>& current_affiliation)
{
	unsigned i;
	vector <double> tmp_affiliation;
	

	if (new_affiliation.size() != current_affiliation.size())
		flush_exit(ERROR_DATA_MISMATCH, "Size %d of new affiliation does not mtch the size %d of current affiliation.", new_affiliation.size(),  current_affiliation.size());
	
	for (i=0; i<working_set_size; i++)
		if (core_samples_affiliation[i] == new_affiliation_number)
			histogram.set_label_sum_in_cell(*working_set.sample(i), new_affiliation_number); 
		else if (new_affiliation_number < 0.0)
			histogram.set_label_sum_in_cell(*working_set.sample(i), core_samples_affiliation[i]); 
		else
			histogram.set_label_sum_in_cell(*working_set.sample(i), 0.0); 
		
	tmp_affiliation = histogram.predict_labels(sample_set, LABEL_SUM_HISTO);
	
	for (i=start_index; i<stop_index; i++)
		new_affiliation[i] = max(tmp_affiliation[i], current_affiliation[i]);
}

#endif
