// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (DOMIAIN_PARTITIONER_CPP) 
 #define DOMIAIN_PARTITIONER_CPP


#include "sources/cluster/solver/partitioner.h" 


#include "sources/shared/basic_functions/math_misc.h"

#include "sources/shared/basic_types/vector.h"
#include "sources/shared/basic_functions/random_subsets.h"




//**********************************************************************************************************************************


void Tcover_cell::clear()
{
	center_sample.clear();
	center_index = 0;
	member_indices.clear();
	
	radius = -1.0;
	cell_number = -1;
}


//**********************************************************************************************************************************


void Tcover_cell::copy(const Tcover_cell& source_cell)
{
	center_sample = source_cell.center_sample;
	center_index = source_cell.center_index;
	member_indices = source_cell.member_indices;
	
	radius = source_cell.radius;
	cell_number = source_cell.cell_number;
}


//**********************************************************************************************************************************


Tdomain_partitioner::Tdomain_partitioner()
{
	clear();
}


//**********************************************************************************************************************************



Tdomain_partitioner::~Tdomain_partitioner()
{
	
}



//**********************************************************************************************************************************

void Tdomain_partitioner::clear()
{
	dataset.clear();
	clear_tree();
}

//**********************************************************************************************************************************

void Tdomain_partitioner::clear_tree()
{
	metric = METRICS_MAX;
	cover_tree_built = false;
	cover_tree.clear();
}


//**********************************************************************************************************************************

void Tdomain_partitioner::load(const Tdataset& dataset)
{
	clear();
	
	Tdomain_partitioner::dataset = dataset;
	Tdomain_partitioner::dataset.enforce_ownership();
}

//**********************************************************************************************************************************

bool Tdomain_partitioner::is_cover_tree_built() const
{
	return cover_tree_built;
}


//**********************************************************************************************************************************

void Tdomain_partitioner::build_cover_tree(const Tworking_set_control& working_set_control, unsigned metric)
{
	Tsubset_info subset_info;
	
	if (dataset.size() == 0)
		flush_exit( ERROR_DATA_MISMATCH, "Cannot build cover tree for empty data set.");

	clear_tree();
	
	Tdomain_partitioner::partitioner_type = partitioner_type;
	Tdomain_partitioner::working_set_control = working_set_control;
	
	subset_info = id_permutation(dataset.size());
	build_cover_tree_recursively(dataset, subset_info, &cover_tree.root_node, 1, working_set_control.size_of_cells, metric);
	
	cover_tree_built = true;
	Tdomain_partitioner::metric = metric;
}



//**********************************************************************************************************************************

void Tdomain_partitioner::build_cover_tree_recursively(const Tdataset& current_working_set, Tsubset_info current_subset_info, Tcover_tree_node* current_tree_node, unsigned current_depth, unsigned NNs, unsigned metric)
{
	unsigned c;
	unsigned i;
	unsigned current_target_cell_size;
	vector <unsigned> new_center_indices;
	Tdataset new_center_dataset;
	vector <double> new_radii;
	vector <unsigned> cell_affiliation;
	Tdataset new_working_subset;
	Tsubset_info new_subset_info;

	
// If the current working set is small enoug, it suffices to store the current subset info
// into the cells member_indices.
	
	if (current_working_set.size() <= NNs)
	{
		current_tree_node->member_indices = current_subset_info;
		
		flush_info(INFO_2, "\nCreating a cell with %d members at depth %d and radius %2.4f.", current_working_set.size(), current_depth, current_tree_node->radius);
		
		return;
	}
	
	
// Else, we first approximately determine the current target cell size
	
	if ((working_set_control.max_tree_depth == 0) or (current_depth < working_set_control.max_tree_depth))
		current_target_cell_size = double(current_working_set.size()) / working_set_control.tree_reduction_factor;
	else
		current_target_cell_size = NNs;
	
	if ((working_set_control.max_theoretical_node_width > 0) and (current_working_set.size() < working_set_control.max_theoretical_node_width * NNs))
		current_target_cell_size = NNs;

	flush_info(INFO_2, "\nSplitting a working set of size %d at a node of depth %d with target cell size %d.", current_working_set.size(), current_depth, current_target_cell_size);

	
// Determine the indices and squared radii of a suitable covering net and adjust the squared radii


	new_center_indices = current_working_set.create_cover_subset_info_by_kNN(current_target_cell_size, working_set_control.random_seed, false, new_radii, false, working_set_control.size_of_dataset_to_find_partition, metric);


	if (metric == EUCLIDEAN)
		for (i=0; i<new_radii.size(); i++)
			new_radii[i] = sqrt(new_radii[i]);
	
	
// Create the corresponding set of covering samples
	
	current_working_set.create_subset(new_center_dataset, new_center_indices, true);
	

// Determine the new cell each sample of the current working set falls into
	
	cell_affiliation.resize(current_working_set.size());
	for (i=0; i<current_working_set.size(); i++)
		cell_affiliation[i] = new_center_dataset.get_index_of_closest_sample(current_working_set.sample(i));


// Translate the indices of the current working set into indices of the previous working set.
	
	for (c=0; c<new_center_indices.size(); c++)
		new_center_indices[c] = current_subset_info[new_center_indices[c]]; 
	
	
// Finally, we need to build the new cover
	
	current_tree_node->child_nodes.resize(new_center_indices.size());
	for (c=0; c<new_center_indices.size(); c++)
	{
// 	Create new working set for the c-th cell

		new_working_subset.clear();
		new_subset_info.clear();
		for (i=0; i<cell_affiliation.size(); i++)
			if (cell_affiliation[i] == c)
			{
				new_working_subset.push_back(current_working_set.sample(i));
				new_subset_info.push_back(current_subset_info[i]);
			}
		
		
// 	Create new cover for this cell
			
		current_tree_node->child_nodes[c] = new Tcover_tree_node;
		
		current_tree_node->child_nodes[c]->center_index = new_center_indices[c];
		current_tree_node->child_nodes[c]->center_sample = *new_center_dataset.sample(c);
		current_tree_node->child_nodes[c]->radius = new_radii[c];
		
		build_cover_tree_recursively(new_working_subset, new_subset_info, current_tree_node->child_nodes[c], current_depth + 1, NNs, metric);
	}
}


//**********************************************************************************************************************************

void Tdomain_partitioner::get_tau_neighborhood(Tsample* sample, double tau, Tsubset_info& subset_info, vector <double>& distances, unsigned metric, const vector <double>& mask) const
{
	unsigned i;
	unsigned ii;
	unsigned c;
	unsigned leafs_visited;
	unsigned distance_computations;
	double distance_to_member;
	double distance_to_center;
	double critical_distance;
	double tau_squared;
	Tcover_tree_node copy_of_root_node;
	Tcover_tree_node* current_node;
	stack <Tcover_tree_node* > stack_of_nodes;
	
	
	if (tau <= 0.0)
		flush_exit(ERROR_DATA_MISMATCH, "Cannot build neighborhood list for tau = %2.4f.", tau);
	
	if (metric == EUCLIDEAN)
		tau_squared = tau * tau;
	else
		tau_squared = tau;
		
	leafs_visited = 0;
	distance_computations = 0;
	
	subset_info.clear();
	distances.clear();
	
	copy_of_root_node = cover_tree.root_node;
	stack_of_nodes.push(&copy_of_root_node);

	while (stack_of_nodes.empty() == false)
	{
		current_node = stack_of_nodes.top();
		stack_of_nodes.pop();
		
		if (current_node->child_nodes.size() > 0)
			for (c=0; c<current_node->child_nodes.size(); c++)
			{
				critical_distance = tau + current_node->child_nodes[c]->radius;
			
				if (metric == EUCLIDEAN)
					distance_to_center = sqrt(squared_distance(sample, &(current_node->child_nodes[c]->center_sample)));  
				else
					distance_to_center = 0.5 * sup_distance(sample, &(current_node->child_nodes[c]->center_sample));
				
				if (distance_to_center <= critical_distance)
						stack_of_nodes.push(current_node->child_nodes[c]);
				
				distance_computations++;
			}
		else
		{
			leafs_visited++;
			for (i=0; i<current_node->member_indices.size(); i++)
			{
				ii = current_node->member_indices[i];
				if (mask[ii] == TRUE_MASK_ENTRY)
				{
					if (metric == EUCLIDEAN)
						distance_to_member = squared_distance(sample, dataset.sample(ii));
					else
						distance_to_member = sup_distance(sample, dataset.sample(ii));
					if (distance_to_member <= tau_squared)
					{
						subset_info.push_back(ii);
						if (metric == EUCLIDEAN)
							distances.push_back(sqrt(distance_to_member));
						else
							distances.push_back(distance_to_member);
					}
						
					distance_computations++; 
				}
			}
		}
	}
}



//**********************************************************************************************************************************

void Tdomain_partitioner::build_neighborhood_lists(unsigned start_index, unsigned stop_index, double tau, unsigned metric, Tneighborhood_lists& neighborhood_lists)
{
	unsigned i;
	unsigned j;
	unsigned size;
	double distance;
	vector <double> mask;
	
	
// Perform various sanity checks before the actual computation
	
	size = dataset.size();
	
	if (stop_index > size)
		flush_exit(ERROR_DATA_MISMATCH, "Cannot build neighborhood list for %d samples since the dataset only conists of %d samples.", stop_index, size);
	
	if (start_index > stop_index)
		flush_exit(ERROR_DATA_MISMATCH, "Cannot build neighborhood list for %d samples between %d and %d", start_index, stop_index);
	
	if ((neighborhood_lists.neighbors.size() != size) or (neighborhood_lists.distances.size() != size))
	{
		if ((start_index == 0) and (stop_index == size))
		{
			neighborhood_lists.neighbors.resize(size);
			neighborhood_lists.distances.resize(size);
		}
		else
			flush_exit(ERROR_DATA_MISMATCH, "Computing neighborhood lists from %d to %d requires a neighborhood list of size %d.", start_index, stop_index, size);
	}
	
	if (((start_index != 0) or (stop_index != size)) and (cover_tree_built == false))
		flush_exit(ERROR_DATA_MISMATCH, "Computing neighborhood lists from %d to %d requires a built cover tree.", start_index, stop_index);
	

// Now, we can start with the actual computation 

	if (cover_tree_built == true)
	{
		if (Tdomain_partitioner::metric != metric)
			flush_exit(ERROR_DATA_MISMATCH, "Metric %d of cover tree does not fit to metric %d of neighborhood.", Tdomain_partitioner::metric, metric);
		
		mask.assign(size, TRUE_MASK_ENTRY);
		for (i=start_index; i<stop_index; i++)
			get_tau_neighborhood(dataset.sample(i), tau, neighborhood_lists.neighbors[i], neighborhood_lists.distances[i], metric, mask);
	}
	else
	{
		for (i=0; i<size; i++)
		{
			neighborhood_lists.neighbors[i].clear();
			neighborhood_lists.distances[i].clear();
			neighborhood_lists.neighbors[i].reserve(1024);
			neighborhood_lists.distances[i].reserve(1024);
		}
		for (i=0; i<size; i++)
			for (j=0; j<i; j++)
			{
				if (metric == EUCLIDEAN)
					distance = sqrt(squared_distance(dataset.sample(i), dataset.sample(j)));
				else
					distance = sup_distance(dataset.sample(i), dataset.sample(j));
				
				if (distance <= tau)
				{
					neighborhood_lists.neighbors[i].push_back(j);
					neighborhood_lists.distances[i].push_back(distance);
					
					neighborhood_lists.neighbors[j].push_back(i);
					neighborhood_lists.distances[j].push_back(distance);
				}
			}
	}
}


//**********************************************************************************************************************************





#endif
