// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.






//**********************************************************************************************************************************


#include "sources/shared/basic_functions/flush_print.h"




//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>::Ttree_node()
{
	depth = 0;
	parent_node = NULL;
	
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nCreating empty node at %p.", this);
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>::Ttree_node(const Ttree_node<Tcontainer_type>& tree_node)
{
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nCreating copied node at %p.", this);
	copy(tree_node);
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>::~Ttree_node()
{
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nDestroying node at %p.", this);
}




//**********************************************************************************************************************************

template <class Tcontainer_type> void Ttree_node<Tcontainer_type>::clear()
{
	if (child_nodes.size() > 0)
		flush_exit(ERROR_DATA_MISMATCH, "Cannot clear node at %p since it has %d children.", this, child_nodes.size());
	
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nClearing node at %p.", this);

	Tcontainer_type::clear();
	child_nodes.clear();
	depth = 0;
	parent_node = NULL;
}



//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>& Ttree_node<Tcontainer_type>::operator = (const Ttree_node<Tcontainer_type>& tree_node)
{
	copy(tree_node);
	return *this;
}



//**********************************************************************************************************************************

template <class Tcontainer_type> void Ttree_node<Tcontainer_type>::copy(const Ttree_node<Tcontainer_type>& source_node)
{
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nCopying node at %p and %d children into node at %p.", &source_node, source_node.child_nodes.size(), this);
	
	Tcontainer_type::copy(source_node);
	depth = source_node.depth;
	child_nodes = source_node.child_nodes;
	parent_node = source_node.parent_node;
}


//**********************************************************************************************************************************

template <class Tcontainer_type> void Ttree_node<Tcontainer_type>::display(unsigned info_level)
{
	flush_info(info_level, "\nNumber of children is %d.", child_nodes.size());
	flush_info(info_level, "\nRAM address is %p.", this);
	
	Tcontainer_type::display(info_level);
}



//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************






//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree <Tcontainer_type>::Ttree()
{
	
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree <Tcontainer_type>::Ttree(const Ttree <Tcontainer_type>& tree)
{
	copy(tree);
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree <Tcontainer_type>::~Ttree()
{
	clear();
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree <Tcontainer_type>& Ttree <Tcontainer_type>::operator = (const Ttree <Tcontainer_type>& tree)
{
	copy(tree);
	return *this;
}



//**********************************************************************************************************************************

template <class Tcontainer_type> void  Ttree <Tcontainer_type>::display( unsigned info_level)
{
	
}



//**********************************************************************************************************************************

template <class Tcontainer_type> void Ttree <Tcontainer_type>::clear()
{
	unsigned c;
	Ttree_node <Tcontainer_type>* current_node;

	
	stack_of_nodes.push(&root_node);
	while (stack_of_nodes.empty() == false)
	{
		current_node = stack_of_nodes.top();
		stack_of_nodes.pop();
		
		for (c=0; c<current_node->child_nodes.size(); c++)
			if (current_node->child_nodes[c] != NULL)
			{
				if (current_node->child_nodes[c]->child_nodes.size() == 0)
				{
					delete current_node->child_nodes[c];
					current_node->child_nodes[c] = NULL;
				}
				else
					stack_of_nodes.push(current_node->child_nodes[c]);
			}
			
		current_node->child_nodes.clear();
		current_node->clear();
	}
}



//**********************************************************************************************************************************

template <class Tcontainer_type> void Ttree <Tcontainer_type>::copy(const Ttree <Tcontainer_type>& source_tree)
{
	unsigned c;
	Ttree_node <Tcontainer_type>* current_node;
	Ttree_node <Tcontainer_type>* current_source_node;
	Ttree_node <Tcontainer_type> source_root_node;
	stack <Ttree_node <Tcontainer_type>* > stack_of_nodes;
	stack <Ttree_node <Tcontainer_type>* > stack_of_source_nodes;
	
	
	stack_of_nodes.push(&root_node);
	source_root_node = source_tree.root_node;
	stack_of_source_nodes.push(&source_root_node);
	
	while (stack_of_nodes.empty() == false)
	{
		current_node = stack_of_nodes.top();
		current_source_node = stack_of_source_nodes.top();
		stack_of_nodes.pop();
		stack_of_source_nodes.pop();
		
		current_node->child_nodes.resize(current_source_node->child_nodes.size());
		for (c=0; c<current_source_node->child_nodes.size(); c++)
			if (current_source_node->child_nodes[c] != NULL)
			{
				current_node->child_nodes[c] = new Ttree_node <Tcontainer_type>;
				stack_of_nodes.push(current_node->child_nodes[c]);
				stack_of_source_nodes.push(current_source_node->child_nodes[c]);
			}
			else
				current_node->child_nodes[c] = NULL;
	}
}



//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>* Ttree <Tcontainer_type>::start()
{
	unsigned c;
	Ttree_node <Tcontainer_type>* current_node;

	
	while (stack_of_nodes.empty() == false)
		stack_of_nodes.pop();
	
	while (queue_of_nodes.empty() == false)
		queue_of_nodes.pop();
		
	stack_of_nodes.push(&root_node);
	while (stack_of_nodes.empty() == false)
	{
		current_node = stack_of_nodes.top();
		stack_of_nodes.pop();
		queue_of_nodes.push(current_node);
		
		for (c=0; c<current_node->child_nodes.size(); c++)
			if (current_node->child_nodes[c] != NULL)
				stack_of_nodes.push(current_node->child_nodes[c]);
	}

	if (queue_of_nodes.empty() == false)
		return queue_of_nodes.front();
	else
		return NULL; 
}



//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>* Ttree <Tcontainer_type>::next()
{
	Ttree_node <Tcontainer_type>* current_node;
	
	queue_of_nodes.pop();
	if (queue_of_nodes.empty() == true)
		return NULL;
	
	current_node = queue_of_nodes.front();
	
	
	return current_node;
}



//**********************************************************************************************************************************



template <class Tcontainer_type> unsigned Ttree <Tcontainer_type>::nodes()
{
	unsigned number_of_nodes;
	Ttree_node <Tcontainer_type>* current_node;
	
	
	number_of_nodes = 0;
	current_node = start();
		while (current_node != NULL)
		{
			number_of_nodes++;
			current_node = next();
		}
	
	return number_of_nodes;
}




//**********************************************************************************************************************************



template <class Tcontainer_type> unsigned Ttree <Tcontainer_type>::leafs()
{
	unsigned number_of_leafs;
	Ttree_node <Tcontainer_type>* current_node;
	
	
	number_of_leafs = 0;
	current_node = start();
		while (current_node != NULL)
		{
			if (current_node->child_nodes.size() == 0)
				number_of_leafs++;
			current_node = next();
		}
	
	return number_of_leafs;
}



//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************



template <class Tcontainer_type> Tforest<Tcontainer_type>::Tforest()
{
	delete_queue = true;
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Tforest<Tcontainer_type>::Tforest(const Tforest <Tcontainer_type>& forest)
{
	copy(forest);
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Tforest <Tcontainer_type>& Tforest <Tcontainer_type>::operator = (const Tforest <Tcontainer_type>& forest)
{
	copy(forest);
	return *this;
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Tforest<Tcontainer_type>::~Tforest()
{
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nDestroying forest with %d roots.", root_nodes.size());
	clear();
}




//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::clear()
{
	typename std::map < double, Ttree_node <Tcontainer_type> >::iterator it;
	Ttree_node<Tcontainer_type>* current_tree_node;

	
	delete_queue = true;
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nClearing forest with %d roots.", root_nodes.size());
	
	for (it = root_nodes.begin(); it != root_nodes.end(); ++it)
	{
		current_tree_node = &(it->second);
		flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nClearing tree root at %p.", current_tree_node);
		clear_recursively(current_tree_node);
	}

	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nDeleting %d roots.", root_nodes.size());
	root_nodes.clear();
}



//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::clear_recursively(Ttree_node<Tcontainer_type>* current_tree_node)
{
	unsigned c;
	
	delete_queue = true;
	for (c=0; c<current_tree_node->child_nodes.size(); c++)
		if (current_tree_node->child_nodes[c] != NULL)
			clear_recursively(current_tree_node->child_nodes[c]);
	current_tree_node->child_nodes.clear();
		
	flush_info(INFO_EXTREMELY_PEDANTIC_DEBUG, "\nGoing to clear leaf node at %p.", current_tree_node);
	current_tree_node->clear();
}



//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::new_root(double target_key)
{
	Ttree_node <Tcontainer_type>* new_node;
	
	if (root_exists(target_key) == true)
		flush_exit(ERROR_DATA_MISMATCH, "Root with key %1.4f does already exist.", target_key);
	
	new_node = new Ttree_node<Tcontainer_type>();
	root_nodes[target_key] = *new_node;
}


//**********************************************************************************************************************************

template <class Tcontainer_type> bool Tforest<Tcontainer_type>::root_exists(double key)
{
	typename std::map < double, Ttree_node <Tcontainer_type> >::iterator it;
	
	it = root_nodes.find(key);
	if (it != root_nodes.end())
		return true;
	else
		return false;
}


//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::check_key(double key)
{
	if (root_exists(key) == false)
		flush_exit(ERROR_DATA_MISMATCH, "Root with key %1.4f does not exist in tree.", key);
}


//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::merge_roots(double target_key, double source_key)
{
	Ttree_node <Tcontainer_type>* target_tree_node;
	Ttree_node <Tcontainer_type>* source_tree_node;
	
	
	check_key(target_key);
	check_key(source_key);
	
	source_tree_node = new Ttree_node<Tcontainer_type>(root_nodes[source_key]);
	target_tree_node = new Ttree_node<Tcontainer_type>(root_nodes[target_key]);


	target_tree_node->child_nodes.push_back(source_tree_node);
	target_tree_node->merge(*source_tree_node);
	
	root_nodes.erase(source_key);
	root_nodes.erase(target_key);
	root_nodes[target_key] = *target_tree_node;
}


//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::merge_roots(double target_key, double source_key_1, double source_key_2)
{
	Ttree_node <Tcontainer_type>* target_tree_node;
	Ttree_node <Tcontainer_type>* source_tree_node_1;
	Ttree_node <Tcontainer_type>* source_tree_node_2;
	
	
	check_key(source_key_1);
	check_key(source_key_2);
	
	source_tree_node_1 = new Ttree_node<Tcontainer_type>(root_nodes[source_key_1]);
	source_tree_node_2 = new Ttree_node<Tcontainer_type>(root_nodes[source_key_2]);
	target_tree_node = new Ttree_node<Tcontainer_type>;

	target_tree_node->child_nodes.resize(2);
	target_tree_node->child_nodes[0] = source_tree_node_1;
	target_tree_node->child_nodes[1] = source_tree_node_2;

	target_tree_node->merge(*source_tree_node_1, *source_tree_node_2);

	root_nodes.erase(source_key_1);
	root_nodes.erase(source_key_2);
	root_nodes.erase(target_key);
	root_nodes[target_key] = *target_tree_node;
}

//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::unite_roots(double target_key, double source_key_1, double source_key_2)
{
	unsigned c;
	Ttree_node <Tcontainer_type>* target_tree_node;
	Ttree_node <Tcontainer_type>* source_tree_node_1;
	Ttree_node <Tcontainer_type>* source_tree_node_2;
	
	
	check_key(source_key_1);
	check_key(source_key_2);
	
	source_tree_node_1 = new Ttree_node<Tcontainer_type>(root_nodes[source_key_1]);
	source_tree_node_2 = new Ttree_node<Tcontainer_type>(root_nodes[source_key_2]);
	target_tree_node = new Ttree_node<Tcontainer_type>;

	
	target_tree_node->child_nodes.reserve(1 + source_tree_node_2->child_nodes.size());
	
	target_tree_node->child_nodes.push_back(source_tree_node_1);
	target_tree_node->Tcontainer_type::copy(*source_tree_node_1);
	
	for (c=0; c<source_tree_node_2->child_nodes.size(); c++)
	{
		target_tree_node->child_nodes.push_back(source_tree_node_2->child_nodes[c]);
		target_tree_node->merge(*(source_tree_node_2->child_nodes[c]));
	}

	target_tree_node->merge(*source_tree_node_1, *source_tree_node_2);

	root_nodes.erase(source_key_1);
	root_nodes.erase(source_key_2);
	root_nodes.erase(target_key);
	root_nodes[target_key] = *target_tree_node;
}

//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::unite_roots(double target_key, double source_key)
{
	unsigned c;
	Ttree_node <Tcontainer_type>* target_tree_node;
	Ttree_node <Tcontainer_type>* source_tree_node;
	

	check_key(target_key);
	check_key(source_key);
	
	source_tree_node = new Ttree_node<Tcontainer_type>(root_nodes[source_key]);
	target_tree_node = new Ttree_node<Tcontainer_type>(root_nodes[target_key]);


	for (c=0; c<source_tree_node->child_nodes.size(); c++)
	{
		target_tree_node->child_nodes.push_back(source_tree_node->child_nodes[c]);
		target_tree_node->merge(*(source_tree_node->child_nodes[c]));
	}
	
	root_nodes.erase(source_key);
	root_nodes.erase(target_key);
	root_nodes[target_key] = *target_tree_node;
}



//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>* Tforest<Tcontainer_type>::start()
{
	typename  std::map < double, Ttree_node <Tcontainer_type> >::iterator it;
	
	
	while (queue_of_nodes.empty() == false)
		queue_of_nodes.pop();

	if (root_nodes.size() > 0)
	{
		delete_queue = false;
		for (it = root_nodes.begin(); it != root_nodes.end(); ++it)
			start(it->first);
		delete_queue = true;

		return queue_of_nodes.front();
	}
	else
		return NULL;
}


//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>* Tforest<Tcontainer_type>::start(double key)
{
	unsigned c;
	Ttree_node <Tcontainer_type>* current_node;

	
	
	while (stack_of_nodes.empty() == false)
		stack_of_nodes.pop();
	
	if (delete_queue == true)
		while (queue_of_nodes.empty() == false)
			queue_of_nodes.pop();
		
	if (root_exists(key) == true)
	{
		stack_of_nodes.push(&root_nodes[key]);
		while (stack_of_nodes.empty() == false)
		{
			current_node = stack_of_nodes.top();
			stack_of_nodes.pop();
			queue_of_nodes.push(current_node);
			
			for (c=0; c<current_node->child_nodes.size(); c++)
				if (current_node->child_nodes[c] != NULL)
					stack_of_nodes.push(current_node->child_nodes[c]);
		}

		return queue_of_nodes.front();
	}
	else
		return NULL;
}



//**********************************************************************************************************************************

template <class Tcontainer_type> Ttree_node<Tcontainer_type>* Tforest<Tcontainer_type>::next()
{
	Ttree_node <Tcontainer_type>* current_node;
	
	queue_of_nodes.pop();
	if (queue_of_nodes.empty() == true)
		return NULL;
	
	current_node = queue_of_nodes.front();
	
	
	return current_node;
}




//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::display(unsigned info_level)
{
	typename std::map < double, Ttree_node <Tcontainer_type> >::iterator it;
	
	
	flush_info(info_level, "\n\nForest contains %d trees.", root_nodes.size());
	for (it = root_nodes.begin(); it != root_nodes.end(); ++it)
	{
		flush_info(info_level, "\nRoot with key %2.4f has the following content:", it->first);
		it->second.display(info_level);
	}
}


//**********************************************************************************************************************************

template <class Tcontainer_type> void Tforest<Tcontainer_type>::copy(const Tforest <Tcontainer_type>& source_forest)
{
	unsigned c;
	double source_key;
	Ttree_node <Tcontainer_type>* current_node;
	Ttree_node <Tcontainer_type>* current_source_node;
	Ttree_node <Tcontainer_type>* root_node;
	Ttree_node <Tcontainer_type> source_root_node;
	Ttree_node <Tcontainer_type>* source_root_node_tmp;
	Ttree_node <Tcontainer_type> source_node_copy;
	
	stack <Ttree_node <Tcontainer_type>* > stack_of_nodes;
	stack <Ttree_node <Tcontainer_type>* > stack_of_source_nodes;
	typename std::map < double, Ttree_node <Tcontainer_type> >::const_iterator it;
	
	
	clear();

	for (it = source_forest.root_nodes.begin(); it != source_forest.root_nodes.end(); ++it)
	{
		source_key = it->first;
		source_root_node = it->second;
		
		source_root_node_tmp = new Ttree_node <Tcontainer_type>(source_root_node);
		root_node = new Ttree_node <Tcontainer_type>(source_root_node);

		stack_of_nodes.push(root_node);
		stack_of_source_nodes.push(source_root_node_tmp);

		while (stack_of_nodes.empty() == false)
		{
			current_node = stack_of_nodes.top();
			current_source_node = stack_of_source_nodes.top();

			stack_of_nodes.pop();
			stack_of_source_nodes.pop();

			current_node->child_nodes.resize(current_source_node->child_nodes.size());
			for (c=0; c<current_source_node->child_nodes.size(); c++)
				if (current_source_node->child_nodes[c] != NULL)
				{
					current_node->child_nodes[c] = new Ttree_node <Tcontainer_type>(*current_source_node->child_nodes[c]);

					stack_of_nodes.push(current_node->child_nodes[c]);
					stack_of_source_nodes.push(current_source_node->child_nodes[c]);
				}
				else
					current_node->child_nodes[c] = NULL;
		}
		root_nodes[source_key] = *root_node;
	}
}


