// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (CLUSTER_INFO_CPP) 
	#define CLUSTER_INFO_CPP



#include "sources/cluster/training_validation/cluster_info.h"





#include "sources/shared/basic_functions/flush_print.h"


//**********************************************************************************************************************************

Tcluster_info::Tcluster_info()
{
	clear();
}


//**********************************************************************************************************************************

Tcluster_info::Tcluster_info(const Tdensity_info& density_info)
{
	clear();
	Tdensity_info::copy(density_info);
}


//**********************************************************************************************************************************

Tcluster_info::Tcluster_info(const Tcluster_info& cluster_info)
{
	copy(cluster_info);
}


//**********************************************************************************************************************************

void Tcluster_info::clear()
{
	Tdensity_info::clear();
	
	tau = 0.0;
	eps = 0.0;
	rho_start = NO_RHO;
	rho_stop = NO_RHO;
	rho = NO_RHO;
	rho_star = NO_RHO;
	rho_step = 0.0;
	min_rho_step = 0.0;

	changed_level_set = false;
	number_of_clusters = 0;
	
	number_of_merges = 0;
	number_of_new_merges = 0;
	number_of_peaks = 0;
	number_of_new_peaks = 0;
	number_of_assigned_samples = 0;
	number_of_unassigned_samples = 0;
	number_of_newly_assigned_samples = 0;
	number_of_samples_with_new_cluster_number = 0;

	clear_time();
};



//**********************************************************************************************************************************

void Tcluster_info::clear_time()
{
	Tdensity_info::clear_time();
	
	cluster_connect_time = 0.0;
	level_set_compute_time = 0.0;
	cluster_admin_time = 0.0;
	cluster_search_time = 0.0;
	cluster_expand_time = 0.0;
	cluster_tree_build_time = 0.0;
}


//**********************************************************************************************************************************

double Tcluster_info::get_full_time() const
{
	double full_time;
	
	full_time = Tdensity_info::get_full_time();
	full_time = full_time + level_set_compute_time;
	full_time = full_time + cluster_search_time;
	full_time = full_time + cluster_admin_time;
	full_time = full_time + cluster_connect_time;
	full_time = full_time + cluster_expand_time;
	full_time = full_time + cluster_tree_build_time;
	
	return full_time;
}


//**********************************************************************************************************************************


string Tcluster_info::displaystring_train_time(unsigned display_mode) const
{
	string output;
	
	
	if (display_mode == TRAIN_INFO_DISPLAY_FORMAT_TIMINGS_FULL)
	{
		output = Tdensity_info::displaystring_train_time(display_mode);
		
		output = output + "\nCluster search time:         " +  number_to_string(cluster_search_time, 4, DISPLAY_FLOAT);
		output = output + "\nCluster connect time:        " + number_to_string(cluster_connect_time, 4, DISPLAY_FLOAT);
		output = output + "\nCluster expand time:         " + number_to_string(cluster_expand_time, 4, DISPLAY_FLOAT);
		output = output + "\nCluster admin time:          " + number_to_string(cluster_admin_time, 4, DISPLAY_FLOAT);
		output = output + "\nCluster tree build time:     " + number_to_string(cluster_tree_build_time, 4, DISPLAY_FLOAT);
	}
	
	return output;
} 


//**********************************************************************************************************************************


string Tcluster_info::displaystring_detailed(unsigned display_mode) const
{
	string output;
	
	if ((display_mode != TRAIN_INFO_DISPLAY_FORMAT_SUMMARIZED) and (display_mode != TRAIN_INFO_DISPLAY_FORMAT_TIMINGS_FULL))
	{
		output = output + "del: " + number_to_string(width, 3, DISPLAY_SCIENTIFIC);
		output = output + "  tau: " + number_to_string(tau, 3, DISPLAY_SCIENTIFIC);
		output = output + "  eps: " + number_to_string(eps, 3, DISPLAY_SCIENTIFIC);
		if (rho_start < 0.0)
			output = output + "  rbeg: " + number_to_string(rho_start, 2, DISPLAY_SCIENTIFIC);
		else
			output = output + "  rbeg: " + number_to_string(rho_start, 3, DISPLAY_SCIENTIFIC);
		if (rho < 0.0)
			output = output + "  rho: " + number_to_string(rho, 2, DISPLAY_SCIENTIFIC);
		else
			output = output + "  rho: " + number_to_string(rho, 3, DISPLAY_SCIENTIFIC);
		if (rho_star >= 0.0)
			output = output + "  rho*: " + number_to_string(rho_star, 3, DISPLAY_SCIENTIFIC);
		else
			output = output + "  rho*:       ---";
		output = output + "  clu: " + pos_number_to_string(number_of_clusters, 2);
		output = output + "  mer: " + pos_number_to_string(number_of_merges, 2);
		output = output + "  pea: " + pos_number_to_string(number_of_peaks, 2);
		output = output + "  nmer: " + pos_number_to_string(number_of_new_merges, 2);
		output = output + "  npea: " + pos_number_to_string(number_of_new_peaks, 2);
		output = output + "  as: " + pos_number_to_string(number_of_assigned_samples, 5);
		output = output + "  nas: " + pos_number_to_string(number_of_newly_assigned_samples, 5);
		output = output + "  snc: " + pos_number_to_string(number_of_samples_with_new_cluster_number, 5);
	}
	return output;
}



//**********************************************************************************************************************************

Tcluster_info Tcluster_info::operator + (const Tcluster_info& cluster_info)
{
	Tdensity_info result_base;
	Tcluster_info result;

	
	result_base = *this;
	result_base = result_base + cluster_info;
	result = Tcluster_info(result_base);
	
	result.copy_hyper_parameters(*this);
	result.copy_cluster_results(*this);

	result.level_set_compute_time = level_set_compute_time + cluster_info.level_set_compute_time;
	result.cluster_search_time = cluster_search_time + cluster_info.cluster_search_time;
	result.cluster_admin_time = cluster_admin_time + cluster_info.cluster_admin_time;
	result.cluster_connect_time = cluster_connect_time + cluster_info.cluster_connect_time;
	result.cluster_tree_build_time = cluster_tree_build_time + cluster_info.cluster_tree_build_time;             

	return Tcluster_info(result);
}


//**********************************************************************************************************************************


Tcluster_info Tcluster_info::operator = (const Tcluster_info& cluster_info)
{
	copy(cluster_info);
	return *this;
}


//**********************************************************************************************************************************

void Tcluster_info::copy(const Tcluster_info& cluster_info)
{
	Tdensity_info::copy(cluster_info);
	
	copy_hyper_parameters(cluster_info);
	copy_cluster_results(cluster_info);
	
	level_set_compute_time = cluster_info.level_set_compute_time;
	cluster_search_time = cluster_info.cluster_search_time;
	cluster_admin_time = cluster_info.cluster_admin_time;
	cluster_connect_time = cluster_info.cluster_connect_time;
	cluster_tree_build_time = cluster_info.cluster_tree_build_time;
}



//**********************************************************************************************************************************

void Tcluster_info::copy_hyper_parameters(const Tcluster_info& cluster_info)
{
	width = cluster_info.width;
	tau = cluster_info.tau;
	eps = cluster_info.eps;
	rho_start = cluster_info.rho_start;
	rho_stop = cluster_info.rho_stop;
	rho_step = cluster_info.rho_step;
	rho = cluster_info.rho;
}


//**********************************************************************************************************************************

void Tcluster_info::copy_cluster_results(const Tcluster_info& cluster_info)
{
	rho_star = cluster_info.rho_star;
	
	changed_level_set = cluster_info.changed_level_set;
	number_of_clusters = cluster_info.number_of_clusters;
	number_of_merges = cluster_info.number_of_merges;
	number_of_new_merges = cluster_info.number_of_new_merges;
	number_of_peaks = cluster_info.number_of_peaks;
	number_of_new_peaks = cluster_info.number_of_new_peaks;
	number_of_assigned_samples = cluster_info.number_of_assigned_samples;
	number_of_unassigned_samples = cluster_info.number_of_unassigned_samples;
	number_of_newly_assigned_samples = cluster_info.number_of_newly_assigned_samples;
	number_of_samples_with_new_cluster_number = cluster_info.number_of_samples_with_new_cluster_number;
	
	working_set_size = cluster_info.working_set_size;
	dataset_size = cluster_info.dataset_size;
}

//**********************************************************************************************************************************


Tcluster_info max(const Tcluster_info& cluster_info_1, const Tcluster_info& cluster_info_2)
{
	Tdensity_info result_base_1;
	Tdensity_info result_base_2;
	Tcluster_info result;
	
	
	result_base_1 = cluster_info_1;
	result_base_2 = cluster_info_2;
	result_base_1 = max(result_base_1, result_base_2);
	result = Tcluster_info(result_base_1);
	
	result.level_set_compute_time = max(cluster_info_1.level_set_compute_time, cluster_info_2.level_set_compute_time);
	result.cluster_search_time = max(cluster_info_1.cluster_search_time, cluster_info_2.cluster_search_time);
	result.cluster_admin_time = max(cluster_info_1.cluster_admin_time, cluster_info_2.cluster_admin_time);
	result.cluster_connect_time = max(cluster_info_1.cluster_connect_time, cluster_info_2.cluster_connect_time);
	result.cluster_tree_build_time = max(cluster_info_1.cluster_tree_build_time, cluster_info_2.cluster_tree_build_time);         

	return Tcluster_info(result);
}


#endif
