// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (DENSITY_INFO_CPP) 
	#define DENSITY_INFO_CPP



#include "sources/cluster/training_validation/density_info.h"





#include "sources/shared/basic_functions/flush_print.h"


//**********************************************************************************************************************************

Tdensity_info::Tdensity_info()
{
	clear();
}


//**********************************************************************************************************************************

Tdensity_info::Tdensity_info(const Tdensity_info& density_info)
{
	copy(density_info);
}


//**********************************************************************************************************************************

void Tdensity_info::clear()
{
	width = 0.0;

	
	working_set_size = 0;
	dataset_size = 0;

	clear_time();
};



//**********************************************************************************************************************************

void Tdensity_info::clear_time()
{
	data_preparation_time = 0.0;
	data_load_time = 0.0;
	tau_neighbor_build_time = 0.0;
	tau_neighbor_sort_time = 0.0;
	histogram_build_time = 0.0;
	density_compute_time = 0.0;
	
	build_time = 0.0;
	train_time = 0.0;
	val_time = 0.0;
	test_time = 0.0;
};


//**********************************************************************************************************************************

double Tdensity_info::get_full_time() const
{
	double full_time;
	
	full_time = data_load_time;
	full_time = full_time + data_preparation_time;
	full_time = full_time + tau_neighbor_build_time;
	full_time = full_time + tau_neighbor_sort_time;
	full_time = full_time + histogram_build_time;
	full_time = full_time + density_compute_time;
	full_time = full_time + test_time;
	
	return full_time;
}


//**********************************************************************************************************************************

void Tdensity_info::display(unsigned display_mode, unsigned info_level) const
{
	string output;
	double full_time;
	
	
	if (display_mode == TRAIN_INFO_DISPLAY_FORMAT_TIMINGS_FULL)
	{
		output = this->displaystring_train_time(display_mode); 
		
		output = output + "\nTest time:                   " + number_to_string(test_time, 4, DISPLAY_FLOAT);
		flush_info(info_level, output.c_str());
		
		full_time = this->get_full_time();
		flush_info(info_level, "\nMeasured time:               %2.4f", full_time);
	}
	else
	{
		output = this->displaystring_detailed(display_mode);

		flush_info(info_level, "\n");
		flush_info(info_level, output.c_str());
	}
}


//**********************************************************************************************************************************


string Tdensity_info::displaystring_train_time(unsigned display_mode) const
{
	string output;
	
	
	if (display_mode == TRAIN_INFO_DISPLAY_FORMAT_TIMINGS_FULL)
	{
		output = "\n\nData loading time:           " +  number_to_string(data_load_time, 4, DISPLAY_FLOAT);
		output = output + "\nData preparation time:       " + number_to_string(data_preparation_time, 4, DISPLAY_FLOAT);
		output = output + "\nTau-neighbor build time:     " + number_to_string(tau_neighbor_build_time, 4, DISPLAY_FLOAT);
		output = output + "\nTau-neighbor sort time:      " + number_to_string(tau_neighbor_sort_time, 4, DISPLAY_FLOAT);
		output = output + "\nHistogram build time:        " + number_to_string(histogram_build_time, 4, DISPLAY_FLOAT);
		output = output + "\nDensity compute time:        " + number_to_string(density_compute_time, 4, DISPLAY_FLOAT);
	}
	
	return output;
}

//**********************************************************************************************************************************


string Tdensity_info::displaystring_detailed(unsigned display_mode) const
{
	string output;
	
	if ((display_mode != TRAIN_INFO_DISPLAY_FORMAT_SUMMARIZED) and (display_mode != TRAIN_INFO_DISPLAY_FORMAT_TIMINGS_FULL))
	{
		output = output + "del: " + number_to_string(width, 3, DISPLAY_SCIENTIFIC);
	}
	return output;
}



//**********************************************************************************************************************************

Tdensity_info Tdensity_info::operator + (const Tdensity_info& density_info)
{
	Tdensity_info result;

	
	result.copy_hyper_parameters(*this);

	result.data_preparation_time = data_preparation_time + density_info.data_preparation_time;
	result.data_load_time = data_load_time + density_info.data_load_time;
	result.tau_neighbor_build_time = tau_neighbor_build_time + density_info.tau_neighbor_build_time;
	result.tau_neighbor_sort_time = tau_neighbor_sort_time + density_info.tau_neighbor_sort_time;
	result.histogram_build_time = histogram_build_time + density_info.histogram_build_time;
	result.density_compute_time = density_compute_time + density_info.density_compute_time;

	result.build_time = build_time + density_info.build_time;
	result.train_time = train_time + density_info.train_time;
	result.val_time = val_time + density_info.val_time;
	result.test_time = test_time + density_info.test_time;                  

	return Tdensity_info(result);
}


//**********************************************************************************************************************************


Tdensity_info Tdensity_info::operator = (const Tdensity_info& density_info)
{
	copy(density_info);
	return *this;
}


//**********************************************************************************************************************************

void Tdensity_info::copy(const Tdensity_info& density_info)
{
	copy_hyper_parameters(density_info);
	
	data_preparation_time = density_info.data_preparation_time;
	data_load_time = density_info.data_load_time;
	tau_neighbor_build_time = density_info.tau_neighbor_build_time;
	tau_neighbor_sort_time = density_info.tau_neighbor_sort_time;
	histogram_build_time = density_info.histogram_build_time;
	density_compute_time = density_info.density_compute_time;
	
	build_time = density_info.build_time;
	train_time = density_info.train_time;
	val_time = density_info.val_time;
	test_time = density_info.test_time;
}



//**********************************************************************************************************************************

void Tdensity_info::copy_hyper_parameters(const Tdensity_info& density_info)
{
	width = density_info.width;
	
	working_set_size = density_info.working_set_size;
	dataset_size = density_info.dataset_size;
}



//**********************************************************************************************************************************


Tdensity_info max(const Tdensity_info& density_info_1, const Tdensity_info& density_info_2)
{
	Tdensity_info result;
	
	
	result.data_preparation_time = max(density_info_1.data_preparation_time, density_info_2.data_preparation_time);
	result.data_load_time = max(density_info_1.data_load_time, density_info_2.data_load_time);
	result.tau_neighbor_build_time = max(density_info_1.tau_neighbor_build_time, density_info_2.tau_neighbor_build_time);
	result.tau_neighbor_sort_time = max(density_info_1.tau_neighbor_sort_time, density_info_2.tau_neighbor_sort_time);
	result.histogram_build_time = max(density_info_1.histogram_build_time, density_info_2.histogram_build_time);
	result.density_compute_time = max(density_info_1.density_compute_time, density_info_2.density_compute_time);

	result.build_time = max(density_info_1.build_time, density_info_2.build_time);
	result.train_time = max(density_info_1.train_time, density_info_2.train_time);
	result.val_time = max(density_info_1.val_time, density_info_2.val_time);
	result.test_time = max(density_info_1.test_time, density_info_2.test_time);                  

	return Tdensity_info(result);
}


#endif
