library(mvtnorm)




# -------------------------------------------------------------------------------------------------------------------------------------------------------

rotation_matrix <- function(angle)
{
	rot_angle = pi/180 * angle

	rot_matrix = matrix(nrow = 2, ncol = 2)

	rot_matrix[1,1] = cos(rot_angle)
	rot_matrix[1,2] = -sin(rot_angle)
	rot_matrix[2,1] = sin(rot_angle)
	rot_matrix[2,2] = cos(rot_angle)

	return(rot_matrix)
}





# -------------------------------------------------------------------------------------------------------------------------------------------------------


covariance_matrix <- function(angle, sigma)
{
	return(rotation_matrix(angle) %*% diag(sigma) %*% rotation_matrix(-angle))
}


# -------------------------------------------------------------------------------------------------------------------------


create_3gaussian_bg_data <- function(name, size = 50, include_boundary = FALSE, sigma_factor = 1.0)
{
	data = create_grid_data(size)
	
	pos_dens = dmvnorm(data[ ,2:3], mean = c(0.75, 0.95), sigma = covariance_matrix(40, sigma_factor * c(0.01, 0.015)))
	pos_dens = pos_dens + dmvnorm(data[ ,2:3], mean = c(0.65, 0.55), sigma = covariance_matrix(20, sigma_factor * c(0.01, 0.015)))
	neg_dens = dmvnorm(data[ ,2:3], mean = c(0.2, 0.25), sigma = covariance_matrix(115, sigma_factor * c(0.03, 0.04)))
	
	pos_dens = normalize_dens(pos_dens)
	neg_dens = normalize_dens(neg_dens)
	
	data[ ,1] = get_eta(pos_dens, neg_dens)
	
	write.table(data, file = get_full_bg_filename(get_toy_name(name, size) , "train"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
	
	if (include_boundary == TRUE)
	{
		number_of_colors = 255
		pure_noise_level = ceiling(number_of_colors/2)
		
		data[ ,1] = 1 + round((number_of_colors - 1) * data[ ,1])
		boundary_points = data[data[ ,1] == pure_noise_level ,2:3]
		boundary_points = boundary_points[order(boundary_points[ ,1], rev(boundary_points[ ,2])), ]

		write.table(boundary_points, file = get_full_bg_filename(name, "boundary"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
	}
}


# -------------------------------------------------------------------------------------------------------------------------


create_dog_data <- function(grid_size = 50, sample_size = 1000)
{
	name = "dogs"

	data = create_grid_data(grid_size)
	dens1 = dmvnorm(data[ ,2:3], mean = c(0.75, 0.80), sigma = covariance_matrix(40, 1.0 * c(0.01, 0.015)))
	dens2 = dmvnorm(data[ ,2:3], mean = c(0.55, 0.50), sigma = covariance_matrix(20, 1.5 * c(0.01, 0.015)))
	dens3 = dmvnorm(data[ ,2:3], mean = c(0.35, 0.35), sigma = covariance_matrix(20, 1.5 * c(0.01, 0.015)))
	
	dens = 0.4 * dens1 + 0.3 * dens2 + 0.3 * dens3
	write_dens_data(dens, name, grid_size)

	
	data1 = rmvnorm(1.2 * sample_size, mean = c(0.75, 0.80), sigma = covariance_matrix(40, 1.0 * c(0.01, 0.015)))
	data2 = rmvnorm(0.9 * sample_size, mean = c(0.55, 0.50), sigma = covariance_matrix(20, 1.5 * c(0.01, 0.015)))
	data3 = rmvnorm(0.9 * sample_size, mean = c(0.35, 0.35), sigma = covariance_matrix(20, 1.3 * c(0.01, 0.015)))
	
	write_sample_data(rbind(data1, data2, data3), name, sample_size)
}



# -------------------------------------------------------------------------------------------------------------------------


create_cat_data <- function(grid_size = 50, sample_size = 2000)
{
	name = "cats"

	data = create_grid_data(grid_size)
	dens = dmvnorm(data[ ,2:3], mean = c(0.15, 0.25), sigma = covariance_matrix(145, 0.3 * c(0.005, 0.04)))
	write_dens_data(dens, name, grid_size)

	data = rmvnorm(3 * sample_size, mean = c(0.15, 0.25), sigma = covariance_matrix(145, 0.3 * c(0.005, 0.04)))
	write_sample_data(data, name, sample_size)
}

# -------------------------------------------------------------------------------------------------------------------------


combine_cat_dog <- function(grid_size, sample_size)
{
	dog_data = as.matrix(read.table(get_full_2d_filename(get_toy_name("dogs", sample_size), "train"), sep=","), nrow=sample_size)
	cat_data = as.matrix(read.table(get_full_2d_filename(get_toy_name("cats", 2*sample_size), "train"), sep=","), nrow=2*sample_size)

	dog_label = rep(-1, sample_size)
	cat_label = rep(1, 2*sample_size)

	dog_data = cbind(dog_label, dog_data)
	cat_data = cbind(cat_label, cat_data)
	
	data = rbind(dog_data, cat_data)

	write_sample_data(data, "dogs-vs-cats", 1.5*sample_size, 2, 3)
	

	# 	-----
	
	dog_density = read.table(get_full_bg_filename(get_toy_name("dogs", grid_size), "dens"), sep = ",")
	cat_density = read.table(get_full_bg_filename(get_toy_name("cats", grid_size), "dens"), sep = ",")
	
	density = 2/3 * cat_density[ ,1] + 1/3 * dog_density[ ,1] 
	
	write_dens_data(density, "dog-vs-cat", grid_size)
	
	data = create_grid_data(grid_size)
	data[ ,1] = 1 - get_eta(cat_density[ ,1], dog_density[ ,1], 2/3)

	write.table(data, file = get_full_bg_filename(get_toy_name("dogs-vs-cats", grid_size) , "train"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
	
	
	# 	-----
	
	
	number_of_colors = 255
	pure_noise_level = ceiling(number_of_colors/2)
	
	data[ ,1] = 1 + round((number_of_colors - 1) * data[ ,1])
	boundary_points = data[data[ ,1] == pure_noise_level ,2:3]
# 	boundary_points = boundary_points[order(boundary_points[ ,1], rev(boundary_points[ ,2])), ]

	write.table(boundary_points, file = get_full_bg_filename("dogs-vs-cats", "boundary"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
}


# -------------------------------------------------------------------------------------------------------------------------

write_dens_data <- function(dens, name, grid_size)
{
	data = create_grid_data(grid_size)
	data[ ,1] = normalize_dens(dens)

	write.table(data, file = get_full_bg_filename(get_toy_name(name, grid_size), "dens"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")

	data = marginalize_dens(data, 2)
	write.table(data, file = get_full_1d_filename(get_toy_name(name, grid_size), "dens"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
}



# -------------------------------------------------------------------------------------------------------------------------

write_sample_data <- function(data, name, sample_size, dim1=1, dim2=2)
{
	data = restrict_data(data, 2 * sample_size, dim1, dim2)
	
	write.table(data[1:sample_size, ], file = get_full_2d_filename(get_toy_name(name, sample_size), "train"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
	
	write.table(data[(sample_size+1):nrow(data), ], file = get_full_2d_filename(get_toy_name(name, sample_size), "test"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
}


# -------------------------------------------------------------------------------------------------------------------------

get_toy_name <- function(name, size)
{
	paste(name, size, sep="_")
}

# -------------------------------------------------------------------------------------------------------------------------


get_eta <- function(pos_dens, neg_dens, pos_prob)
{
	pos_dens = normalize_dens(pos_dens)
	neg_dens = normalize_dens(neg_dens)
	
	eta = pos_prob * pos_dens / (pos_prob * pos_dens + (1-pos_prob) * neg_dens)

# 	eta = 0.5 * (pos_dens - neg_dens) / (pos_dens + neg_dens) + 0.5
	
	return(eta)
}


# -------------------------------------------------------------------------------------------------------------------------

normalize_dens <- function(density)
{
	return(density / sum(density))
}



# -------------------------------------------------------------------------------------------------------------------------

marginalize_dens <- function(density_on_grid, coordinate_integrated_over)
{
	data = matrix(0.0, nrow = sqrt(nrow(density_on_grid)), ncol=2)

	xgrid_points = sort(unique(density_on_grid[ ,coordinate_integrated_over]))
	
	for (i in 1:nrow(data))
	{
		data[i, 2] = xgrid_points[i] 
		data[i, 1] = sum(density_on_grid[density_on_grid[ ,coordinate_integrated_over] == xgrid_points[i], 1])
	}
	
	return (data)
}




# -------------------------------------------------------------------------------------------------------------------------


convert_density_to_aux_data <-function(density, grid_size, y_res = 10000)
{
	x.bin = seq(0, 1, length=grid_size+1)
	
	dens_sum = rep(0.0, grid_size)
	
	for (i in 1:grid_size)
		dens_sum[i] = round(y_res * sum(density[findInterval(density[ ,2], x.bin) == i, 1]))
		

	data = matrix(0, ncol = 2, nrow = sum(dens_sum)) 
	
	start = 1
	for (i in 1:grid_size)
		if (dens_sum[i] > 0)
		{
			data[start:(start + dens_sum[i]-1), 1] = rep((i - 0.5) / grid_size, dens_sum[i])
			start = start + dens_sum[i]
		}
		
	return(data)
}







