library(graphics)
library(grDevices)


source("filename_scripts.r")
source("create_data.r")
source("plot_data.r")

source("art_cluster_data.r")

background = "black"
non_assigned = "white"

# ----------------------------------------------------------------------------------------------------------------------


plot_cluster_datasets <- function(name, color_permutation, hier_depth)
{
	result_data = read.table(get_control_filename(name, "result.csv"), sep=",")

	if (hier_depth != -1)
		image_name = paste(name, "_", hier_depth, sep = "")
	else
		image_name = name

	image_name = get_image_filename(image_name, "result", FALSE, "")
	open_image(image_name, background)
	plot_cluster_points(result_data, color_permutation)
	close_image()


	result_data = read.table(get_control_filename(name, "predict.csv"), sep=",")

	if (hier_depth != -1)
		image_name = paste(name, "_", hier_depth, sep = "")
	else
		image_name = name

	image_name = get_image_filename(image_name, "predict", FALSE, "")
	open_image(image_name, background)
	largest_cluster_number = max(color_permutation)
	pal = get_colors(largest_cluster_number)

	plot_bg(result_data, pal[color_permutation], TRUE)
	close_image()
}

# ----------------------------------------------------------------------------------------------------------------------


plot_cluster_truth <- function(name, color_permutation, hier_depth)
{
	result_data = read.table(get_control_filename(name, "truth.csv"), sep=",")

	if (hier_depth != -1)
		image_name = paste(name, "_", hier_depth, sep = "")
	else
		image_name = name

	image_name = get_image_filename(image_name, "truth", FALSE, "")
	open_image(image_name, background)
	plot_cluster_points(result_data, color_permutation)
	close_image()
	
	
	image_name = name
	result_data[ ,1] = 0.0
	image_name = get_image_filename(image_name, "raw", FALSE, "")
	open_image(image_name, background)
	plot_cluster_points(result_data, color_permutation)
	close_image()



	result_data = read.table(get_control_filename(name, "dens.csv"), sep=",")

	if (hier_depth != -1)
		image_name = paste(name, "_", hier_depth, sep = "")
	else
		image_name = name

	image_name = get_image_filename(image_name, "inf", FALSE, "")
	open_image(image_name, background)
	largest_cluster_number = max(color_permutation)
	pal = get_colors(largest_cluster_number)

	plot_bg(result_data, pal[color_permutation], TRUE)
	close_image()
	
	
	result_data = read.table(get_data_filename(name, "dens"), sep=",")

	if (hier_depth != -1)
		image_name = paste(name, "_", hier_depth, sep = "")
	else
		image_name = name

	image_name = get_image_filename(image_name, "dens", FALSE, "")
	open_image(image_name, background)
	
	terrain_cols = 20
	pal = terrain.colors(n=terrain_cols)
	result_data[ ,1] = ceiling(terrain_cols * result_data[ ,1] / max(result_data[ ,1]))

	plot_bg(result_data, pal, TRUE)
	close_image()
}

# ----------------------------------------------------------------------------------------------------------------------

get_cluster_numbers <- function(data)
{
	cluster_numbers = sort(unique(data[ ,1]))
	cluster_numbers = as.matrix(cluster_numbers, ncol = 1)

	if (cluster_numbers[1] != 0)
		cluster_numbers = rbind(0, cluster_numbers)
	
	return (cluster_numbers)
}


# ----------------------------------------------------------------------------------------------------------------------

get_colors <- function(largest_cluster_number)
{
	number_of_clusters = largest_cluster_number + 1

	pal = rainbow(number_of_clusters)
	pal = append(non_assigned, pal)

	return (pal)
}


# ----------------------------------------------------------------------------------------------------------------------

plot_cluster_points <- function(data, color_permutation, maxsize=200000)
{
	data = restrict_data(data, maxsize)

	largest_cluster_number = max(color_permutation)
	palette = get_colors(largest_cluster_number)

	pointsize = 1.5
	pointform = 20 

	cluster_numbers = seq(0, largest_cluster_number)
	number_of_clusters = length(cluster_numbers)

	for (i in 1:number_of_clusters)
	{
		cluster_data = data[data[ ,1] == cluster_numbers[i] ,2:3]

		points(cluster_data, col = palette[color_permutation[i]], pch = pointform, cex = pointsize)
	}
}


# ----------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------



gmixtures <- function(means, covariances, fractions, size, type)
{
	peaks = max(1, nrow(means))

	sigma = covariance_matrix(covariances[1, 1], covariances[1, 2:3])
	data = create_gauss_data(size, means[1, ], sigma, fractions[1], type)

	if (peaks > 1)
		for (i in 2:peaks)
		{
			sigma = covariance_matrix(covariances[i, 1], covariances[i, 2:3])
			data_tmp = create_gauss_data(size, means[i, ], sigma, fractions[i], type)

			data = combine_data(data, data_tmp, type)
		}

	if (type == "sample")
		data = restrict_data(data, size, 1, 2)

	return(data)
}


# ----------------------------------------------------------------------------------------------------------------------


all_gmixtures <- function(name, grid_size, sample_size, means, covariances, fractions, plot_only)
{
	print(name)

	if (plot_only == FALSE)
	{
		print("Creating density")
		data = gmixtures(means, covariances, fractions, grid_size, "dens")
		if (nrow(means) > 100)
			data[ ,1] = round(data[ ,1], floor(2 * log10(grid_size)))
		write.table(data, file = get_data_filename(name, "dens"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")

		print("Creating samples")
		data = gmixtures(means, covariances, fractions, 2 * sample_size, "sample")
		write.table(data[1:sample_size, ], file = get_data_filename(name, "train"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
		write.table(data[(sample_size + 1):(2 * sample_size), ], file = get_data_filename(name, "test"), row.names = FALSE, col.names = FALSE, append = FALSE, sep = ", ", eol = "\n")
		
		print("Datasets written to disk")
	}
}



# ----------------------------------------------------------------------------------------------------------------------


toy_gmixtures <- function(grid_size, sample_size, peaks, plot_only)
{
	name = paste("toy-", peaks, "g.", as.character(sample_size), sep = "")

	all_peaks = 12
	means = matrix(0, ncol = 2, nrow = all_peaks)
	covariances = matrix(0, ncol = 3, nrow = all_peaks)
	fractions = matrix(0, ncol = 1, nrow = all_peaks)

	means[1, ] = c(0.75, 0.8)
	means[2, ] = c(0.52, 0.5)
	means[3, ] = c(0.1, 0.2)  
	means[4, ] = c(0.11, 0.79)
	means[5, ] = c(0.08, 0.69)
	means[6, ] = c(0.21, 0.74)
	means[7, ] = c(0.77, 0.09)
	means[8, ] = c(0.82, 0.11)
	means[9, ] = c(0.80, 0.14)
	means[10, ] = c(0.87, 0.19)
	means[11, ] = c(0.92, 0.21)
	means[12, ] = c(0.88, 0.24)

	covariances[1, ] = c(40, 0.005, 0.015)
	covariances[2, ] = c(20, 0.015, 0.030)
	covariances[3, ] = c(70, 0.010, 0.010)
	covariances[4, ] = c(75, 0.0006, 0.0003)
	covariances[5, ] = c(30, 0.0003, 0.0002)
	covariances[6, ] = c(60, 0.0002, 0.0004)
	covariances[7, ] = c(35, 0.00002, 0.00001)
	covariances[8, ] = c(60, 0.00001, 0.00001)
	covariances[9, ] = c(25, 0.00001, 0.00002)
	covariances[10, ] = c(35, 0.000002, 0.000002)
	covariances[11, ] = c(60, 0.000002, 0.000002)
	covariances[12, ] = c(25, 0.000002, 0.000002)

	fractions[1] = 0.3
	fractions[2] = 0.3
	fractions[3] = 0.3
	fractions[4] = 0.03
	fractions[5] = 0.04
	fractions[6] = 0.03
	fractions[7] = 0.003
	fractions[8] = 0.004
	fractions[9] = 0.003
	fractions[10] = 0.003
	fractions[11] = 0.004
	fractions[12] = 0.003

	if (peaks < all_peaks)
	fractions[(peaks+1):all_peaks] = 0
	fractions = fractions / sum(fractions)
	
	all_gmixtures(name, grid_size, sample_size, means[1:peaks, ], covariances[1:peaks, ], fractions[1:peaks], plot_only)
	
	return(name)
}


# ----------------------------------------------------------------------------------------------------------------------


crosses_gmixtures <- function(grid_size, sample_size, sigma, plot_only)
{
	name = paste("crosses-", sigma, ".", as.character(sample_size), sep = "")

	all_peaks = 6
	means = matrix(0, ncol = 2, nrow = all_peaks)
	covariances = matrix(0, ncol = 3, nrow = all_peaks)
	fractions = matrix(0, ncol = 1, nrow = all_peaks)
	
	means[1, ] = c(0.31, 0.79)
	means[2, ] = c(0.31, 0.79)
	means[3, ] = c(0.79, 0.31)
	means[4, ] = c(0.79, 0.31)
	means[5, ] = c(0.33, 0.30)
	means[6, ] = c(0.33, 0.30)
	
	covariances[1, ] = c(40, 0.05, 0.005)
	covariances[2, ] = c(150, 0.05, 0.005)
	covariances[3, ] = c(20, 0.05, 0.005)
	covariances[4, ] = c(120, 0.05, 0.005)
	covariances[5, ] = c(10, 0.5, 0.05)
	covariances[6, ] = c(100, 0.5, 0.05)
	
	covariances[ ,2:3] = sigma/10 * covariances[ ,2:3]

	fractions[1] = 0.05
	fractions[2] = 0.05
	fractions[3] = 0.05
	fractions[4] = 0.05
	fractions[5] = 0.4
	fractions[6] = 0.4

	all_gmixtures(name, grid_size, sample_size, means, covariances, fractions, plot_only)
	
	return(name)
}


# ----------------------------------------------------------------------------------------------------------------------


s2_gmixtures <- function(grid_size, sample_size, sigma, plot_only)
{
	name = paste("s2-", sigma, ".", as.character(sample_size), sep = "")

	all_peaks = 16
	means = matrix(0, ncol = 2, nrow = all_peaks)
	covariances = matrix(0, ncol = 3, nrow = all_peaks)
	fractions = matrix(0, ncol = 1, nrow = all_peaks)
	
	
	means[1, ] = c(0.3502, 0.1711)
	means[2, ] = c(0.6629, 0.1543) 
	means[3, ] = c(0.5280, 0.2177) 
	means[4, ] = c(0.5158, 0.4068) 
	means[5, ] = c(0.3483, 0.3756) 
	means[6, ] = c(0.1894, 0.4518) 
	means[7, ] = c(0.7646, 0.6154) 
	means[8, ] = c(0.6871, 0.4695) 
	means[9, ] = c(0.4757, 0.8456) 
	means[10, ] = c(0.5830, 0.6958) 
	means[11, ] = c(0.7508, 0.7803) 
	means[12, ] = c(0.4117, 0.5892) 
	means[13, ] = c(0.1383, 0.2217) 
	means[14, ] = c(0.2391, 0.7237) 
	means[15, ] = c(0.7955, 0.2311) 
	means[16, ] = c(0.5158, 0.4068)
	
	covariances[ ,1] = 0
	covariances[ ,2:3] = sigma / 1000
	covariances[5, ] = c(130, sigma / 1000, sigma / 3000)
	covariances[11, ] = c(170, sigma / 1000, sigma / 3000)
	covariances[12, ] = c(100, sigma / 1000, sigma / 10000)
	covariances[16, ] = c(0, 1/4, 1/4)
	
	fractions[] = 1/all_peaks
	
	all_gmixtures(name, grid_size, sample_size, means, covariances, fractions, plot_only)
	
	return(name)
}


# ----------------------------------------------------------------------------------------------------------------------


bananas_gmixtures <- function(grid_size, sample_size, sigma, plot_only)
{
	name = paste("bananas-", sigma, ".", as.character(sample_size), sep = "")

	resolution = min(1000, sample_size)
	means = matrix(0, ncol = 2, nrow = resolution)
	covariances = matrix(0, ncol = 3, nrow = resolution)
	fractions = matrix(0, ncol = 1, nrow = resolution)

	
	for (i in 1:(resolution/2))
	{
		means[i, 1] = 0.41 + 0.25 * cos(pi * 2 * i / resolution + pi / 4)
		means[i, 2] = 0.41 + 0.25 * sin(pi * 2 * i / resolution + pi / 4)
		
		means[i + (resolution/2), 1] = 0.59 + 0.25 * cos(pi * 2 * i / resolution + 5 * pi / 4)
		means[i + (resolution/2), 2] = 0.59 + 0.25 * sin(pi * 2 * i / resolution + 5 * pi / 4)
	}
	
	covariances[ ,1] = 0
	covariances[ ,2:3] = sigma / 1000
	
	fractions[] = 1/resolution
	
	all_gmixtures(name, grid_size, sample_size, means, covariances, fractions, plot_only)

	return(name)
}


  
# ----------------------------------------------------------------------------------------------------------------------

cluster <- function(name, mode, plot_only, hier_depth = -1, compare_mode = 0, rho_tolerance = 1.0)
{
	cluster_call = "../../bin/dbc -T 0 -d 1 -P 2 -D 0.75 0.25 50 1 -t 2 -e 3"
	cluster_call = paste(cluster_call, "-H", hier_depth, "-s", compare_mode, rho_tolerance, sep = " ")
		
	modifier_weighted = "+l 0 +w 1 +i 0 +g 0"
	modifier_labeled = "+l 1 +w 0 +i 0 +g 0"
	modifier_unlabeled = "+l 0 +w 0 +i 0 +g 0"

	if (mode == "dens")
	{
		kde_call = "-E 0"
		
		train_name = paste(get_data_filename(name, "dens"), modifier_weighted, sep = " ")
		result_name = get_control_filename(name, "dens.csv")
		
		test_name = paste(get_data_filename(name, "train"), modifier_unlabeled, sep = " ")
		predict_name = paste(get_control_filename(name, "truth.csv"), modifier_labeled, sep = " ")
	}
	
	if (mode == "sample")
	{
		kde_call = "-E 3"
		
		train_name = paste(get_control_filename(name, "truth.csv"), modifier_labeled, sep = " ")
		result_name = get_control_filename(name, "result.csv")
		
		test_name = paste(get_control_filename(name, "dens.csv"), modifier_weighted, sep = " ")
		predict_name = get_control_filename(name, "predict.csv")
	}
	
	if (plot_only == FALSE)
	{
		full_call = paste(cluster_call, kde_call, train_name, result_name, test_name, predict_name, sep = " ")
		print(full_call)
		system(full_call)
	}

	result_data = read.table(result_name, sep=",")
	cluster_numbers = get_cluster_numbers(result_data)

	return (max(cluster_numbers))
}
  

# ----------------------------------------------------------------------------------------------------------------------


cluster_and_plot <- function(name, plot_only, color_permutation, hier_depth = -1, compare_mode = 0)
{
	dens_max = cluster(name, "dens", plot_only, hier_depth)
	sample_max = cluster(name, "sample", plot_only, hier_depth, compare_mode, 1.0)
	
	largest_cluster_number = max(dens_max, sample_max)

	if (length(color_permutation) != largest_cluster_number + 1)
	{
		if (largest_cluster_number == 1)
			color_permutation = c(1,2)
		else
		{
			color_permutation = sample(2:(largest_cluster_number+1))
			color_permutation = append(1, color_permutation)
		}
	}
	print(color_permutation)

	plot_cluster_datasets(name, color_permutation, hier_depth)
	plot_cluster_truth(name, color_permutation, hier_depth)
}

  
# ----------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------

grid_size = 1000
plot_only = FALSE

# for (sigma in c(1, 5, 10, 15))
# 	for (j in c(5, 10, 25, 50, 100))
# 	{
# 		size = 1000 * j
# 		name = crosses_gmixtures(grid_size, size, sigma, plot_only)
# 		cluster_and_plot(name, plot_only, 0)
# 	}

# peaks = 9
# # j = 10
# # for (peaks in c(3, 6, 9))
# 	for (j in c(5, 10, 25, 50, 100))
# 	{
# 		size = 1000 * j
# 		name = toy_gmixtures(grid_size, size, peaks, plot_only)
# 		
# 		color_permutation = c(1, 2, 3, 4, 5, 6)
# 		cluster_and_plot(name, FALSE, color_permutation, -1, 1)
# 	}
# 

sigma = 15
# j = 10
# for (sigma in c(1, 2, 3, 4))
# 	for (j in c(5, 10, 25, 50, 100))
# 	{
# 		size = 1000 * j
# 		name = s2_gmixtures(grid_size, size, sigma, plot_only)
# 
# 		cluster_and_plot(name, FALSE, 0, -1, 1)
# 	}
# 
# 
# for (sigma in c(1, 5, 10, 20))
	for (j in c(25, 50, 100))
	{
		size = 1000 * j
		name = bananas_gmixtures(grid_size, size, sigma, plot_only)
		cluster_and_plot(name, plot_only, 0)
	}

