// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (RANDOM_GENERATOR_CPP) 
	#define RANDOM_GENERATOR_CPP



 
#include "sources/shared/basic_functions/random_generator.h"


#ifndef __RAND__
  #define __RAND__() rand();
#endif

#ifndef __SRAND__
  #define __SRAND__(i) srand(i);
#endif


#include "sources/shared/basic_functions/math_misc.h"
#include "sources/shared/basic_functions/flush_print.h"

#include <ctime>
#include <cmath>


//**********************************************************************************************************************************



void fix_random_seed(int& random_seed)
{
	if (random_seed == RANDOM_SEED_BY_TIMER)
		random_seed = unsigned(time(NULL));
}

//**********************************************************************************************************************************

void init_random_generator(int random_seed, unsigned extra_seed)
{
	fix_random_seed(random_seed);
	__SRAND__(random_seed + extra_seed);
}

//**********************************************************************************************************************************


int get_random_number(int min_value, int max_value)
{
	int random_value;
	
	random_value = __RAND__();
	if ((min_value == 0) and (max_value == RAND_MAX-1))
		return random_value;
	else
		return random_value % (max_value - min_value + 1) + min_value;
}

//**********************************************************************************************************************************

double get_uniform_random_number()
{
	return (double(get_random_number()) / double(RAND_MAX-1));
}

//**********************************************************************************************************************************


double get_normal_random_number()
{
	double u;
	double v;
	
// 	This compuation uses the classical Box-Mueller formula
	
	u = get_uniform_random_number();
	v = get_uniform_random_number();
	
	return sqrt(-2 * log(u)) * cos(2 * MY_PI * v);
}


//**********************************************************************************************************************************

unsigned get_bernoulli_random_number(double p)
{
	if (get_uniform_random_number() < p)
		return 1;
	else
		return 0;
}


//**********************************************************************************************************************************

void get_random_distribution(vector <double>& distribution, unsigned size)
{
	unsigned i;
	double sum;

	sum = 0.0;	
	distribution.resize(size);

	for (i=0; i<size; i++)
	{
		distribution[i] = get_uniform_random_number();
		sum = sum + distribution[i];
	}
	
	for (i=0; i<size; i++)
		distribution[i] = distribution[i] / sum;
}

//**********************************************************************************************************************************


void get_uniform_random_vector(vector <double>& random_vector, unsigned size, double min_value, double max_value)
{
	unsigned i;
	
	random_vector.resize(size);
	for (i=0; i<size; i++)
		random_vector[i] = (max_value - min_value) * (double(get_random_number()) / double(RAND_MAX)) + min_value;
}



//**********************************************************************************************************************************


void get_normal_random_vector(vector <double>& random_vector, unsigned size)
{
	unsigned i;
	double u;
	double v;
	
	
// 	This compuation uses the classical Box-Mueller formula
	
	random_vector.resize(size);
	for (i=0; i + 1 <size; i = i + 2)
	{
		u = get_uniform_random_number();
		v = get_uniform_random_number();

		random_vector[i] = sqrt(-2 * log(u)) * cos(2 * MY_PI * v);
		random_vector[i+1] = sqrt(-2 * log(u)) * sin(2 * MY_PI * v);
	}
	
	if ((size % 2) == 1)
	{
		u = get_uniform_random_number();
		v = get_uniform_random_number();
	
		random_vector[size - 1] = sqrt(-2 * log(u)) * cos(2 * MY_PI * v);	
	}
}


//**********************************************************************************************************************************

void get_uniform_ball_vector(vector <double>& random_vector, unsigned dim, double radius, unsigned metric)
{
	unsigned i;
	double u;
	double norm;
	double factor;
	vector <double> y;
	
	
	if (metric == SUP_NORM)
		get_uniform_random_vector(random_vector, dim, -radius, radius);
	
	if (metric == EUCLIDEAN)
	{
		u = get_uniform_random_number();
		get_normal_random_vector(y, dim);
		
		norm = 0.0;
		for (i=0; i<dim; i++)
			norm = norm + y[i] * y[i];
		norm = sqrt(norm);
		
		factor = radius * pow(u, 1.0 / double(dim)) / norm;
		
		random_vector.resize(dim);
		for (i=0; i<dim; i++)
			random_vector[i] = y[i] * factor;
	}
}



#endif
