// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.


#if !defined (HISTOGRAM_CPP)
	#define HISTOGRAM_CPP


	
	
#include "sources/shared/kernel/histogram.h"



#include "sources/shared/basic_types/vector.h"
#include "sources/shared/system_support/timing.h"

	
#include <math.h> 
#include <stdlib.h>


int const OUTSIDE_HISTOGRAM = -1;


//**********************************************************************************************************************************

Thistogram::Thistogram()
{
	clear();
}


//**********************************************************************************************************************************

Thistogram::~Thistogram()
{
	clear();
}



//**********************************************************************************************************************************

void Thistogram::clear()
{
	dim = 0;
	cell_width = 1.0;
	
	number_of_bins.clear();
	bin_minima.clear();
	bin_maxima.clear();
	
	density_histogram.clear();
	label_sum_histogram.clear();
	indices_in_cell.clear();
}



//**********************************************************************************************************************************

void Thistogram::reserve(const Tdataset_info& dataset_info, double cell_width)
{
	unsigned i;
	double min_value;
	double max_value;
	double excess_width;
	double current_number_of_bins;
	
	clear();
	
	Thistogram::cell_width = cell_width;
	Thistogram::dataset_info = dataset_info;
	
	dim = dataset_info.dim;
	number_of_bins.resize(dim);
	bin_minima.resize(dim);
	bin_maxima.resize(dim);
	
	if ((dim == 0) or (dataset_info.size == 0))
		flush_exit(ERROR_DATA_MISMATCH, "Tdataset_info object belongs to empty data set.");
	
	if ((dataset_info.minima.size() != dim) or (dataset_info.maxima.size() != dim))
		flush_exit(ERROR_DATA_MISMATCH, "Tdataset_info object was not fully constructed.");
	
	for (i=0; i<dim; i++)
	{
		min_value = dataset_info.minima[i];
		max_value = dataset_info.maxima[i];

		current_number_of_bins = ceil((max_value - min_value) / cell_width);
		excess_width = current_number_of_bins * cell_width - (max_value - min_value);
		
		number_of_bins[i] = unsigned(current_number_of_bins);
		bin_minima[i] = min_value - excess_width/2.0;
		bin_maxima[i] = max_value + excess_width/2.0;
	}
}


//**********************************************************************************************************************************

void Thistogram::assign(const Tdataset& dataset, double& build_time)
{
	unsigned i;
	hashkey_type cell_key;
	
	
	build_time = get_process_time_difference(build_time);
	
	if (number_of_bins.size() == 0)
		flush_exit(ERROR_DATA_MISMATCH, "Thistogram object has not been reserved.");
		
	
	density_histogram.clear();
	label_sum_histogram.clear();
	indices_in_cell.clear();
	
	for (i=0; i<dataset.size(); i++)
	{
		cell_key = get_key(dataset.sample(i));
		if (density_histogram.count(cell_key) == 0)
		{
			density_histogram.insert(std::pair<hashkey_type, unsigned>(cell_key, 1));
			label_sum_histogram.insert(std::pair<hashkey_type, double>(cell_key, dataset.sample(i)->label));
		}
		else
		{
			density_histogram[cell_key]++;
			label_sum_histogram[cell_key] = label_sum_histogram[cell_key] + dataset.sample(i)->label;
		}
		indices_in_cell[cell_key].push_back(i);
	}
	
	build_time = get_process_time_difference(build_time);
}


//**********************************************************************************************************************************

vector <double> Thistogram::predict_labels(const Tdataset& dataset, unsigned label_aggregation_method)
{
	unsigned i;
	hashkey_type cell_key;
	vector <double> labels;


	labels.resize(dataset.size());
	for (i=0; i<dataset.size(); i++)
	{
		cell_key = get_key(dataset.sample(i));
		
		switch(label_aggregation_method)
		{
			case CELL_HITS_HISTO:
				labels[i] = double(density_histogram[cell_key]);
				break;
			case AVERAGE_LABEL_SUM_HISTO:
				labels[i] = label_sum_histogram[cell_key] / max(1.0, double(density_histogram[cell_key]));
				break;
			case LABEL_SUM_HISTO:
				labels[i] = label_sum_histogram[cell_key];
				break;
		}
	}
	
	return labels;
}

//**********************************************************************************************************************************

void Thistogram::get_mid_point_dataset(Tdataset& mp_dataset)
{
	unsigned i;
	unsigned j;
	unsigned bin_number;
	Tsample sample;
	vector <double> density;
	vector <hashkey_type> cell_keys;
	my_unordered_map <hashkey_type, unsigned>::const_iterator cell_position;
	
	
	density.reserve(density_histogram.size());
	cell_keys.reserve(density_histogram.size());
	for (cell_position = density_histogram.begin(); cell_position != density_histogram.end(); ++cell_position)
	{
		cell_keys.push_back(cell_position->first);
		density.push_back(double(cell_position->second));
	}
	sort_up(cell_keys, density);


	sample = Tsample(CSV, dim);
	mp_dataset.clear();
	mp_dataset.enforce_ownership();	
	for (i=0; i<cell_keys.size(); i++)
	{
		sample.label = density[i];
		for (j=0; j<dim; j++)
		{
			bin_number = strtoul(cell_keys[i].substr(2 * j, 2).c_str(), NULL, 16);
			sample.change_coord(j, get_bin_mid_point(bin_number, bin_minima[j]));
		}
		mp_dataset.push_back(sample);
	}
}


//**********************************************************************************************************************************

unsigned Thistogram::size() const
{
	return density_histogram.size();
}

//**********************************************************************************************************************************


vector <unsigned> Thistogram::get_cell_statistics() const
{
	unsigned i;
	vector <unsigned> cell_statistics;
	my_unordered_map <hashkey_type, list <unsigned> >::const_iterator cell_position;
	
	
	i = 0;
	cell_statistics.resize(size());
	for (cell_position = indices_in_cell.begin(); cell_position != indices_in_cell.end(); ++cell_position)
	{
		cell_statistics[i] = cell_position->second.size();
		i++;
	}
	
	return cell_statistics;
}


//**********************************************************************************************************************************



double Thistogram::get_number_of_cells() const
{
	unsigned i;
	double total_number;
	
	if (dim == 0)
		return 0.0;
	
	total_number = double(number_of_bins[0]);
	for (i=1; i<dim; i++)
		total_number = total_number * max(1.0, double(number_of_bins[i]));
	
	return total_number;
}


//**********************************************************************************************************************************


const list <unsigned>& Thistogram::get_indices_in_cell(const Tsample& sample)
{
	hashkey_type cell_key;
	
	
	cell_key = get_key(sample);
	return indices_in_cell[cell_key];
}


//**********************************************************************************************************************************


unsigned Thistogram::get_hits_in_cell(const Tsample& sample)
{
	hashkey_type cell_key;
	
	
	cell_key = get_key(sample);
	return density_histogram[cell_key];
}


//**********************************************************************************************************************************


void Thistogram::set_label_sum_in_cell(const Tsample& sample, double value)
{
	hashkey_type cell_key;
	
	
	cell_key = get_key(sample);
	label_sum_histogram[cell_key] = value;
}


//**********************************************************************************************************************************


double Thistogram::get_label_sum_in_cell(const Tsample& sample)
{
	hashkey_type cell_key;
	
	
	cell_key = get_key(sample);
	return label_sum_histogram[cell_key];
}


//**********************************************************************************************************************************


int Thistogram::get_bin_number(double value, double min_value, double max_value, unsigned number_of_bins) const
{
	if ((value < min_value) or (value > max_value))
		return OUTSIDE_HISTOGRAM;
	else
		return min(number_of_bins - 1, unsigned((value - min_value) / cell_width));
}


//**********************************************************************************************************************************


double Thistogram::get_bin_mid_point(unsigned bin_number, double min_value)
{
	return double(bin_number) * cell_width + min_value + 0.5 * cell_width;
}


//**********************************************************************************************************************************



hashkey_type Thistogram::get_key(const Tsample& sample) const
{
	unsigned i;
	int bin_number;
	char cell_number_str[4];
	hashkey_type key;
	
	for (i=0; i<dim; i++)
	{
		bin_number = get_bin_number(sample.coord(i), bin_minima[i], bin_maxima[i], number_of_bins[i]);
		if (bin_number == OUTSIDE_HISTOGRAM)
			sprintf(cell_number_str, "-");
		else
			sprintf(cell_number_str, "%02x", bin_number);
		key.append(cell_number_str);
	}
	
	return key;
}






//**********************************************************************************************************************************




#endif
