// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.



#include "../shared/system_support/os_specifics.h"
#include "../shared/basic_functions/flush_print.h"
#include "../shared/basic_types/dataset_info.h"
#include "../shared/system_support/timing.h"
#include "../shared/command_line/command_line_parser.h"


#include "../svm/training_validation/svm_manager.h"




//**********************************************************************************************************************************


unsigned const ERROR_clp_ted_c = 120;
unsigned const ERROR_clp_ted_l = 121;
unsigned const ERROR_clp_ted_L = 122;
unsigned const ERROR_clp_ted_t = 123;

//**********************************************************************************************************************************


class Tcommand_line_parser_extract_data: public Tcommand_line_parser
{
	public:
		Tcommand_line_parser_extract_data();
		void parse();

		unsigned task;
		unsigned cell;
		double label;
		bool cells_to_labels;
		bool only_by_label;
		string aux_filename;
		Tsample_file_format train_file_format;
		Tsample_file_format input_file_format;
		Tsample_file_format output_file_format;

	protected:
		void exit_with_help();
		void display_help(unsigned error_code);
};


//**********************************************************************************************************************************


Tcommand_line_parser_extract_data::Tcommand_line_parser_extract_data()
{
	task = 0;
	cell = 0;
	cells_to_labels = true;
	only_by_label = false;
	command_name = "extract-data";
};

//**********************************************************************************************************************************


void Tcommand_line_parser_extract_data::parse()
{
	check_parameter_list_size();
	for(current_position=1; current_position<parameter_list_size; current_position++)
		if (Tcommand_line_parser::parse("-h-r-d") == false)
		{
			if(parameter_list[current_position][0] != '-') 
				break;
			if (string(parameter_list[current_position]).size() > 2)
				Tcommand_line_parser::exit_with_help(ERROR_clp_gen_unknown_option);
			
			switch(parameter_list[current_position][1])
			{
				case 'c':
					cell = get_next_number(ERROR_clp_ted_c, 0);
					cells_to_labels = false;
					break;
				case 'l':
					current_position++;
					input_file_format = get_next_data_file_format(ERROR_clp_ted_l);
					current_position--;
					break;
				case 'L':
					label = get_next_number(ERROR_clp_ted_L, -get_limits_max<double>());
					only_by_label = true;
					break;
				case 't':
					task = get_next_number(ERROR_clp_ted_t, 0);
					break;
				default:
					Tcommand_line_parser::exit_with_help(ERROR_clp_gen_unknown_option);
			}
		}
		
	train_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
	check_labeled_data_format(train_file_format);
	if (only_by_label == false)
		aux_filename = get_next_aux_filename(ERROR_clp_gen_missing_data_file_name);
	else
		input_file_format.clear();
	output_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
};


//**********************************************************************************************************************************


void Tcommand_line_parser_extract_data::exit_with_help()
{
	flush_info(INFO_SILENCE,
	"\n\nextract-data [options] <train_data_file> <train_aux_file> <output_data_file>\n"
	"\nextract-data reads <training_file> and, if option -L is not used, the auxiliary"
	"\nfile <train_aux_file> created by svm-train to write a subset of <training_file>,"
	"\nor the file specified by -l, to <output_file> according to the set options.\n"
	"\nAllowed extensions:\n"
		"<data_files>:  .csv (labeled) and .lsv\n");
	display_help_file_formats();

	if (full_help == false)
		flush_info(INFO_SILENCE, "\nOptions:");
	display_help(ERROR_clp_gen_d);
	display_help(ERROR_clp_ted_c);
	display_help(ERROR_clp_gen_h);
	display_help(ERROR_clp_ted_l);
	display_help(ERROR_clp_ted_L);
	display_help(ERROR_clp_gen_r);
	display_help(ERROR_clp_ted_t);
	
	flush_info(INFO_SILENCE,"\n\n");
	copyright();
	flush_exit(ERROR_SILENT, "");
};


//**********************************************************************************************************************************


void Tcommand_line_parser_extract_data::display_help(unsigned error_code)
{
	Tfold_control fch;
	
	Tcommand_line_parser::display_help(error_code);
	
	if (error_code == ERROR_clp_ted_c)
	{
		display_separator("-c <cell>");
		flush_info(INFO_1, 
		"Takes the samples from the specified <cell> of the task specified by -t.\n");

		display_ranges();
		flush_info(INFO_1, "<cell>:  unsigned integer\n");
		
		display_defaults();
		flush_info(INFO_1, "<cell> = 0\n");
	}
	
	if (error_code == ERROR_clp_ted_l)
	{
		display_separator("-l <data_file>");
		flush_info(INFO_1, 
		"Manipulates the labels before writing to the output file. Two different ways"
		"\nfor the manipulation are available:\n\n");

		flush_info(INFO_1, "<training_file> == <data_file>\n");
		flush_info(INFO_1, "       The samples of the task specified by -t are labeled according to the\n");
		flush_info(INFO_1, "       cells they are contained in. The original ordering gets lost and\n");
		flush_info(INFO_1, "       samples occur more than once, if they are contained in multiple cells.\n\n");
		
		flush_info(INFO_1, "<training_file> != <data_file>\n");
		flush_info(INFO_1, "       The samples of <data_file> are read and labeled according to the\n");
		flush_info(INFO_1, "       smallest cell number they are contained in for the task specified by -t.\n"); 
	}
	
	if (error_code == ERROR_clp_ted_L)
	{
		display_separator("-L <label>");
		flush_info(INFO_1, 
		"Extract only the samples having the specified label. This option ignores all"
		"\nother options.\n\n");
	}
		
	if (error_code == ERROR_clp_ted_t)
	{
		display_separator("-t <task>");
		flush_info(INFO_1, 
		"Take the samples from the cell specified by -c of the task <task>.\n");

		display_ranges();
		flush_info(INFO_1, "<task>:  unsigned integer\n");
		
		display_defaults();
		flush_info(INFO_1, "<task> = 0\n");
	}
}


//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************


int main(int argc, char **argv)
{
	Tcommand_line_parser_extract_data command_line_parser;

	unsigned i;
	unsigned cell;
	vector <unsigned> cells_of_sample;
	Tdataset training_set;
	Tdataset data_set;
	Tdataset working_set;
	Tdataset working_set_tmp;
	Tsubset_info subset_info;
	Tsvm_manager svm_manager;
	Tworking_set_manager working_set_manager;
	FILE* fp_aux_train_read;
	
	double write_time;
	double read_time;
	double full_time;

	
// Read command line

	full_time = get_wall_time_difference();
	
	command_line_parser.setup(argc, argv);
	command_line_parser.parse();



//-------- Load data --------------------------------------------------------------------------------------------

	read_time = get_process_time_difference();
	training_set.read_from_file(command_line_parser.train_file_format);

	if ((command_line_parser.input_file_format.filename != command_line_parser.train_file_format.filename) and (command_line_parser.input_file_format.filename.size() > 0))
		data_set.read_from_file(command_line_parser.input_file_format);
	else
	{
		data_set.enforce_ownership();
		data_set = training_set;
		data_set.enforce_ownership();
	}
	read_time = get_process_time_difference(read_time);
	
	
//-------- get the working set ----------------------------------------------------------------------------------

	
	if (command_line_parser.only_by_label == false)
	{
		svm_manager.load(training_set);

		read_time = get_process_time_difference(read_time);
		fp_aux_train_read = open_file(command_line_parser.aux_filename, "r");
		svm_manager.read_train_aux_from_file(fp_aux_train_read);
		close_file(fp_aux_train_read);
		read_time = get_process_time_difference(read_time);
		
		working_set_manager = svm_manager.get_working_set_manager();

		if (command_line_parser.cells_to_labels == false)
			working_set_manager.build_working_set(working_set, data_set, command_line_parser.task, command_line_parser.cell);
		else
		{
			working_set = data_set;
			working_set.enforce_ownership();
			
			if (command_line_parser.input_file_format.filename != command_line_parser.train_file_format.filename)
				for (i=0; i<working_set.size(); i++)
				{
					cells_of_sample = working_set_manager.cells_of_sample(working_set.sample(i), command_line_parser.task);
					if (cells_of_sample.size() > 0)
						working_set.set_label_of_sample(i, double(cells_of_sample[0]));
					else
						working_set.set_label_of_sample(i, -1.0);
				}
			else
				for (cell=0; cell<working_set_manager.number_of_cells(command_line_parser.task); cell++)
				{
					working_set_manager.build_working_set(working_set_tmp, command_line_parser.task, cell);
					for (i=0; i<working_set_tmp.size(); i++)
						working_set_tmp.set_label_of_sample(i, double(cell)); 
					working_set.push_back(working_set_tmp);
				}
		}
	}
	else
	{
		subset_info = training_set.create_subset_info_with_label(command_line_parser.label);
		training_set.create_subset(working_set, subset_info, true);
	}
	
	
// Write to file

	write_time = get_process_time_difference();
	working_set.write_to_file(command_line_parser.output_file_format);
	write_time = get_process_time_difference(write_time);
	

// Clean up

 	full_time = get_wall_time_difference(full_time);

	flush_info(INFO_1,"\n\n%4.2f seconds used to run create-tt.", full_time);
	flush_info(INFO_1,"\n%4.2f seconds used for read from file operations.", read_time);
	flush_info(INFO_1,"\n%4.2f seconds used for write to file operations.", write_time);


	command_line_parser.copyright();
	
	flush_info(INFO_1,"\n\n");
}


