// Copyright 2017 Ingo Steinwart
//
// This file is part of liquidCluster.
//
// liquidCluster is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as 
// published by the Free Software Foundation, either version 3 of the 
// License, or (at your option) any later version.
//
// liquidCluster is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.

// You should have received a copy of the GNU Affero General Public License
// along with liquidCluster. If not, see <http://www.gnu.org/licenses/>.



#include "../shared/system_support/os_specifics.h"
#include "../shared/basic_functions/flush_print.h"
#include "../shared/basic_types/dataset_info.h"
#include "../shared/basic_types/vector.h"
#include "../shared/system_support/timing.h"
#include "../shared/command_line/command_line_parser.h"



//**********************************************************************************************************************************


unsigned const ERROR_clp_tmc_m = 100;





//**********************************************************************************************************************************


class Tcommand_line_parser_modify_mc: public Tcommand_line_parser
{
	public:
		Tcommand_line_parser_modify_mc();
		void parse();
		
		
		unsigned modification_mode;
		Tsample_file_format read_file_format;
		Tsample_file_format write_file_format;

	protected:
		void exit_with_help();
		void display_help(unsigned error_code);
};


//**********************************************************************************************************************************

Tcommand_line_parser_modify_mc::Tcommand_line_parser_modify_mc()
{
	command_name = "modify-multi-class";
};

//**********************************************************************************************************************************

void Tcommand_line_parser_modify_mc::parse()
{
	check_parameter_list_size();
	for(current_position=1; current_position<parameter_list_size; current_position++)
		if (Tcommand_line_parser::parse("-d-h") == false)
		{
			if(parameter_list[current_position][0] != '-') 
				break;
			if (string(parameter_list[current_position]).size() > 2)
				Tcommand_line_parser::exit_with_help(ERROR_clp_gen_unknown_option);

			switch(parameter_list[current_position][1])
			{
				case 'm':
					modification_mode = get_next_number(ERROR_clp_tmc_m, 0, 0);
					break;
				
				default:
					Tcommand_line_parser::exit_with_help(ERROR_clp_gen_unknown_option);
			}
		}
		
	read_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
	check_labeled_data_format(read_file_format);
	if (current_position < parameter_list_size)
	{
		write_file_format = get_next_data_file_format(ERROR_clp_gen_missing_data_file_name);
		check_labeled_data_format(write_file_format);
	}
	else
		write_file_format = read_file_format;
};


//**********************************************************************************************************************************

void Tcommand_line_parser_modify_mc::exit_with_help()
{
	flush_info(INFO_SILENCE,
	"\n\nchange-labels [options] <read_data_file> [<write_data_file>]\n"
	"\nChanges the labels of the data set contained in <read_data_file> and saves"
	"\nthe results to <write_data_file>. If the latter is missing, the result is"
	"\nwritten to <read_data_file> instead.\n"
	"\nAllowed extensions:\n"
		"<x_data_file>:  .csv (labeled) and .lsv\n");
	display_help_file_formats();

	if (full_help == false)
		flush_info(INFO_SILENCE, "\nOptions:");
	display_help(ERROR_clp_gen_d);
	display_help(ERROR_clp_gen_h);
	display_help(ERROR_clp_tmc_m);
	
	flush_info(INFO_SILENCE,"\n\n");
	copyright();
	flush_exit(ERROR_SILENT, "");
};


//**********************************************************************************************************************************


void Tcommand_line_parser_modify_mc::display_help(unsigned error_code)
{
	Tcommand_line_parser::display_help(error_code);
	
	if (error_code == ERROR_clp_tmc_m)
	{
		display_separator("-c <old_label> <new_label>");
		flush_info(INFO_1, 
		"Change <old_label> to <new_label>.\n");

		display_ranges();
		flush_info(INFO_1, "<old_label>:  real number\n");
		flush_info(INFO_1, "<new_label>:  real number\n");
	}
}


//**********************************************************************************************************************************


vector <double> convert_to_class_probabilities(const Tsample& sample, double eps = 0.0000001)
{
	unsigned i;
	unsigned j;
	unsigned l;
	unsigned s;
	unsigned m;
	unsigned t;
	unsigned dim;
	double Qp;
	double pQp;
	double sum_p;
	double diff;
	double eta_limits;
	vector <double> raw_estimates;
	vector <double> probabilities;
	vector <double> old_probabilities;
	vector <vector <double> > q_matrix;
	vector <vector <double> > r_matrix;
	
	
	dim = sample.dim();
	m = 0.5 + sqrt(2.0 * dim + 0.25);
	raw_estimates = sample.get_x_part(); 
	eta_limits = 0.9999;
	
	
// Create r_matrix with pairwise probability estimates
	
	r_matrix.resize(m);
	for (i=0; i<m; i++)
		r_matrix[i].resize(m);
	
	s=0;
	for (i=0; i<m; i++)
	{
		r_matrix[i][i] = 0.0;
		
		for (j=i+1; j<m; j++)
		{
			raw_estimates[s] = 0.5 * (max(-eta_limits, min(eta_limits, raw_estimates[s])) + 1.0);
			r_matrix[i][j] = 1.0 - raw_estimates[s];
			r_matrix[j][i] = raw_estimates[s];
			s++;
		}
	}

	
// Create q_matrix 
	
	q_matrix.resize(m);
	for (i=0; i<m; i++)
	{
		q_matrix[i].assign(m, 0.0);
		for (j=0; j<m; j++)
		{
			if (i == j)
				for (s=0; s<m; s++)
					q_matrix[i][j] = q_matrix[i][j] + r_matrix[s][i] * r_matrix[s][i];
			else
				q_matrix[i][j] = - r_matrix[j][i] * r_matrix[i][j];
		}
	}

	
// Follow Algorithm 2 of Wu, Lin, and Weng, JMLR 2004
	
	l = 0;
	probabilities.assign(m, 1.0/ double(m));
	do
	{
		old_probabilities = probabilities;
		
		pQp = 0.0;
		for (i=0; i<m; i++)
			for (j=0; j<m; j++)	
				pQp = pQp + probabilities[i] * probabilities[j] * q_matrix[i][j];
		
		sum_p = 0.0;
		for (t=0; t<m; t++)
		{
			Qp = 0.0;
			for (j=0; j<m; j++)
				if (j != t)
					Qp = Qp + q_matrix[t][j] * probabilities[j];
			
			probabilities[t] = (-Qp + pQp) / q_matrix[t][t];
			sum_p = sum_p + probabilities[t];
		}

		diff = 0.0;
		for (t=0; t<m; t++)
		{
			probabilities[t] = probabilities[t] / sum_p;
			diff = diff + abs(old_probabilities[t] - probabilities[t]);
		}
		
		l++;
	}
	while ((l<1000) and (diff / double(m) > eps));
	

	return probabilities;
}





//**********************************************************************************************************************************
//**********************************************************************************************************************************
//**********************************************************************************************************************************


int main(int argc, char **argv)
{
	unsigned i;
	Tcommand_line_parser_modify_mc command_line_parser;
	Tsample sample;
	Tdataset data_set;
	Tdataset new_data_set;
	Tdataset_info data_set_info;
	vector <double> probabilities;
	double label;
	double file_time;
	double full_time;


// Read command line
	
	full_time = get_wall_time_difference();

	command_line_parser.setup(argc, argv);
	command_line_parser.parse();


// Load data set

	file_time = get_process_time_difference();
	data_set.read_from_file(command_line_parser.read_file_format);
	file_time = get_process_time_difference(file_time);
	
	
	new_data_set.enforce_ownership();
	data_set_info = Tdataset_info(data_set, true);
	for (i=0; i<data_set.size(); i++)
	{
		probabilities = convert_to_class_probabilities(data_set.sample(i));
		label = data_set_info.label_list[argmax(probabilities)];
		if (data_set.sample(i)->label != label)
		{
			flush_info(INFO_1, "\nChanging prediction %d to %d for sample %d. Max probability is %1.3f.", unsigned(data_set.sample(i)->label), unsigned(label), i, probabilities[argmax(probabilities)]);
			if (probabilities[argmax(probabilities)] < 3.0 / double(data_set_info.label_list.size()))
			{
				label = data_set.sample(i)->label;
				flush_info(INFO_1, " Not enough confidence.");
			}
		}
		
		sample = Tsample(probabilities, label);
		new_data_set.push_back(sample);
		
	}
	
	
// Save data set
	
	file_time = get_process_time_difference(file_time);
	new_data_set.write_to_file(command_line_parser.write_file_format);
	file_time = get_process_time_difference(file_time);
	flush_info(INFO_1,"\n%d samples written to file %s", data_set.size(), command_line_parser.write_file_format.filename.c_str());


// Clean up

 	full_time = get_wall_time_difference(full_time);

	flush_info(INFO_1,"\n\n%4.2f seconds used to run data-stats.", full_time);
	flush_info(INFO_1,"\n%4.2f seconds used for file operations.", file_time);

	command_line_parser.copyright();
	
	flush_info(INFO_1,"\n\n");
}


